unit CheckOverlay;

interface

uses Classes, ExpPlot, Overlay;

function CheckOverlays(Overlays: TOverlays; S: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; SortBoth, Interpolate: Boolean):Double;

implementation

uses Units, MathSup, SysUtils;

function CheckOverlays(Overlays: TOverlays; S: TStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; SortBoth, Interpolate: Boolean):Double;
var
  i, junk: Integer;
  Ref, Test: TExperimentalPlot;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then
      WriteLn('Point Number ', i, ':', Name);
  end;
  procedure CheckNo(n1, n2, tol: Double; Name: string);
  var
    Err: Double;
  begin
    if isNanOrInf(n2) then
      Fail(Format('%s fails: %g Nan', [Name, n1]))
    else begin
      if tol < 0 then begin
        if Abs(n1-n2) < 1e-15 then
          Err := Abs(n1-n2) { Permit small absolute errors }
        else
          Err := 2*Abs(n1-n2)/(Abs(n1)+Abs(n2));
      end else
        Err := Abs(n1-n2);
      if Err > Abs(tol) then
        Fail(Format('%s out of tolerance: %g %g %g', [Name, n1, n2, Err]));
      if (Err > Result) and (Failures >= 0) then
        Result := Err;
    end;
  end;
begin
  nTried := 0;
  Result := 0;
  Ref := TExperimentalPlot.Create(nil);
  try
    Ref.WeightedLoadFromStream(S, '.dat', False);
    Test := Overlays.Items[0] as TExperimentalPlot;
//    if SortBoth or Interpolate then begin ??Always sorted now
//      Test.Sort;
//      Ref.Sort;
//    end;
    if Test.Units > High(TEnergyUnits) then
      Ref.Sort(False);
    if not Interpolate then begin
      i := -1;
      CheckNo(High(Ref.Frequency), High(Test.Frequency), Tolerance, 'x points');
      CheckNo(High(Ref.Intensity), High(Test.Intensity), Tolerance, 'y points');
      CheckNo(High(Ref.Weights), High(Test.Weights), Tolerance, 'weights');
      for i := 0 to High(Ref.Frequency) do begin
        CheckNo(Convert(Ref.Frequency[i],cm1,Test.Units), Test.Frequency[i], Tolerance, 'Frequency');
        CheckNo(Ref.Intensity[i], Test.Intensity[i], Tolerance, 'Intensity');
        if High(Ref.Weights) > 0 then
          CheckNo(Ref.Weights[i], Test.Weights[i], Tolerance, 'Weights');
        Inc(nTried);
      end;
    end else begin
      for i := 0 to High(Test.Frequency) do begin
        CheckNo(Ref.IntensityAt(Convert(Test.Frequency[i],Test.Units,cm1),junk), Test.Intensity[i], Tolerance, 'Intensity');
        Inc(nTried);
      end;
    end;
  finally
    Ref.Free;
  end;
end;

end.
