unit DatLoader;

interface

uses ExpPlot, TextStrm;

type

  TDatFileLoader = class(TExperimentalPlotLoader)
  private
    S: TTextStream;
    procedure ReadBody;
    function DetectChannels: Integer;
  public
    function Load(AskAboutWeights: Boolean): Boolean; override;
    destructor Destroy; override;
  end;

implementation

uses Math, SysUtils;

{ TDatFileLoader }

destructor TDatFileLoader.Destroy;
begin
  FreeAndNil(S);
  inherited;
end;

function TDatFileLoader.DetectChannels: Integer;
var
  nCol, i, FChan, DChan: Integer;
  Changes: array of Integer;
  LastValue: array of Double;
  Value: Double;
  MinCols: Integer; { Number of complete rows of data }
begin
  MinCols := 0;
  Result := 0;
  { Autodetect frequency channel by counting number of changes of each sign in each channel }
  while not S.AtEndOfFile do begin
    nCol := 0;
    while not S.AtEndOfLine do
      if S.GetFreeItem then begin
        if not S.IsRealToken(Value) then
          Break;
        if nCol >= Length(ChannelDest) then begin
          SetLength(ChannelDest, nCol+1);
          SetLength(Changes, nCol+1);
          SetLength(LastValue, nCol+1);
          Changes[nCol] := 0;
          LastValue[nCol] := Value;
        end else begin
          Value := S.RealToken;
          Changes[nCol] := Changes[nCol] + Sign(Value-LastValue[nCol]);
          LastValue[nCol] := Value;
        end;
        Inc(nCol);
      end;
    S.NextLine;
    if nCol > 0 then begin
      Inc(Result);
      if MinCols = 0 then
        MinCols := nCol
      else
        MinCols := Min(nCol, MinCols);
    end;
  end;
  SetLength(ChannelDest,MinCols);
  if MinCols < 1 then
    Exit;
  { Channel with the largest next changes is probably the frequency channel }
  FChan := 0;
  for i := 1 to High(ChannelDest) do
    if Abs(Changes[i]) > Abs(Changes[FChan]) then
      FChan := i;
  DChan := 1;
  for i := 0 to High(ChannelDest) do
    if i = FChan then
      ChannelDest[i] := 0 { Frequency channel }
    else begin
      ChannelDest[i] := DChan; { Default data channel is first non-frequency channel }
      DChan := -1;
    end;
end;

function TDatFileLoader.Load(AskAboutWeights: Boolean): Boolean;
var
  N: Integer;
begin
  S := TTextStream.Create(Stream);
  S.ItemSeparators := [isTab, isComma];
  { Determine number of Columns and Rows of data }
  N := DetectChannels;
  Result := False;
  if (N = 0) or (Length(ChannelDest) = 0) then
    Exit;
  CheckAvailableChannels(Length(ChannelDest), N, AskAboutWeights);
  ReadBody;
  if Length(ChannelDest) = 1 then
    ExperimentalPlot.IsLineList := True;
  Result := True;
end;

procedure TDatFileLoader.ReadBody;
var
  i, nCol: Integer;
  Value: Double;
begin
  { And repeat read logic, this time setting arrays }
  S.Position := 0;
  i := 0;
  while not S.AtEndOfFile do begin
    nCol := 0;
    while not S.AtEndOfLine do
      if S.GetFreeItem then begin
        if not S.IsRealToken(Value) then
          Break;
        SetValue(i, nCol, Value);
        Inc(nCol);
      end;
    S.NextLine;
    if nCol > 0 then
      Inc(i);
  end;
end;

end.
