unit ExplicitTest;

interface

uses Mixture;

procedure DoExplicitTest(Eigenvalues: TEigenvalues);

implementation

uses SysUtils, QuantumNumber, HundA, Math;

{ Explicit equations for triplet sigma - states }
{ Checked with the help of Maple against generated matrices and
  Expressions given by Rouille at al, JMS 154 372 (1992). }
function TryTripletSigmaEnergy(HundA: THundA; J: TQuantumNumber; Index: Integer; var E: Double):Boolean;
var
  t1,t2,t3,t4,t5,t6,t7,t8,t9,t10,t11,t12,t15,t16,
  t17,t18,t19,t20,t21,t23,t25,t26,t28,t29,t31: Double;
  av, df, AvJ, DvJ: Double;
begin
  Result := False;
  if not ((HundA.S = 2) and (HundA.Lambda=-1) and HundA.gerade) then
    Exit;
  Result := True;
  J := J div 2;
  with HundA do
    if Odd(J) then begin
      t1 := J*J;
      t2 := t1*J;
      t5 := t1*t1;
      t26 := t1*B-2.0*D*t2-D*t1+2.0/3.0*t1*lambdaD-t1*gammaD+2.0/3.0*J*lambdaD
             -J*gammaD-gamma+2.0/3.0*lambdaSS-D*t5+J*B;
      E := t2*H-2.0*t2*gammaH+t5*t1*H+3.0*t5*J*H+3.0*t5*H-t1*gammaH+2.0/3.0*t5*
           lambdaH+4.0/3.0*t2*lambdaH+2.0/3.0*t1*lambdaH-t5*gammaH+t26;
    end else begin
//      t1 := J*J;
//      t4 := t1*J;
//      t6 := t1*t1;
//      t21 := -t1*lambdaD/3-7.0/2.0*t1*gammaD-11.0*t4*gammaH-11.0/2.0*t6*gammaH
//             -39.0/2.0*t1*gammaH-14.0*J*gammaH+3.0*t6*J*H+18.0*t6*H+31.0*t4*H+t6*t1*H+33.0*t1*
//             H+18.0*J*H-lambdaSS/3-J*lambdaD/3-4.0*J*lambdaH-2.0/3.0*t4*lambdaH;
//      t31 := -t6*lambdaH/3-13.0/3.0*t1*lambdaH-7.0/2.0*J*gammaD-D*t6+t1*B-2.0
//             *D-3.0/2.0*gamma+4.0*H-6.0*D*J+J*B-2.0*D*t4-7.0*D*t1+B-4.0/3.0*lambdaD-8.0/3.0*
//             lambdaH-2.0*gammaD-4.0*gammaH;
//      av := t21+t31;

      av := J*B+J*J*B+B
            -6*J*D-2*D-IntPower(J,4)*D-7*J*J*D-2*J*J*J*D
            +IntPower(J,6)*H +3*IntPower(J,5)*H+18*IntPower(J,4)*H+31*J*J*J*H+33*J*J*H
              +18*J*H +4*H
            -lambdaSS/3
            -4/3*lambdaD-J*lambdaD/3-J*J*lambdaD/3
            -IntPower(J,4)*lambdaH/3 -2/3*J*J*J*lambdaH -13/3*J*J*lambdaH -4*J*lambdaH
              -8/3*lambdaH
            -3/2*gamma
            -7/2*J*gammaD -2*gammaD- 7/2*J*J*gammaD
            -4*gammaH -14*J*gammaH -39/2*J*J*gammaH -11*J*J*J*gammaH -11/2*IntPower(J,4)*gammaH;

//      t1 := J*gammaD;
//      t3 := J*H;
//      t4 := J*lambdaH;
//      t5 := J*gammaH;
//      t6 := J*J;
//      t7 := t6*gammaH;
//      t8 := t6*J;
//      t9 := t8*gammaH;
//      t10 := t6*t6;
//      t11 := t10*gammaH;
//      t12 := t6*lambdaH;
//      t15 := -8.0*H+t1+2.0*J*lambdaD+4.0*gammaD-20.0*t3+40.0/3.0*t4+12.0*t5+
//             13.0*t7+2.0*t9+t11+46.0/3.0*t12+4.0*t8*lambdaH+2.0*t10*lambdaH;
//      t16 := t10*H;
//      t17 := t6*H;
//      t18 := t8*H;
//      t19 := D*J;
//      t20 := D*t6;
//      t21 := t6*gammaD;
//      t23 := -6.0*t16-26.0*t17-12.0*t18+8.0*gammaH+gamma+4.0*D-2.0*B+4.0*t19+4.0*t20
//             +2.0*lambdaSS+t21+2.0*t6*lambdaD+16.0/3.0*lambdaH+8.0/3.0*lambdaD;
//      t25 := IntPower(t15+t23,2);
//      t28 := -6.0*B+3.0*gamma+12.0*t20+12.0*t19+12.0*D+2.0*lambdaD+12.0*gammaD+3.0
//             *t21+3.0*t1-18.0*t16-36.0*t18;
//      t29 := -78.0*t17-60.0*t3-24.0*H+4.0*t12+4.0*t4+4.0*lambdaH+39.0*t7+36.0*
//             t5+24.0*gammaH+3.0*t11+6.0*t9;
//      t31 := IntPower(t28+t29,2);
//      df := t25/4+J*(J+1.0)*t31/9;
//      if Index = 1 then
//        E := av - Sqrt(df)
//      else
//        E := av + Sqrt(df);

      AvJ := (2.0*J+1)*(
               B-2*D*(J*J+J+1)
               +H*(3*IntPower(J,4)+6*J*J*J+13*J*J+10*J+4)
               -gamma/2-gammaD*(J*J+J+4)/2
               -gammaH*(IntPower(J,4)+2*J*J*J+13*J*J+12*J+8)/2)
             -1/(2*J+1)*(
                 lambdaSS+lambdaD*(7*J*J+7*J+4)/3
                 +lambdaH*(11*IntPower(J,4)+22*J*J*J+39*J*J+28*J+8)/3);

      DvJ := 2/(2*J+1)*(lambdaSS+lambdaD*(J*J+J+1)
                        +lambdaH*(IntPower(J,4)+2*J*J*J+7*J*J+6*J+2));
      df := Sqrt(Sqr(AvJ)+J*(J+1)*Sqr(DvJ));
      if Index = 1 then
        E := av - df
      else
        E := av + df;
    end;
end;

procedure DoExplicitTest(Eigenvalues: TEigenvalues);
var
  i: Integer;
  G, E: Double;
begin
  if Eigenvalues.Size = 0 then
    Exit;
  if (Eigenvalues.Manifold is TLinearManifold)
     and (Eigenvalues.Manifold.nStates = 1)
     and TryTripletSigmaEnergy(Eigenvalues.Manifold.State[0] as THundA, Eigenvalues.J, 1, E)
   then
    for i := 1 to Eigenvalues.Size do begin
      TryTripletSigmaEnergy(Eigenvalues.Manifold.State[0] as THundA, Eigenvalues.J, i, E);
      G := Eigenvalues.E[i];
      if not SameValue(G, E) then
        WriteLn(Eigenvalues.J, ' ', Eigenvalues.Sym, ' ', i, ' ', G, E, G-E);
    end;
end;

end.
