unit JDXLoader;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses ExpPlot, Classes;

type

  TJDXFileLoader = class(TExperimentalPlotLoader)
  private
    c: Char;
    EOF: Boolean;
    CharPending: Boolean;
    PendingChar: Char;
    StreamPosition: Int64;
    SavedChar: Char;
    NLChar: Char;
    function IsSpace(Ch: Char): Boolean;
    function IsDigit(Ch: Char): Boolean;
    function NextC: Char;
    function PeekChar: Char;
    procedure SaveStreamPosition;
    procedure RestoreStreamPosition;
  private
    LastLongValue, DUPoutstanding: Integer;
    LastValue, FirstX, LastX, xFactor, yFactor, LastDiff: Double;
    LastWasDiff, HadNewline, HadPeaks, HadData: Boolean;
    function DXAddToTitle: string;
    function AFFN: string;
    function AFFNtail(Start: string): string;
    function AFFNSkipnl: string;
    function ASDF:Boolean;
    procedure DiscardDXComments;
    function DXdiscardable(Ac: Char):Boolean;
    procedure DXexpect(e: Char);
    function DXGetItem(Term: Char): string;
    procedure DXNextLine;
    function GetDATA_LABEL: string;
    function IsAFFNterm(Ch: Char):Boolean;
    procedure WantAFFN(Item: string);
  public
    function Load(AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Units, SysUtils, Math, MathSup;

{ TJDXFileLoader }

function TJDXFileLoader.DXAddToTitle: string;
var
  FirstNonBlank, NextMarkerSeen: Boolean;
begin
  Result := '';
  while not EOF and (c in [' ',#9]) do
    NextC;
  DiscardDXComments;
  NextMarkerSeen := False;
  repeat
    { For each line }
    FirstNonBlank := True;
    while not EOF and not (c in [#10, #13]) do begin
      if not IsSpace(c) then begin
        if FirstNonBlank and (c = '#') and (PeekChar = '#') then begin
          NextMarkerSeen := True;
          Break;
        end;
        FirstNonBlank := False;
      end;
      Result := Result + c;
      NextC;
      DiscardDXComments();
    end;
    if NextMarkerSeen then Break;
    if PeekChar in [#10, #13] then
      NextC;
    Result := Result + ' ';
    NextC;
    DiscardDXComments();
  until EOF or NextMarkerSeen;
  Result := Trim(Result);
  if ExperimentalPlot.Comment <> '' then
    Result := ' ' + Result;
  ExperimentalPlot.Comment := ExperimentalPlot.Comment + Result;
end;

function TJDXFileLoader.AFFNSkipnl: string;
begin
  repeat
    DiscardDXComments;
    if EOF or not IsSpace(c) then
      Break;
    NextC;
  until False;
  Result := AFFNtail('');
end;

function TJDXFileLoader.AFFNtail(Start: string): string;
var
  Exponent: string;
  nDigits: Integer;
  LastType: TArrayType;
begin
  LastType := tyI16;
  Result := Start;
  if not EOF and (c in ['+','-']) then begin
    Result := Result + c;
    NextC;
  end;
  while not EOF and ( (c in ['0'..'9']) or ((LastType = tyI16) and (c = '.')) ) do begin
    if c = '.' then
      LastType := tyDouble;
    Result := Result + c;
    NextC;
  end;
  if Result = '' then
    Exit;
  { look for E+nn or E+nnn only (sign must be present; 2 or 3 digits only)
    allow 1 digit and lower case e as an extension }
  SaveStreamPosition;
  if not EOF and (c in ['E', 'e']) and (PeekChar in ['+','-']) then begin
    Exponent := 'E' + NextC;
    NextC;
    nDigits := 0;
    while IsDigit(c) and (nDigits <= 3) do begin
      Exponent := Exponent + c;
      NextC;
      Inc(nDigits);
    end;
    if (nDigits in [1..3]) and IsAFFNterm(c) then begin
      Result := Result+Exponent;
      LastType := tyDouble;
    end else
      RestoreStreamPosition;
  end else if LastType <> tyDouble then begin
    if TryStrToInt(Result, LastLongValue) then begin
      LastValue := LastLongValue;
      if (LastLongValue < Low(SmallInt)) or (LastLongValue > High(SmallInt)) then
        LastType := tyI32;
    end else
      LastType := tyDouble;
  end;
  if LastType = tyDouble then
    if not TryStrToFloat(Result, LastValue) then
      FileErr('Bad number:' + Result);
  while not EOF and (c in [' ',#9]) do
    NextC;
  LastWasDiff := False;
end;

procedure TJDXFileLoader.DiscardDXComments;
begin
  if not EOF and (PeekChar = '$') then
    repeat
      NextC;
    until EOF or (c = #10);
end;

function TJDXFileLoader.DXdiscardable(Ac: Char): Boolean;
begin
  begin
    Result := Ac in [' ', #9, '-', '/', '_'];
  end;
end;

procedure TJDXFileLoader.DXexpect(e: Char);
begin
  while not EOF and DXdiscardable(c) do
    NextC;
  if EOF then
    raise Exception.Create('File too short');
  if c <> e then begin
    if c = #0 then c := '?';
    FileErr(Format('Expecting %s found %s', [e, c]));
  end;
  while not EOF and ( c <> e ) do
    NextC; { Look for one anyway }
  NextC; { and discard it }
end;

function TJDXFileLoader.DXGetItem(Term: Char): string;
begin
  Result := '';
  repeat
    while not EOF and DXdiscardable(c) do
      NextC;
    DiscardDXComments;
    if EOF or (c = Term) or (c = #10) then
      Exit;
    if not IsSpace(c) then
      Result := Result + c;
    NextC;
  until EOF;
end;

procedure TJDXFileLoader.DXNextLine;
begin
  while not EOF and not (c in [#10, #13]) do
    NextC;
end;

function TJDXFileLoader.GetDATA_LABEL: string;
begin
  DiscardDXComments;
  while not EOF and IsSpace(c) do
    NextC;
  if EOF then
    raise Exception.Create('File too short');
  DXexpect('#');
  DXexpect('#');
  Result := DXGetItem('=');
  if c <> '=' then
    FileErr('= not found after ' + Result)
  else
    NextC;
end;

function TJDXFileLoader.IsAFFNterm(Ch: Char): Boolean;
begin
  Result := (Ch in [#0,',']) or IsSpace(Ch);
end;

function TJDXFileLoader.IsDigit(Ch: Char): Boolean;
begin
  Result := Ch in ['0'..'9'];
end;

function TJDXFileLoader.IsSpace(Ch: Char): Boolean;
begin
  Result := Ch in [' ', #9, #10, #13];
end;

function TJDXFileLoader.Load(AskAboutWeights: Boolean): Boolean;
var
  Item, DataForm: string;
  x, y, yValue, dx: Double;
  nPoints, i, nCh: Integer;
  LastyWasDiff, PeakTable: Boolean;
  buf: string[6];
  Section: string;
  procedure SkipPunct;
  begin
    if c in [',',';'] then
      NextC;
  end;
  procedure CheckPoints;
  begin
    if i = nPoints then Exit;
    FileErr('Wrong number of points' + Section);
    SetLength(ExperimentalPlot.Frequency, i);
    SetLength(ExperimentalPlot.Intensity, i);
    if Length(ExperimentalPlot.Weights) > 0 then
      SetLength(ExperimentalPlot.Weights, i);
  end;
  procedure AddFrequency(V: Double);
  begin
    if i >= nPoints then begin
      FileErr('Too many points read' + Section);
      Exit;
    end;
    ExperimentalPlot.Frequency[i] := V*FrequencyFactor;
  end;
begin
  NLChar := #0;
  Result := False;
  EOF := False;
  CharPending := False;
  LineNumber := 1;
  { Defaults }
  SetNan(FirstX);
  SetNan(LastX);
  xFactor := 1;
  yFactor := 1;
  HadNewline := False;
  HadPeaks := False;
  HadData := False;
  DUPoutstanding := 0;
  nPoints := -1;

  if Ext = '' then begin
    SaveStreamPosition;
    NextC;
    while not EOF and IsSpace(c) do
      NextC;
    if (c = '#') and (PeekChar = '#') then begin
      NextC;
      SetLength(buf,Stream.Read(buf[1], 6));
      if AnsiCompareText(buf,'TITLE=') = 0 then
        Ext := '.jdx';
    end;
    RestoreStreamPosition;
  end;
  if (Ext <> '.jdx') and (Ext <> '.dx') and (Ext <> '.jcm') then
    Exit;
  NextC;
  Result := True;
  ExperimentalPlot.Comment := '';
  Item := GetDATA_LABEL;
  if AnsiCompareText(Item, 'TITLE') <> 0 then
    FileErr('Expected TITLE found ' + Item)
  else
    DXAddToTitle;
  Item := GetDATA_LABEL;
  if AnsiCompareText(Item, 'JCAMPDX') <> 0 then
    FileErr('Expected JCAMPDX found ' + Item)
  else begin
    DXNextLine; // Don't bother checking for 4.24
    Item := GetDATA_LABEL;
  end;
  while AnsiCompareText(Item, 'END') <> 0 do begin
    if AnsiCompareText(Item, 'DATATYPE') = 0 then begin
      Item := DXAddToTitle;
      if AnsiCompareText(Item, 'LINK') = 0 then
        raise Exception.Create('LINK blocks not implemented');
    end else if AnsiCompareText(Item, 'BLOCK') = 0 then
      raise Exception.Create('LINK blocks not implemented')
    else if AnsiCompareText(Item, 'NTUPLES') = 0 then
      raise Exception.Create('NTUPLES not implemented')
    else if (AnsiCompareText(Item, 'ORIGIN') = 0)
            or (AnsiCompareText(Item, 'YUNITS') = 0)
            or (AnsiCompareText(Item, 'OWNER') = 0) then
      DXAddToTitle
    else if AnsiCompareText(Item, 'XUNITS') = 0 then begin
      Item := AnsiUpperCase(DXGetItem(#0));
      if Item = '1/CM' then
        IndicateUnits(cm1)
      else if Item = 'NANOMETERS' then begin
        IndicateUnits(nmAir);
      end else if Item = 'MICROMETERS' then begin
        IndicateUnits(nmVac);
        FrequencyFactor := 1e-3;
      end else if Item = 'MHZ' then
        IndicateUnits(MHz)
      else if Item = 'KHZ' then begin
        IndicateUnits(MHz);
        FrequencyFactor := 1e-3;
      end else if Item = 'HZ' then begin
        IndicateUnits(MHz);
        FrequencyFactor := 1e-6;
      end else if Item = 'GHZ' then begin
        IndicateUnits(MHz);
        FrequencyFactor := 1e3;
      end;
    end else if AnsiCompareText(Item, 'NPOINTS') = 0 then begin
      WantAFFN(Item);
      nPoints := Round(LastValue);
    end else if AnsiCompareText(Item, 'FIRSTX') = 0 then begin
      WantAFFN(Item);
      FirstX := LastValue;
    end else if AnsiCompareText(Item, 'LASTX') = 0 then begin
      WantAFFN(Item);
      LastX := LastValue;
    end else if AnsiCompareText(Item, 'XFACTOR') = 0 then begin
      WantAFFN(Item);
      xFactor := LastValue;
    end else if AnsiCompareText(Item, 'YFACTOR') = 0 then begin
      WantAFFN(Item);
      yFactor := LastValue;
    end else if (AnsiCompareText(Item, 'XYDATA') = 0)
                or (AnsiCompareText(Item, 'XYPOINTS') = 0)
                or (AnsiCompareText(Item, 'PEAKTABLE') = 0) then begin
      PeakTable := AnsiCompareText(Item, 'PEAKTABLE') = 0;
      if PeakTable then begin
        if HadPeaks then
          FileErr('More than one ' + Item);
        HadPeaks := True;
      end else begin
        if HadData then
          FileErr('More than one data block (' + Item +')');
        HadData := True;
      end;
      DataForm := DXGetItem(#0);
      if AnsiCompareText(DataForm, '(X++(Y..Y))') = 0 then begin
        CheckAvailableChannels(2, nPoints, AskAboutWeights);
        Section := ' for ' + Item + ' ' + DataForm;
        i := 0;
        y := 0;
        LastyWasDiff := False;
        if IsNan(FirstX) then
          FileErr('FIRSTX missing' + Section);
        x := FirstX;
        if IsNan(LastX) then begin
          FileErr('LASTX missing' + Section);
          LastX := 1000;
        end;
        if npoints <= 1 then begin
          FileErr('NPOINTS missing' + Section);
          dx := 1;
        end else
          dx := (LastX - FirstX) / (nPoints - 1);
        while ASDF do begin
          if HadNewline then begin
            LastValue := LastValue * xFactor;
            if LastyWasDiff and (i > 0) then
              x := x - dx;
            if not SameValue(x, LastValue)
               and (Abs(x - lastvalue) >= Max(Abs(dx),1/xFactor)*0.9) then
                FileErr('X-Value check fails' + Section);
            { x := lastvalue; Spec says use value from firstx }
            if not ASDF then begin
              FileErr('Missing y value' + Section);
              Continue;
            end else begin
              yValue := LastValue * yFactor;
              if LastyWasDiff and (i > 0) then begin
                { Y-VALUE CHECK }
                if LastWasDiff then
                  FileErr('Y value check not present' + Section);
                LastyWasDiff := LastWasDiff;
                if not SameValue(y, yValue) then begin
                  FileErr('Y value check fails' + Section);
                  yValue := y;
                end;
                x := x + dx;
                Continue;
              end;
            end
          end else
            yValue := LastValue * yFactor;
          LastyWasDiff := LastWasDiff;
          y := yValue;
          if PeakTable then
            ExperimentalPlot.AddPeak(x,ExperimentalPlot.Units,False)
          else begin
            AddFrequency(x);
            if i < nPoints then
              ExperimentalPlot.Intensity[i] := y;
          end;
          Inc(i);
          x := x + dx;
        end;
        CheckPoints;
      end else if (AnsiCompareText(DataForm, '(XY..XY)') = 0)
                  or (AnsiCompareText(DataForm, '(XYZ..XYZ)') = 0) then begin
        nCh := 2;
        if AnsiCompareText(DataForm, '(XY..XY)') <> 0 then
          Inc(nCh);
        i := 0;
        if not PeakTable then
          CheckAvailableChannels(nCh, nPoints, AskAboutWeights);
        while AFFNSkipnl <> '' do begin
          if PeakTable then
            ExperimentalPlot.AddPeak(LastValue,ExperimentalPlot.Units,False)
          else
            AddFrequency(LastValue);
          SkipPunct;
          WantAFFN(Item);
          if not PeakTable and (i < nPoints) then
            ExperimentalPlot.Intensity[i] := LastValue;
          if nCh = 3 then begin
            SkipPunct;
            WantAFFN(Item);
            if (High(ExperimentalPlot.Weights) > 0) and (i < nPoints) then
              ExperimentalPlot.Weights[i] := LastValue;
          end;
          SkipPunct;
          Inc(i);
        end;
        CheckPoints;
      end else
        raise Exception.Create('Bad dataform' + Section);
    end else begin
      { Unwanted LDR }
      repeat
        DXNextLine;
        while not EOF and IsSpace(c) do NextC;
        while not EOF and DXdiscardable(c) do NextC;
      until EOF or (c = '#');
    end;
    Item := GetDATA_LABEL;
  end;
end;

function TJDXFileLoader.NextC: Char;
begin
  if CharPending then begin
    CharPending := False;
    Result := PendingChar;
  end else
    EOF := Stream.Read(Result, 1) < 1;
  if NLChar = #0 then begin
    if c in [#10,#13] then
      NLChar := c;
  end;
  if (c = NLChar) and (NLChar <> #0) then
    Inc(LineNumber);
  c := Result;
end;

function TJDXFileLoader.PeekChar: Char;
begin
  if not CharPending then
    CharPending := Stream.Read(PendingChar, 1) = 1;
  if CharPending then
    Result := PendingChar
  else
    Result := #0;
end;

procedure TJDXFileLoader.RestoreStreamPosition;
begin
  Stream.Position := StreamPosition;
  CharPending := False;
  c := SavedChar;
end;

procedure TJDXFileLoader.SaveStreamPosition;
begin
  StreamPosition := Stream.Position;
  if CharPending then
    Dec(StreamPosition);
  SavedChar := c;
end;

procedure TJDXFileLoader.WantAFFN(Item: string);
begin
  if AFFNSkipnl = '' then
    FileErr('Missing ' + Item);
end;

function TJDXFileLoader.ASDF: Boolean;
var
  r: Double;
  procedure Diff(s: string);
  begin
    r := LastValue;
    NextC;
    AFFNtail(s);
    LastDiff := LastValue;
    LastWasDiff := True;
    LastValue := r + LastValue;
  end;
  function Dup(Ch: Char):Boolean;
  begin
    DUPoutstanding := Ord(Ch) - Ord('S') + 1;
    repeat
      NextC;
      if not EOF and IsDigit(c) then
        DUPoutstanding := DUPoutstanding*10 + Ord(c) - Ord('0')
      else
        Break;
    until False;
    Dec(DUPoutstanding);
    Result := ASDF();
  end;
  procedure Tail(s: string);
  begin
    NextC;
    AFFNtail(s);
  end;
begin
  HadNewline := False;
  Result := True;
  if DUPoutstanding <> 0 then begin
    Dec(DUPoutstanding);
    if LastWasDiff then
      LastValue := LastValue + LastDiff;
    Exit;
  end;
  while not EOF and IsSpace(c) do begin
    HadNewline := HadNewline or (c = #10);
    NextC;
  end;
  case c of
    '+','-','0'..'9', '.': AFFN;
    '@':      Tail('0');
    'A'..'I': Tail(Char(Ord(c) - Ord('A') + Ord('1')));
    'a'..'i': Tail('-' + Char(Ord(c) - Ord('a') + Ord('1')));
    'j'..'r': Diff('-' + Char(Ord(c) - Ord('j') + Ord('1')));
    '%':      Diff('0');
    'J'..'R': Diff(Char(Ord(c) - Ord('J') + Ord('1')));
    's':      Result := Dup(Char(Ord('S')+8));
    'S'..'Z': Result := Dup(c);
    '?':      begin
                LastWasDiff := False;
                LastValue := 0; { TODO: Could consider deleting point.  NaN would not be helpful }
                NextC;
              end;
    '$':      begin
                if NextC <> '$' then
                  FileErr('Invalid char after $:' + c)
                else
                  DXNextLine;
                Result := ASDF();
              end;
    else
      Result := False;
  end;
end;

function TJDXFileLoader.AFFN: string;
begin
  while not EOF and (c in [' ',#9]) do
    NextC;
  Result := AFFNtail('');
end;

initialization
  RegisterPlotLoader(TJDXFileLoader, 'JCAMP-DX Files', '*.dx;*.jdx;*.jcm');
end.
