unit MNodeClip;

{$I DFS.inc}

interface

uses
  Classes, SysUtils, MNode, DOM;

function MNodeClipboardFormat: Cardinal;
function ReadXMLMNodeFromClipboard: TXMLDocument;
function PasteXMLNodeOnto(TargetMNode: TMNode):TMNode;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Huge, Mixture, ClipBrd, XMLRead, RenForm;

var
  MNodeClipboardFormatVar: Cardinal = 0;

function MNodeClipboardFormat: Cardinal;
begin
  if MNodeClipboardFormatVar = 0 then
    MNodeClipboardFormatVar := RegisterClipboardFormat('MNode');
  Result := MNodeClipboardFormatVar;
end;

function ReadXMLMNodeFromClipboard: TXMLDocument;
var
  Stream: THugeMemoryStream;
begin
  Result := nil;
  Stream := THugeMemoryStream.Create;
  try
    Stream.SetFromClipboard(MNodeClipboardFormat);
    ReadXMLFile(Result, TStream(Stream), '<Clipboard>');
  finally
    Stream.Free;
  end;
end;

function PasteXMLNodeOnto(TargetMNode: TMNode):TMNode;
var
  MNode, NewMNode: TMNode;
  XMLDocument: TXMLDocument;
  Child: TDOMNode;
  NewSubClass: TMNodeClass;
  NameOK: Boolean;
  i: Integer;
  OldName: string;
begin
  Result := nil;
  if Assigned(TargetMNode) and Clipboard.HasFormat(MNodeClipboardFormat) then begin
    XMLDocument := ReadXMLMNodeFromClipboard;
    try
      Child := XMLDocument.FirstChild;
      if Assigned(Child) and (Child.NodeName = 'List') then
        Child := Child.FirstChild;
      OldName := '';
      while Assigned(Child) do begin
        NewSubClass := ReadMNodeClass(Child, nil);
        MNode := nil;
        if Assigned(TargetMNode.Owner)
           and (TargetMNode.Owner is NewSubClass.OwnerType) then
          MNode := TargetMNode.Owner as TMNode
        else if TargetMNode is NewSubClass.OwnerType then
          MNode := TargetMNode
        else if NewSubClass.InheritsFrom(TTransitionMoment) then begin
          { Find Molecule }
          MNode := TargetMNode;
          while Assigned(MNode) and not (MNode is TMolecule) do
            MNode := MNode.Owner as TMNode;
          if Assigned(MNode) then
            MNode := MNode.ChildWithClass(TTransitionMoments)
        end;
        if not Assigned(MNode) then
          raise Exception.Create('Nowhere to paste ' + NewSubClass.ClassName + '(' + TargetMNode.ClassName + ')');
        NewMNode := MNode.ReadMNode(Child);
        NameOk := True;
        { TODO: Make unique name }
        for i := 0 to MNode.ComponentCount-1 do
          if (NewMNode <> MNode.Items[i])
             and (NewMNode.Name = MNode.Items[i].Name) then begin
            NameOK := False;
            Break;
          end;
        if (NewMNode.Name = '') and (OldName <> '') and (NewMNode is TMNodePair) then
          with TMNodePair(NewMNode) do begin
            if FBraName = OldName then
              FBraName := 'Copy of ' + OldName;
            if FKetName = OldName then
              FKetName := 'Copy of ' + OldName;
          end
        else if (NewMNode.Name <> '') and not NameOK then begin
          if OldName = '' then
            OldName := NewMNode.Name;
          NewMNode.Name := 'Copy of ' + NewMNode.Name;
        end;
        try
          RenameForm.RenameMNodePairsAsRequired(NewMNode);
        except
          NewMNode.Free;
          raise;
        end;
        if not Assigned(Result) then
          Result := MNode;
        Child := Child.NextSibling;
      end;
    finally
      XMLDocument.Free;
    end;
  end;
end;

end.

