unit MNodeColour;
{ Helper functions for MNode Colour Properties }
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics;

function GetColourNames(var LowestValue: Integer): PPChar; { for MNode.GetSettingsNames }
function ColourToStr(Colour: TColor): string;
function StrToColour(const s: string): TColor;

const
MNODEColourValues: array [0..17] of TColor =
 ( clNone, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clGray, clSilver, clRed,
   clLime, clYellow, clBlue, clFuchsia, clAqua, clBlack, clWhite, clDefault);


implementation

uses {$IFNDEF CONSOLE} Dialogs, Forms, Graphs, {$ENDIF}
     SysUtils;

const
  ColourNames: array [0..18] of PChar =
   ( 'None', 'Maroon', 'Green', 'Olive', 'Navy', 'Purple', 'Teal', 'Gray', 'Silver', 'Red',
     'Lime', 'Yellow', 'Blue', 'Fuchsia', 'Aqua', 'Black', 'White', 'Other...', nil);

var
  LastColour: TColor;

function GetColourNames(var LowestValue: Integer): PPChar;
{ for MNode.GetSettingsNames }
begin
  LowestValue := 0;
  Result := @ColourNames[0];
end;

function ColourToStr(Colour: TColor): string;
var
  i: Integer;
begin
  LastColour := Colour;
  for i := 0 to High(MNodeColourValues) do
    if Colour = MNodeColourValues[i] then begin
      Result := ColourNames[i];
      Exit;
    end;
  Result := Format('$%.6x', [Colour]);
  if (Colour and not clWhite) <> 0 then
    raise Exception.Create('Bad Colour:' + Result);
end;

function StrToColour(const s: string): TColor;
var
  i: Integer;
  {$IFNDEF CONSOLE}
  Dialog: TColorDialog;
  {$ENDIF}
begin
  if s = '' then
    Result := clNone
  else if Length(s) <= 2 then begin
    { Backward compatibility only }
    i := StrToInt(s);
    if i = 0 then
      Result := clNone
    else
      Result := {$IFDEF CONSOLE} clBlack {$ELSE} default_pen_colour[i] {$ENDIF};
  end else begin
    for i := 0 to High(MNodeColourValues)-1 do
      if s = ColourNames[i] then begin
        Result := MNodeColourValues[i];
        Exit;
      end;
    if TryStrToInt(s, i) then
      Result := i
    else
      Result := clDefault;
    if (Result and not clWhite) <> 0 then begin
      {$IFNDEF CONSOLE}
      Dialog := TColorDialog.Create(Application);
      Result := LastColour;
      try
        Dialog.Color := LastColour;
        if Dialog.Execute then
          Result := Dialog.Color;
      finally
        Dialog.Free;
      end;
      {$ELSE}
      WriteLn('Bad Colour Name:'+s); { Don't crash as probably harmless in concole }
      {$ENDIF}
    end;
  end;
end;

end.
