unit StateInfo;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, Split, Grids, NGrid, StdCtrls, AlEdit, EditExtr, NewSpin,
  ExtCtrls, ChildForm, MNode;

type
  TStatesForm = class(TChildForm)
    EigLabel: TLabel;
    LabelPanel: TPanel;
    StateComboBox: TComboBox;
    JSpinBox: TSpinBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    SymSpinBox: TSpinBox;
    HamGrid: TNumberGrid;
    HamiltonianLabel: TLabel;
    HamPanel: TPanel;
    Split: TSplit;
    EigPanel: TPanel;
    EigenValuesLabel: TLabel;
    EigGrid: TNumberGrid;
    PerturbationsCheckBox: TCheckBox;
    HamLabel: TLabel;
    procedure EigGridGetCellLabel(Sender: TObject; Col, Row: Integer;
      var s: string; var IsText: Boolean);
    procedure JSpinBoxChange(Sender: TObject);
  protected
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
  end;

var
  StatesForm: TStatesForm;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     MatrixVec, Mixture;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

{ TStatesForm }

function TStatesForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TStatesForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  JSpinBoxChange(nil);
end;

procedure TStatesForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  JSpinBoxChange(nil);
end;

procedure TStatesForm.ShowItem(MNode: TMNode);
var
  Manifold: TManifold;
  State: TState;
  EigenValues: TEigenvalues;
  i, ii, Woff: Integer;
  M: TMatrix;
  Sym: Integer;
begin
  if not Assigned(MNode) then Exit;
  BeginUpdate;
  try
    State := nil;
    if MNode is TState then begin
      Manifold := TState(MNode).Manifold;
      if Manifold.nStates > 1 then
        State := TState(MNode); { Need to select State }
    end else
      Manifold := MNode as TManifold;
    Manifold.CheckProperties;
    CurrentMNode := MNode;

    { Force correct even/oddness of J }
    JSpinBox.MinValue := 0; { Avoid getting range check }
    JSpinBox.Value := (JSpinBox.Value and not 1) or Manifold.OddSpin;
    JSpinBox.MinValue := Manifold.OddSpin;
    { Force symmetry in range }
    SymSpinBox.MaxValue := SymSpinBox.Value; { Avoid getting range check }
    if SymSpinBox.Value >= Manifold.nSym then
      SymSpinBox.Value := Manifold.nSym - 1;
    SymSpinBox.MaxValue := Manifold.nSym - 1;

    Sym := Manifold.Molecule.JAdjustedSym(JSpinBox.Value, SymSpinBox.Value);
    { Do not attempt to use cache, to avoid problems with different settings }
    EigenValues := Manifold.MakeBasis(JSpinBox.Value, Sym, State);
    try
      with EigenValues do
        if Size > 0 then begin
          MakeHamiltonian(Manifold.Mixture.BasisOrder, PerturbationsCheckBox.Checked); { TODO: Make checkbox }
          HamGrid.RowCount := Size+1;
          HamGrid.ColCount := Size+1;
          for i := 1 to Size do begin
            HamGrid.Title[i] := Basis[i].BasisLabel;
            HamGrid.Lbl[i] := Basis[i].BasisLabel;
            for ii := 1 to Size do
              HamGrid.Value[i, ii] := C[i,ii];
            HamGrid.AutoAdjustColumn(i, HamGrid.Width < HamPanel.Width);
          end;
          HamiltonianLabel.Caption := 'Hamiltonian matrix for Sym = ' + Manifold.Molecule.SymName(Sym);
          // HamGrid.Title[0] := '<' + Basis[1].BasisLabelKey + '|';
          HamGrid.AutoAdjustColumn(0);
          HamLabel.Caption := '| ' + Basis[1].BasisLabelKey + ' >';
          Diagonalize(True);
          if Manifold.EigenSearch then { TODO: Make checkboxes }
            if Manifold.LimitSearch then
              DoLimitedEigenSearch
            else
              DoEigenSearch;
          Woff := 0;
          if Manifold.HasWidth then begin
            M := MatrixElements(meWidth);
            try
              Widths := Transform(M);
              Woff := 1;
            finally
              M.Free;
            end;
          end;
          EigGrid.RowCount := Size + 1;
          EigGrid.ColCount := Size + 3 + Woff;
          EigGrid.Title[2] := 'E';
          if Woff > 0 then
            EigGrid.Title[3] := 'Width';
          for i := 1 to Size do begin
            EigGrid.Lbl[i] := DominantBasis[i].StateLabel;
            EigGrid.Title[i+Woff+2] := Basis[i].BasisLabel;
            EigGrid.Value[i,2] := E[i];
            if Woff > 0 then
              EigGrid.Value[i, 3] := Widths[i];
            for ii := 1 to Size do
              EigGrid.Value[i, ii+2+Woff] := C[ii, i];
            EigGrid.AutoAdjustColumn(2+Woff+i, EigGrid.Width < EigPanel.Width);
          end;
          EigenvaluesLabel.Caption := 'Eigenvalues for Sym = '+Manifold.Molecule.SymName(Sym);
          EigGrid.Title[1] := Basis[1].StateLabelKey;
          EigGrid.AutoAdjustColumn(1);
          EigGrid.AutoAdjustColumn(0);
          EigLabel.Caption := '| ' + Basis[1].BasisLabelKey + ' >';
        end else begin
          { Empty basis - show blank matrices }
          HamGrid.Blank;
          EigGrid.Blank;
        end;
    finally
      EigenValues.Free;
    end;
  finally
    EndUpdate(False);
  end;
end;

procedure TStatesForm.StuffSelector(MNode: TMNode);
begin
  inherited;
  if (MNode = nil) or (csDestroying in MNode.ComponentState) then Exit;
  BeginUpdate;
  StateComboBox.Clear;
  StateComboBox.OnChange := {$IFDEF FPC}@{$ENDIF}ComboBoxChange;
  Stuff(MNode, '', nil);
  if StateComboBox.Items.Count > 0 then
    StateComboBox.ItemIndex := 0;
  EndUpdate(False);
  ComboBoxChange(StateComboBox);
  // if Assigned(CurrentMNode) then
  //  Caption := CurrentMNode.Name + ' - Details'
  // else
    Caption := 'State Details';
end;

function TStatesForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if ((MNode is TManifold) and (TManifold(MNode).nStates > 1)) or (MNode is TState) then begin
    StateComboBox.AddItem(Name1 + Name2, MNode);
    Result := StateComboBox.Items.Count-1;
  end;
end;

procedure TStatesForm.JSpinBoxChange(Sender: TObject);
begin
  if Lockout > 0 then Exit;
  ShowItem(CurrentMNode);
end;

procedure TStatesForm.EigGridGetCellLabel(Sender: TObject; Col, Row: Integer;
  var s: string; var IsText: Boolean);
begin
  if Row > 0 then
    if Col = 0 then
      s := IntToStr(Row)
    else if Col = 1 then begin
      s := EigGrid.Lbl[Row];
      IsText := True;
    end;
end;

initialization
{$IFDEF LCL_}
  {$I StateInfo.lrs}
{$ENDIF}
end.
