unit LevelFrm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ScrlCtrl, Graphs, StdCtrls, ExtCtrls,
  WhatsThis, ChildForm, MNode, Mixture, EditExtr, NewSpin, AlEdit, Buttons;

type

  TPlottedPoint = record
    X, Y:  Single;  { Plot Position }
    Index: Integer; { Observation number }
  end;

  TLevelForm = class(TChildForm)
    Panel: TPanel;
    Label1: TLabel;
    StateComboBox: TComboBox;
    PerturbationsCheckBox: TCheckBox;
    G: TGraphD;
    StatusBar: TStatusBar;
    TopEdit: TAlignedEdit;
    Label2: TLabel;
    Label3: TLabel;
    BottomEdit: TAlignedEdit;
    JMinSpin: TSpinBox;
    Label4: TLabel;
    Label5: TLabel;
    JMaxSpin: TSpinBox;
    Label6: TLabel;
    SubJJP1: TAlignedEdit;
    Label7: TLabel;
    AllButton: TSpeedButton;
    ZoomButton: TSpeedButton;
    ZoomOutButton: TSpeedButton;
    PointsButton: TSpeedButton;
    LineButton: TSpeedButton;
    LinePointsButton: TSpeedButton;
    JoinHighest: TCheckBox;
    NoObsButton: TSpeedButton;
    ObsUpperButton: TSpeedButton;
    ObsLowerButton: TSpeedButton;
    ReloadButton: TSpeedButton;
    JoinObsButton: TSpeedButton;
    AnotherButton: TSpeedButton;
    ListButton: TButton;
    TrackState: TCheckBox;
    TrackQno: TCheckBox;
    LabelsComboBox: TComboBox;
    Label8: TLabel;
    procedure ListButtonClick(Sender: TObject);
    procedure PerturbationsCheckBoxClick(Sender: TObject);
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure AllButtonClick(Sender: TObject);
    procedure ZoomButtonClick(Sender: TObject);
    procedure ZoomOutButtonClick(Sender: TObject);
    procedure ReloadButtonClick(Sender: TObject);
    procedure AnotherButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
  private
    JJP1: Double;
    { Keep track of vertical plot range }
    EMin, EMax: Double;
    FirstPoint: Boolean;
    FEBottom: Double;
    FETop: Double;
    LogHeaderShown: Boolean;
    PlottedPoints: array of TPlottedPoint;
    procedure PlotManifold(Manifold: TManifold; State: TState; Log, ShowMolecule, ShowManifold: Boolean);
    function ReadValues: Boolean;
    procedure SetEBottom(const Value: Double);
    procedure SetETop(const Value: Double);
  protected
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); overload; override;
    procedure ShowItem(MNode: TMNode; Log: Boolean); overload;
    procedure StuffSelector(MNode: TMNode); override;
    property ETop: Double read FETop write SetETop;
    property EBottom: Double read FEBottom write SetEBottom;
  end;

var
  LevelForm: TLevelForm;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     Obs, GMarks, GStyle, MathSup, LogUnit, Main, SlaveEdit, QuantumNumber;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

{ TLevelForm }

{ TODO: Save settings from form in TMixture }

function TLevelForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TLevelForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.ShowItem(MNode: TMNode; Log: Boolean);
  procedure PlotMolecule(Molecule: TMolecule; ShowMolecule:Boolean; ShowPartitionFunction: Boolean);
  var
    i: Integer;
  begin
    if ShowPartitionFunction then begin
      LogHeaderShown := Molecule.CalculatePartitionFunction(Molecule.Mixture.Temperature, Molecule.Mixture.Tvib, True,
                                                            ShowMolecule, not LogHeaderShown) > 0
    end else
      for i := 0 to Molecule.nManifolds-1 do
        PlotManifold(Molecule.Manifold[i], nil, Log, ShowMolecule, Molecule.nManifolds > 1);
  end;
  procedure PlotSpecies(Species: TSpecies; ShowMolecule:Boolean; ShowPartitionFunction: Boolean);
  var
    i: Integer;
  begin
    for i := 0 to Species.Count-1 do
      PlotMolecule(Species.Isotopomer[i], ShowMolecule or (Species.Count > 1), ShowPartitionFunction);
  end;
  procedure PlotMixture(Mixture: TMixture; ShowPartitionFunction: Boolean);
  var
    i: Integer;
  begin
    for i := 0 to Mixture.Count-1 do
      PlotSpecies(Mixture.Species[i], Mixture.nMolecules > 1, ShowPartitionFunction);
  end;
  procedure PlotState(State: TState);
  begin
    if State.Manifold.nStates = 1 then
      PlotManifold(State.Manifold, nil, Log, False, False)
    else
      PlotManifold(State.Manifold, State, Log, False, False);
  end;
  procedure PlotKey;
  var
    Molecule: TMolecule;
    Node: TMNode;
    sym: Integer;
    s, s1, s2: string;
  begin
    Node := MNode;
    while not (Node is TMolecule) and Assigned(Node.Owner) do
      Node := Node.Owner as TMNode;
    if not (Node is TMolecule) then
      Exit;
    { Only show a key if it would be identical for all molecules }
    { TODO : enhance above search for molecule to cope with identical symmetries? }
    {$IFDEF OLD}
    MarkStyle := TMarkStyle.Create;
    MarkStyle.DotAtCentre := True;
    MarkStyle.Colour := clBlack;
    Molecule := TMolecule(Node);
    x := JMinSpin.Value/2 + 3*G.cx;
    y := ETop-G.cy;
    for sym := 0 to Molecule.LastSym do begin
      MarkStyle.MarkType := TMarkType(Ord(FirstMark) + Sym);
      G.SymbolT(x-1.5*G.cx, y+G.cy/3, MarkStyle);
      G.Text(x, y, Molecule.SymName(sym));
      x := x + (4+Length(Molecule.SymName(sym)))*G.cx;
    end;
    MarkStyle.Free;
    {$ENDIF}
    s := 'Symmetry';
    Molecule := TMolecule(Node);
    if Molecule.Mixture.JAdjustSym then
      s := s + ' (Jeven/Jodd)';
    s := s + ': ';
    for sym := 0 to Molecule.LastSym do begin
      s := s + #1 + Char(Ord(FirstMark)+sym) + #0':';
      s1 := Molecule.JAdjustedSymName(0,sym);
      if Molecule.Mixture.JAdjustSym then begin
        s2 := Molecule.JAdjustedSymName(2,Molecule.JAdjustedSym(2,Sym));
        if s2 <> s1 then
          s1 := s1 + '/' + s2;
        s := s + s1 + ' (';
        s1 := Molecule.SymName(sym);
        s2 := Molecule.SymName(Molecule.JAdjustedSym(2,Sym));
        if s1 <> s2 then
          s := s + s1 + '/' + s2
        else
          s := s + s1;
        s := s + ')'
      end else
        s := s + s1;
      s := s + ' ';
    end;
    G.Title2(s);
  end;
  procedure Plot(ShowPartitionFunction: Boolean);
  begin
    LogHeaderShown := False;
    if MNode is TMixture then
      PlotMixture(TMixture(MNode), ShowPartitionFunction)
    else if MNode is TSpecies then
      PlotSpecies(TSpecies(MNode), False, ShowPartitionFunction)
    else if MNode is TMolecule then
      PlotMolecule(TMolecule(MNode), False, ShowPartitionFunction)
    else if MNode is TManifold then
      if ShowPartitionFunction then
        PlotMolecule(TManifold(MNode).Molecule, False, ShowPartitionFunction)
      else
        PlotManifold(TManifold(MNode), nil, Log, False, False)
    else if MNode is TState then
      if ShowPartitionFunction then
        PlotMolecule(TState(MNode).Molecule, False, ShowPartitionFunction)
      else
        PlotState(TState(MNode));
  end;
begin
  if not Assigned(MNode) then Exit;
  MNode.CheckProperties;
  if not ReadValues then
    Exit;
  G.Start;
  BeginUpdate;
  try
    { Plot some Levels }
    G.Map(JMinSpin.Value/2, JMaxSpin.Value/2, EBottom, ETop);
    PlotKey;
    FirstPoint := True;
    if JJP1 <> 0 then
      G.Scales('J', '(Energy - ' + FloatToStr(JJP1) + 'J(J+1))/' + (BaseNode as TMixture).UnitsName)
    else
      G.Scales('J', 'Energy/' + (BaseNode as TMixture).UnitsName);
    Plot(False);
    if (Log) then
      Plot(True);
    CurrentMNode := MNode;
  finally
    G.Finish;
    EndUpdate(False);
  end;
end;

procedure TLevelForm.ShowItem(MNode: TMNode);
begin
  ShowItem(MNode, False);
end;

function TLevelForm.StuffObject(const Name1, Name2: string; MNode: TMNode;
  var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if    ( (MNode is TMixture) and (TMixture(MNode).Count >1) )
     or ( (MNode is TSpecies) and (TSpecies(MNode).Count >1) )
     or ( (MNode is TMolecule) and (TMolecule(MNode).nManifolds >1) )
     or ( (MNode is TManifold) and (TManifold(MNode).nStates >1) )
     or ( MNode is TState ) then begin
    StateComboBox.AddItem(Name1 + Name2, MNode);
    Result := StateComboBox.Items.Count-1;
  end;
end;

procedure TLevelForm.StuffSelector(MNode: TMNode);
begin
  inherited;
  if (MNode = nil) or (csDestroying in MNode.ComponentState) then Exit;
  BeginUpdate;
  StateComboBox.Clear;
  StateComboBox.OnChange := {$IFDEF FPC}@{$ENDIF}ComboBoxChange;
  Stuff(MNode, '', nil);
  if StateComboBox.Items.Count > 0 then
    StateComboBox.ItemIndex := 0;
  EndUpdate(False);
  ComboBoxChange(StateComboBox);
  if Assigned(CurrentMNode) then
    Caption := CurrentMNode.Name + ' - Energy Level Plot'
  else
    Caption := 'Energy Level Plot';
end;

procedure TLevelForm.PerturbationsCheckBoxClick(Sender: TObject);
begin
  if Lockout > 0 then Exit;
  ShowItem(CurrentMNode);
end;

procedure TLevelForm.PlotManifold(Manifold: TManifold; State: TState; Log, ShowMolecule, ShowManifold: Boolean);
var
  J, JMin, JMax: TQuantumNumber;
  Sym, AdjSym: Integer;
  Ev: TEigenValues;
  MarkStyle: TMarkStyle;
  ShowPoints, ShowLines: Boolean;
  nPoints, i, nErrors: Integer;
  Points: array of TDPoint;
  Matches: array of Integer;
  LastColour: TColor;
  TIndex, MaxTIndex: Integer;
  function AdjustY(y: Double; J: TQuantumNumber): Double;
  begin
    Result := y - JJP1*J*(J+2)/4;
  end;
  procedure PlotPoint(EvNo: Integer);
  var
    E: Double;
  begin
    E := AdjustY(Ev.E[EvNo], J);
    if FirstPoint then begin
      Emin := E;
      EMax := E;
      FirstPoint := False;
    end else begin
      if E > EMax then EMax := E;
      if E < EMin then EMin := E;
    end;
    LastColour := Ev.DominantBasis[EvNo].State.FindColour;
    if LastColour = clNone then
      LastColour := clBlack;
    MarkStyle.Colour := LastColour;
    MarkStyle.MarkType := TMarkType(Ord(FirstMark) + Sym);
    if ShowPoints then
      G.SymbolT(J/2, E, MarkStyle);
    if ShowLines then begin
      if nPoints >= Length(Points) then
        SetLength(Points, nPoints+10);
      with Points[nPoints] do begin
        x := J/2;
        y := E;
      end;
      Inc(nPoints);
    end;
  end;
  procedure PlotObsPoint(Manifoldo: TManifold; Jo: TQuantumNumber; Symo: Integer; Indexo: Integer;
                         Manifoldc: TManifold; Jc: TQuantumNumber; Symc: Integer; Indexc: Integer;
                         E: Double; Number: Integer);
  begin
    if (Manifoldo = Manifold) and (Jo >= JMin) and (Jo <= JMax) then
      with Observations[Number] do begin
        Indexc := FindEigenvaluesAndIndex(Ev, Manifoldc, Jc, Symc, Indexc, Name,
                                          Manifoldc = ManifoldL, LineNumber, FileName);
        if Indexc <= 0 then begin
          Inc(nErrors);
          Exit;
        end;
        E := Adjusty(E + Ev.E[Indexc], Jo);
        LastColour := Ev.DominantBasis[Indexc].State.FindColour;
        if LastColour = clNone then
          LastColour := clBlack
        else
          Colour := LastColour;
        MarkStyle.Colour := LastColour;
        MarkStyle.MarkType := TMarkType(Ord(FirstMark) + Manifoldo.Molecule.JDeAdjustedSym(Jo,Symo));
        with PlottedPoints[Number] do begin
          X := Jo/2;
          Y := E;
          Index := Number;
          G.SymbolT(X, E, MarkStyle);
        end;
        if JoinObsButton.Down then begin
          Indexo := FindEigenvaluesAndIndex(Ev, Manifoldo, Jo, Symo, Indexo, Name,
                                            Manifoldo = ManifoldL, LineNumber, FileName);
          if Indexo <= 0 then begin
            Inc(nErrors);
            Exit;
          end;
          LastColour := Ev.DominantBasis[Indexo].State.FindColour;
          if LastColour = clNone then
            LastColour := clBlack
          else
            Colour := LastColour;
          G.StyleP(LastColour, G.LineStyle.Width, G.LineStyle.Pattern);
          G.ClipPoint(Jo/2, E);
          G.ClipJoin(Jo/2, Adjusty(Ev.E[Indexo],Jo));
        end;
      end;
  end;
  procedure ShowLog;
  var
    Mixture: TMixture;
    Sym, i: Integer;
  begin
    Mixture := Manifold.Mixture;
    if not LogHeaderShown then begin
      { Header for list }
      WriteLn('Energy Level List');
      if ShowMolecule then
        Write(Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
      if ShowManifold then
        Write('M':Length(Mixture.LongestManifoldName), ' ');
      WriteLn('   J Sym   #        Energy Label');
      LogHeaderShown := True;
    end;
    J := Manifold.FindJmin(JMin);
    while J <= JMax do begin
      for Sym := 0 to Manifold.nSym-1 do begin
        AdjSym := Manifold.Molecule.JAdjustedSym(J, Sym);
        Ev := Manifold.Eigenvalues(J, AdjSym, [], 0, State, PerturbationsCheckBox.Checked);
        for i := 1 to Ev.Size do begin
          if ShowMolecule then
            Write(Ev.Manifold.Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
          if ShowManifold then
            Write(Ev.Manifold.Name:Length(Mixture.LongestManifoldName), ' ');
          WriteLn(Manifold.QuantumNumberToStr(J):4, Sym:4, i:4, Ev.E[i]:14:Manifold.Precision, ' ', Ev.DominantBasis[i].StateLabel);
        end;
      end;
      Inc(J, 2);
    end;
    WriteLn;
  end;

  procedure ShowOneState(Sym: Integer; MatchState: TState);
  var
    EvNo, MaxEvNo, i: Integer;
    IndexL, IndexU, nMatch, IndexT: Integer; { Numbers filtered by State }
    Match: Boolean;
  begin
    EvNo := 1;
    MaxEvNo := 0;
    repeat
      nPoints := 0; { No points plotted yet }
      J := Manifold.FindJmin(JMin);
      while J <= JMax do begin
        AdjSym := Manifold.Molecule.JAdjustedSym(J, Sym);
        Ev := Manifold.Eigenvalues(J, AdjSym, [], 0, State, PerturbationsCheckBox.Checked);
        if Assigned(MatchState) or (TIndex >= 0) then begin
          { Find the number matching this state }
          nMatch := 0;
          for i := 1 to Ev.Size do begin
            Match := not Assigned(MatchState) or (Ev.DominantBasis[i].State = MatchState);
            if TIndex >= 0 then begin
              IndexT := Ev.DominantBasis[i].TrackIndex;
              if IndexT > MaxTIndex then
                MaxTIndex := IndexT;
              Match := Match and (TIndex = IndexT);
            end;
            if Match then begin
              Inc(nMatch);
              if nMatch > Length(Matches) then
                SetLength(Matches, nMatch+20);
              Matches[nMatch-1] := i;
            end;
          end;
          if EvNo <= nMatch then begin
            IndexL := Matches[EvNo-1];
            IndexU := Matches[nMatch-EvNo];
          end;
        end else begin
          nMatch := Ev.Size;
          IndexL := EvNo;
          IndexU := Ev.Size-EvNo+1;
        end;
        if nMatch > MaxEvNo then
          MaxEvNo := nMatch;
        if nMatch >= EvNo then
          if JoinHighest.Checked then
            PlotPoint(IndexU)
          else
            PlotPoint(IndexL);
        Inc(J, 2);
      end;
      if nPoints > 1 then begin
        G.StyleP(LastColour, G.LineStyle.Width, G.LineStyle.Pattern);
        G.ClipPoint(Points[0].X, Points[0].Y);
        for i := 1 to nPoints-1 do
          G.ClipJoin(Points[i].X, Points[i].Y);
      end;
      Inc(EvNo);
    until EvNo > MaxEvNo;
  end;

  procedure ShowStateLabels;
  var
    i: Integer;
  begin
    for i := 0 to Manifold.nStates-1 do
      with Manifold.State[i] do
        if (Origin >= G.Ymin) and (Origin <= G.Ymax) then
          case LabelsComboBox.ItemIndex of
            1: G.TextJ(G.Xmin-G.cx/6, Origin-G.cy/3, taRightJustify, Name); { LeftOut}
            2: G.TextJ(G.Xmin+G.cx*1.5, Origin, taLeftJustify,  Name); { LeftIn}
            3: G.TextJ(G.Xmax, Origin, taRightJustify, Name); { RightIn}
            4: G.TextJ(G.Xmax, Origin, taLeftJustify, Name);  { RightOut}
          end;
  end;

begin
  nErrors := 0;
  Manifold.CheckProperties;
  MarkStyle := TMarkStyle.Create;
  MarkStyle.DotAtCentre := True;
  JMax := JMaxSpin.Value;
  JMin := JMinSpin.Value;
  ShowPoints := not LineButton.Down;
  ShowLines := not PointsButton.Down;
  Setlength(Points, Jmax div 2 + 2);
  if Log then
    ShowLog;
  if TrackQno.Checked then begin
    TIndex := 0;
    MaxTIndex := 0;
  end else begin
    TIndex := -1;
    MaxTIndex := -1;
  end;
  repeat
    for Sym := 0 to Manifold.nSym-1 do
      if TrackState.Checked then
        for i := 0 to Manifold.nStates-1 do
          ShowOneState(Sym,Manifold.State[i])
      else
        ShowOneState(Sym,nil);
    Inc(TIndex);
  until TIndex > MaxTIndex;
  if LabelsComboBox.ItemIndex > 0 then
    ShowStateLabels;

  { Now plot observed energy levels }
  if ObsUpperButton.Down or ObsLowerButton.Down then
    SetLength(PlottedPoints, Length(Observations))
  else
    SetLength(PlottedPoints, 0);
  if ObsUpperButton.Down then
    for i := 0 to High(Observations) do
      with Observations[i] do
        PlotObsPoint(ManifoldU, JU, SymU, IndexU, ManifoldL, JL, SymL, IndexL, Frequency, i)
  else if ObsLowerButton.Down then
    for i := 0 to High(Observations) do
      with Observations[i] do
        PlotObsPoint(ManifoldL, JL, SymL, IndexL, ManifoldU, JU, SymU, IndexU, -Frequency, i);
  MarkStyle.Free;
  if nErrors > 0 then
    G.Title1(Format('%d problems with observations found', [nErrors]));
end;

procedure TLevelForm.GMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  s: string;
  GD: TGraphD;
  i: Integer;
begin
  GD := Sender as TGraphD;
  s := Concat(RoundedStr(GD.LastMouseX, (GD.Xmax-GD.Xmin)/GD.Width), '  ',
              RoundedStr(GD.LastMouseY, (GD.Ymax-GD.Ymin)/GD.Height));
  if GD.AreaSelected then
    with GD.SelectedMathArea do
      s := s + ', [' + RoundedStr(Right-Left, (GD.Xmax-GD.Xmin)/GD.Width) + ', '
                     + RoundedStr(Top-Bottom, (GD.Ymax-GD.Ymin)/GD.Height) + ']';
  if Length(PlottedPoints) = Length(Observations) then
    for i := 0 to High(PlottedPoints) do
      with PlottedPoints[i] do
        if GD.CloseToX(GD.LastMouseX, X) and GD.CloseToY(GD.LastMouseY, Y) then
          s := s + Format(' %s:%d', [ExtractFileName(Observations[Index].FileName), Observations[Index].LineNumber]);
  StatusBar.SimpleText := s;
end;

procedure TLevelForm.AllButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginUpdate;
  d := (EMax-Emin)/20;
  if d = 0 then d := 1;
  ETop := EMax+d;
  EBottom := EMin-d;
  EndUpdate(False);
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.ZoomButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginUpdate;
  if G.AreaSelected then begin
    ETop := G.SelectedMathArea.Top;
    EBottom := G.SelectedMathArea.Bottom;
    if ETop < EBottom then begin
      ETop := G.SelectedMathArea.Bottom;
      EBottom := G.SelectedMathArea.Top;
    end;
  end else if ReadValues then begin
    d := (ETop - EBottom)/4;
    Etop := Etop - d;
    EBottom := EBottom + d;
  end;
  EndUpdate(False);
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.SetEBottom(const Value: Double);
begin
  FEBottom := Value;
  BottomEdit.Text := FloatToStr(Value);
end;

procedure TLevelForm.SetETop(const Value: Double);
begin
  FETop := Value;
  TopEdit.Text := FloatToStr(Value);
end;

function TLevelForm.ReadValues: Boolean;
begin
  Result := TryStrToFloat(BottomEdit.Text, FEBottom)
            and TryStrToFloat(TopEdit.Text, FETop)
            and TryStrToFloat(SubJJP1.Text, JJP1);
end;

procedure TLevelForm.ZoomOutButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginUpdate;
  if ReadValues then begin
    d := (ETop - EBottom)/2;
    Etop := Etop + d;
    EBottom := EBottom - d;
  end;
  EndUpdate(False);
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.ReloadButtonClick(Sender: TObject);
begin
  if not SaveAllWindows then Exit;
  MainForm.LoadAnyObservations;
  PerturbationsCheckBoxClick(nil);
end;

procedure TLevelForm.AnotherButtonClick(Sender: TObject);
var
  NewLevelForm: TLevelForm;
begin
  NewLevelForm := TLevelForm.Create(Application);
  NewLevelForm.Show;
  NewLevelForm.BaseNode := BaseNode;
end;

procedure TLevelForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if Self <> LevelForm then
    CloseAction := caFree;
end;

procedure TLevelForm.ListButtonClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, True);
end;

initialization
{$IFDEF LCL_}
  {$I LevelFrm.lrs}
{$ENDIF}
end.
