unit TranInfo;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  { Windows, Messages, } SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ChildForm, StdCtrls, AlEdit, EditExtr, NewSpin, ExtCtrls, ComCtrls,
  MNode, Split, Grids, NGrid, Buttons;

type
  TTransitionForm = class(TChildForm)
    LabelPanel: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    UpperComboBox: TComboBox;
    UpperJSpinBox: TSpinBox;
    UpperSymSpinBox: TSpinBox;
    LowerComboBox: TComboBox;
    LowerSymSpinBox: TSpinBox;
    LowerJSpinBox: TSpinBox;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    TranComboBox: TComboBox;
    OrgPanel: TPanel;
    Label7: TLabel;
    OrgGrid: TNumberGrid;
    Split: TSplit;
    TranPanel: TPanel;
    Label8: TLabel;
    TranGrid: TNumberGrid;
    SquareCheckBox: TCheckBox;
    AddButton: TRadioButton;
    Label9: TLabel;
    ReplaceButton: TRadioButton;
    SkipButton: TRadioButton;
    PlotCheckBox: TCheckBox;
    Label10: TLabel;
    JChangeSpinBox: TSpinBox;
    PerturbationsCheckBox: TCheckBox;
    OriginalLabel: TLabel;
    TransformedLabel: TLabel;
    SimulateButton: TSpeedButton;
    ExcitationComboBox: TComboBox;
    Label11: TLabel;
    ExcitationRankSpinBox: TSpinBox;
    procedure UpperJSpinBoxChange(Sender: TObject);
    procedure LowerSymSpinBoxSetIntegerValue(Sender: TObject;
      NewValue: Integer; var NewText: String);
    procedure UpperSymSpinBoxGetIntegerValue(Sender: TObject;
      var NewValue: Integer; TheText: String; var StringValid: Boolean);
    procedure JChangeSpinBoxGetIntegerValue(Sender: TObject;
      var NewValue: Integer; TheText: String; var StringValid: Boolean);
    procedure JChangeSpinBoxSetIntegerValue(Sender: TObject;
      NewValue: Integer; var NewText: String);
    procedure SimulateButtonClick(Sender: TObject);
  protected
    procedure FormConstrainedResize(Sender: TObject; var MinWidth,
      MinHeight, MaxWidth, MaxHeight: Integer);
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure ShowTransitions;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure StuffSelector(MNode: TMNode); override;
  end;

var
  TransitionForm: TTransitionForm;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF} Math, Mixture, MatrixVec, Types,
  LineFrm, Main, QuantumNumber, Units, DRintensity;

{$IFNDEF FPC}
{$R *.dfm}
{$ENDIF}

{ TTransitionForm }

function TTransitionForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TTransitionForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  ComboBoxChange(TranComboBox);
end;

procedure TTransitionForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  ComboBoxChange(TranComboBox);
end;

procedure TTransitionForm.ShowItem(MNode: TMNode);
var
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  i: Integer;
  Change: TQuantumNumber;
  procedure StuffManifold(Manifold: TManifold; ComboBox: TComboBox);
  var
    i: Integer;
  begin
    if ComboBox.Items.IndexOfObject(Manifold.State[0]) >= 0 then Exit;
    if Manifold.nStates > 1 then
      ComboBox.AddItem(Manifold.UniqueName, Manifold);
    for i := 0 to Manifold.nStates-1 do
      ComboBox.AddItem(Manifold.State[i].UniqueName, Manifold.State[i]);
  end;
  procedure StuffTransitionMoments(TransitionMoments: TTransitionMoments);
  var
    Bra, Ket: TManifold;
  begin
    Bra := TransitionMoments.Bra as TManifold;
    Ket := TransitionMoments.Ket as TManifold;
    if Bra.Initial then begin
      StuffManifold(Bra, LowerComboBox);
      StuffManifold(Ket, UpperComboBox);
    end else begin
      StuffManifold(Ket, LowerComboBox);
      StuffManifold(Bra, UpperComboBox);
    end;
  end;
  procedure CheckForMultipleManifolds(ComboBox: TComboBox; JSpinBox, SymSpinBox: TSpinBox);
  var
    i: Integer;
    nManifolds: Integer;
    CurManifold, Manifold: TManifold;
  begin
    nManifolds := 0;
    CurManifold := nil;
    for i := 0 to ComboBox.Items.Count-1 do begin
      if ComboBox.Items.Objects[i] is TManifold then
        Manifold := TManifold(ComboBox.Items.Objects[i])
      else if ComboBox.Items.Objects[i] is TState then
        Manifold := TState(ComboBox.Items.Objects[i]).Manifold
      else
        raise Exception.Create('CheckForMultipleManifolds');
      if Manifold <> CurManifold then begin
        CurManifold := Manifold;
        Inc(nManifolds);
      end;
    end;
    if nManifolds > 1 then begin
      ComboBox.AddItem('(All)', nil);
      ComboBox.ItemIndex := ComboBox.Items.Count-1;
    end else
      ComboBox.ItemIndex := 0;
    if ComboBox.Items.Objects[0] is TManifold then
      Manifold := TManifold(ComboBox.Items.Objects[0])
    else if ComboBox.Items.Objects[0] is TState then
      Manifold := TState(ComboBox.Items.Objects[0]).Manifold
    else
      Exit;
    { Force correct even/oddness of J }
    JSpinBox.MinValue := Manifold.OddSpin - 2;
    JSpinBox.Value := (JSpinBox.Value and not 1) or Manifold.OddSpin;
    { Force symmetry in range }
    SymSpinBox.MaxValue := Manifold.nSym - 1;
    if SymSpinBox.Value >= Manifold.nSym then
      SymSpinBox.Value := Manifold.nSym - 1;
  end;
begin
  if not Assigned(MNode) then Exit;
  BeginUpdate;
  if MNode is TMolecule then begin
    Molecule := TMolecule(MNode);
    if Molecule.nTransitionMoments = 1 then
      TransitionMoments := Molecule.GetTransitionMoments(0)
    else
      TransitionMoments := nil;
  end else begin
    if MNode is TTransitionMoment then
      TransitionMoments := MNode.Owner as TTransitionMoments
    else
      TransitionMoments := MNode as TTransitionMoments;
    Molecule := TransitionMoments.Molecule;
  end;
  UpperComboBox.Clear;
  LowerComboBox.Clear;
  if Assigned(TransitionMoments) then
    StuffTransitionMoments(TransitionMoments)
  else
    for i := 0 to Molecule.nTransitionMoments-1 do
      StuffTransitionMoments(Molecule.GetTransitionMoments(i));
  CheckForMultipleManifolds(UpperComboBox, UpperJSpinBox, UpperSymSpinBox);
  CheckForMultipleManifolds(LowerComboBox, LowerJSpinBox, LowerSymSpinBox);
  { Update possible changes using MaxRank, preserving 'all' if required }
  Change := JChangeSpinBox.Value;
  if Change = JChangeSpinBox.MinValue then
    Change := -100;
  JChangeSpinBox.MinValue := -Molecule.MaxRank-2;
  JChangeSpinBox.MaxValue := Molecule.MaxRank;
  JChangeSpinBox.Value := Max(Min(Change,JChangeSpinBox.MaxValue), JChangeSpinBox.MinValue);
  CurrentMNode := MNode;
  ShowTransitions;
  EndUpdate(False);
end;

procedure TTransitionForm.ShowTransitions;
var
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  TransitionMoment: TTransitionMoment;
  ManifoldU, ManifoldL: TManifold;
  StateU, StateL: TState;
  EigenValuesU, EigenValuesL: TEigenvalues;
  SymU, SymL: Integer;
  LineList: TLineList;
  AddLines: Boolean;
  DeltaJ: TQuantumNumber;
  Excitation: TExcitation;
  function Extract(var Manifold: TManifold; var State: TState; var Sym: Integer; ComboBox: TComboBox;
                   JSpinBox, SymSpinBox: TSpinBox; OtherJ, DeltaJ: TQuantumNumber): TEigenvalues;
  var
    MNode: TMNode;
    J: TQuantumNumber;
  begin
    Result := nil;
    MNode := ComboBox.Items.Objects[ComboBox.ItemIndex] as TMNode;
    State := nil;
    Manifold := nil;
    if MNode is TManifold then
      Manifold := TManifold(MNode)
    else if MNode is TState then begin
      State := TState(MNode);
      Manifold := State.Manifold;
    end;
    Sym := SymSpinBox.Value;
    if Sym < 0 then Exit;
    J := JSpinBox.Value;
    if (J < 0) and (OtherJ >= 0) and (Abs(DeltaJ) < 100) then
      J := OtherJ + DeltaJ;
    if J < 0 then Exit;
    if Assigned(Manifold) then begin
      Result := Manifold.MakeBasis(J, Manifold.Molecule.JAdjustedSym(J,Sym), State);
      try
        if Result.Size = 0 then Exit;
        Result.MakeHamiltonian(Manifold.Mixture.BasisOrder); { TODO: Make checkbox }
        Result.Diagonalize(True);
        if Manifold.EigenSearch then
          if Manifold.LimitSearch then
            Result.DoLimitedEigenSearch
          else
            Result.DoEigenSearch;
      except
        Result.Free;
        raise;
      end;
    end;
  end;
  procedure ShowMatrix(D: TMatrix; Grid: TNumberGrid; ShowBasis: Boolean);
  var
    i, ii: Integer;
    function SLabel(i: Integer; Ev: TEigenValues): string;
    begin
      if ShowBasis then
        Result := Ev.Basis[i].BasisLabel
      else
        Result := Ev.DominantBasis[i].StateLabel;
    end;
  begin
    Grid.RowCount := Max(EigenValuesL.Size,1)+1;
    Grid.ColCount := Max(EigenValuesU.Size,1)+1;
    for i := 1 to EigenValuesU.Size do begin
      Grid.Title[i] := SLabel(i, EigenValuesU);
      if EigenValuesL.Size = 0 then
        Grid.Valid[1, i] := False;
      Grid.AutoAdjustColumn(i,Grid.Width < OrgPanel.Width);
    end;
    if EigenValuesL.Size = 0 then
      Grid.Lbl[1] := ' ';
    for i := 1 to EigenValuesL.Size do begin
      Grid.Lbl[i] := SLabel(i, EigenValuesL);
      if EigenValuesU.Size = 0 then
        Grid.Valid[i, 1] := False;
      for ii := 1 to EigenValuesU.Size do
        if Assigned(D) then
          Grid.Value[i, ii] := D[i,ii]
        else
          Grid.Valid[i, ii] := False;
    end;
    if EigenValuesU.Size = 0 then
      Grid.Title[1] := ' ';
    Grid.Caption := Format('<%s||%s>', [Molecule.SymName(EigenvaluesL.Sym), Molecule.SymName(EigenvaluesU.Sym)]);
    Grid.AutoAdjustColumn(0);
  end;
  procedure ShowMatrices;
  var
    D, F: TMatrix;
    s: string;
  begin
    F := nil;
    D := nil;
    if (EigenValuesL.Size > 0) and (EigenValuesU.Size >0)
       and (EigenValuesL.StatisticalWeight = EigenValuesU.StatisticalWeight) then
      if Assigned(TransitionMoment)
         or (Assigned(TransitionMoments) and (TransitionMoments.nRanks = 1)) then begin
        D := TransitionMoments.OriginalTransitionMatrix(EigenValuesL, EigenValuesU, TransitionMoments.MaxRank, TransitionMoment);
        F := D;
      end else if Assigned(TransitionMoments) then
        D := TransitionMoments.TransitionMatrix(EigenValuesL, EigenValuesU, nil, 0);
    try
      ShowMatrix(F, OrgGrid, True);
      s := '< Lower ';
      if EigenValuesL.Size > 0 then
        s := s + EigenValuesL.Basis[1].BasisLabelKey;
      s := s + ' |T| Upper ';
      if EigenValuesU.Size > 0 then
        s := s + EigenValuesU.Basis[1].BasisLabelKey;
      s := s + ' >';
      OriginalLabel.Caption := s;
      if Assigned(F) and (D.Size > 0) then begin
        TransitionMoments.TransformTransitionMatrix(D, EigenValuesL, EigenValuesU);
        SquareCheckBox.Enabled := True;
        if SquareCheckBox.Checked then
          D.Sqr;
        s := '< Lower ';
        if EigenValuesL.Size > 0 then
          s := s + EigenValuesL.Basis[1].StateLabelKey;
        s := s + ' |T| Upper ';
        if EigenValuesU.Size > 0 then
          s := s + EigenValuesU.Basis[1].StateLabelKey;
        s := s + ' >';
        TransformedLabel.Caption := s;
      end else begin
        SquareCheckBox.Enabled := False;
        SquareCheckBox.Checked := True;
      end;
      ShowMatrix(D, TranGrid, False);
    finally
      D.Free;
    end;
  end;
  procedure CalculateMatrices; { If transition selection sufficiently limited,
                                 show transition matrixes }
  var
    DeltaJ: TQuantumNumber;
  begin
    DeltaJ := JChangeSpinBox.Value;
    if DeltaJ = JChangeSpinBox.MinValue then
      DeltaJ := -100;
    EigenvaluesU := Extract(ManifoldU, StateU, SymU, UpperComboBox, UpperJSpinBox, UpperSymSpinBox, LowerJSpinBox.Value, DeltaJ);
    try
      EigenvaluesL := Extract(ManifoldL, StateL, SymL, LowerComboBox, LowerJSpinBox, LowerSymSpinBox, UpperJSpinBox.Value, -DeltaJ);
      try
        if Assigned(EigenvaluesU) and Assigned(EigenvaluesL) then
          ShowMatrices
        else begin
          OrgGrid.Blank;
          TranGrid.Blank;
        end;
      finally
        EigenValuesL.Free;
      end;
    finally
      EigenValuesU.Free;
    end;
  end;
begin
  if CurrentMNode = nil then Exit;
  BeginUpdate;
  TransitionMoment := nil;
  if CurrentMNode is TMolecule then begin
    Molecule := TMolecule(CurrentMNode);
    if Molecule.nTransitionMoments = 1 then
      TransitionMoments := Molecule.GetTransitionMoments(0)
    else
      TransitionMoments := nil;
  end else begin
    if CurrentMNode is TTransitionMoment then begin
      TransitionMoment := TTransitionMoment(CurrentMNode);
      TransitionMoments := CurrentMNode.Owner as TTransitionMoments;
    end else
      TransitionMoments := CurrentMNode as TTransitionMoments;
    Molecule := TransitionMoments.Molecule;
  end;
  Molecule.CheckProperties;
  try
    CalculateMatrices;
    AddLines := AddButton.Checked or ReplaceButton.Checked;
    if AddLines or PlotCheckBox.Checked then begin
      if ReplaceButton.Checked then
        LineForm.DeleteUnassignedClick(Self);
      LineList := TLineList.Create(Molecule.Mixture.Units);
      try
        DeltaJ := JChangeSpinBox.Value;
        if DeltaJ = JChangeSpinBox.MinValue then
          DeltaJ := -10000;
        Excitation := RandomExcitation;
        Excitation.Rank := ExcitationRankSpinBox.Value;
        Excitation.Simple := ExcitationComboBox.ItemIndex <= 0;
        if not Excitation.Simple then begin
          { TODO : Make more general }
          Excitation.DeltaJ := ExcitationComboBox.ItemIndex div 2;
          if Odd(ExcitationComboBox.ItemIndex) then
            Excitation.DeltaJ := -Excitation.DeltaJ;
        end;
        Molecule.AddToSpectrum(LineList, 0, Excitation,
                               LowerJSpinBox.Value, SymL,
                               UpperJSpinBox.Value, SymU,
                               ManifoldL, ManifoldU,
                               DeltaJ, TransitionMoments, TransitionMoment,
                               StateL, StateU, PerturbationsCheckBox.Checked);
        if AddLines and (LowerJSpinBox.Value < 0) and (UpperJSpinBox.Value < 0) and (LineList.nLines > 20) then
          AddLines := MessageDlg(Format('Add %d lines to line list?', [LineList.nLines]), mtConfirmation, [mbYes, mbNo], 0) = mrYes;
        if AddLines then begin
          LineList.FindLines(0, 1e99, LineList.Units);
          LineForm.ShowLines(MainForm.Mixture.Temperature, MainForm.Mixture.Tvib);
          SetFocus;
        end;
        LineList.Selected := True;
      except
        LineList.Free;
        raise;
      end;
      if PlotCheckBox.Checked then begin
        MainForm.LineList := LineList;
        MainForm.ShowSimulation;
      end else
        FreeAndNil(LineList);
    end;
  finally
    EndUpdate(False);
  end;
end;

function TTransitionForm.StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if (MNode is TTransitionMoment)
     or ((MNode is TTransitionMoments) and (TTransitionMoments(MNode).Count > 1))
     or ((MNode is TMolecule) and (TMolecule(MNode).nTransitionMoments >1)) then begin
    TranComboBox.AddItem(Name1 + Name2, MNode);
    Result := TranComboBox.Items.Count-1;
  end;
end;

procedure TTransitionForm.StuffSelector(MNode: TMNode);
begin
  inherited;
  if (MNode = nil) or (csDestroying in MNode.ComponentState) then Exit;
  {$IFNDEF LCL_}
  OnConstrainedResize := FormConstrainedResize;
  {$ENDIF}
  BeginUpdate;
  UpperComboBox.Clear;
  LowerComboBox.Clear;
  TranComboBox.Clear;
  TranComboBox.OnChange := {$IFDEF FPC}@{$ENDIF}ComboBoxChange;
  Stuff(MNode, '', nil);
  if TranComboBox.Items.Count > 0 then
    TranComboBox.ItemIndex := 0;
  EndUpdate(False);
  ComboBoxChange(TranComboBox);
  if Assigned(CurrentMNode) then
    Caption := CurrentMNode.UniqueName + ' - Details'
  else
    Caption := 'Transition Details';
end;

procedure TTransitionForm.UpperJSpinBoxChange(Sender: TObject);
begin
  if Lockout > 0 then Exit;
  ShowTransitions;
end;

procedure TTransitionForm.FormConstrainedResize(Sender: TObject;
  var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
begin
  MinHeight := ClientToScreen(Point(0,0)).Y - Top + LabelPanel.Height;
end;

procedure TTransitionForm.LowerSymSpinBoxSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: String);
begin
  if NewValue < 0 then
    NewText := 'all';
end;

procedure TTransitionForm.UpperSymSpinBoxGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: String; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') then begin
    if (Sender as TSpinBox).MinValue <= -2 then
      NewValue := (NewValue and 1) - 2
    else
      NewValue := -1;
    StringValid := True;
  end;
end;

procedure TTransitionForm.JChangeSpinBoxGetIntegerValue(
  Sender: TObject; var NewValue: Integer; TheText: String;
  var StringValid: Boolean);
var
  Change: Integer;
begin
  if SameText(TheText, 'all') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end else if (Length(TheText) = 1) then begin
    Change := (Ord(LowerCase(TheText)[1])-Ord('q'))*2;
    if Abs(Change) <= 18 then begin
      NewValue := Change;
      StringValid := True;
    end;
  end;
end;

procedure TTransitionForm.JChangeSpinBoxSetIntegerValue(
  Sender: TObject; NewValue: Integer; var NewText: String);
begin
  if NewValue = (Sender as TSpinBox).MinValue then
    NewText := 'all'
  else
    NewText := Char((NewValue div 2) + Ord('Q'));
end;

procedure TTransitionForm.SimulateButtonClick(Sender: TObject);
begin
  ShowTransitions;
end;

initialization
{$IFDEF LCL_}
  {$I TranInfo.lrs}
{$ENDIF}
end.
