unit Clock;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

function Clocks: int64; {$IFDEF CPU86} assembler; {$ENDIF}
function ClockFrequency: Double;

implementation

uses SysUtils;
               
var
  InitialClock: int64;
  InitialTime: TDateTime;

{$IFDEF OLD_MSWINDOWS}
uses mmSystem;
{$IFDEF PENTIUM}
  type TClock = Comp;

  procedure Clocks(var t: TClock); assembler;
  asm
    {$IFDEF WIN32}
      mov ECX, EDI
      mov EDI, t
      DB 0FH, 031H  { RDTSC }
      mov [EDI], eax
      mov [EDI+4], edx
      mov EDI, ECX
    {$ELSE}
      mov CX, DI
      les DI, t
      DB 0FH, 031H  { RDTSC }
      db 066h
      mov [DI], ax
      db 066h
      mov [DI+4], dx
      mov DI, CX
    {$ENDIF}
  end;
  {$IFDEF WIN32}
    function ClockFrequency: Double;
    var
      s, e: Comp;
      st, et, f: TLargeInteger;
      dt: Double;
    begin
      Clocks(s);
      QueryPerformanceCounter(st);
      QueryPerformanceFrequency(f);
      repeat
        Sleep(100);
        Clocks(e);
        QueryPerformanceCounter(et);
        { dt := (et.QuadPart - st.QuadPart)/f.QuadPart; }
        dt := (et - st)/f;
      until dt > 0.1;
      Result := (e-s)/dt;
    end;
  {$ELSE}
    function ClockFrequency: Double;
    var
      s, e: Comp;
      st, et: Longint;
      dt: Double;
    begin
      Clocks(s);
      st := timeGetTime;
      repeat
        Yield;
        Clocks(e);
        et := timeGetTime;
        dt := (et - st)/1000;
      until dt > 0.1;
      Result := (e-s)/dt;
    end;
  {$ENDIF}
{$ELSE}
  {$IFDEF WIN32}
    type TClock = Comp;
    procedure Clocks(var t: TClock);
    begin
      QueryPerformanceCounter(TLargeInteger(t));
    end;
    function ClockFrequency: Double;
    var
      f: Comp;
    begin
      QueryPerformanceFrequency(TLargeInteger(f));
      Result := f;
    end;
  {$ELSE}
    {$IFDEF timeGetTime}
      type TClock = Longint;
      procedure Clocks(var t: TClock);
      begin
        t := timeGetTime;
      end;
      function ClockFrequency: Double;
      begin
        Result := 1000;
      end;
    {$ELSE}
      var
        lpfnVTD: Longint;
      type TClock = Comp;
      procedure Get_VTD_Address; assembler;
      asm
        { Get address of VTD_Get_Real_Time function }
        push di
        push si
        mov ax, 01684h
        mov bx, 5
        xor di, di
        int 02fh
        mov WORD PTR lpfnVTD, di
        mov WORD PTR lpfnVTD+2, es
        pop si
        pop di
      end;
      procedure Clocks(var t: TClock); assembler;
      asm
        push di
        push si
        { Use VTD_Get_Real_Time function }
        mov ax, 0100h
        call lpfnVTD
        les DI, t
        db 066h
        mov [DI], ax
        db 066h
        mov [DI+4], dx
        pop si
        pop di
      end;
      function ClockFrequency: Double;
      begin
        Get_VTD_Address;
        Result := 1.19318e6;;
      end;
    {$ENDIF}
  {$ENDIF}
{$ENDIF}
{$ENDIF}

{$IFDEF CPU86}
{$IFDEF FPC}
  {$ASMMODE intel}
{$ENDIF}
function Clocks: int64; assembler;
asm
  DB 0FH, 031H  { RDTSC }
end;
{$ENDIF}
{$IFDEF DARWIN}
function mach_absolute_time: Int64; cdecl; external;
function Clocks: int64;
begin
  Result := mach_absolute_time;
end;
{$ENDIF}

function ClockFrequency: Double;
var
  CurrentClock: int64;
begin
  CurrentClock := Clocks;
  if Now = InitialTime then
    Sleep(1);
  Result := (CurrentClock-InitialClock)/(Now-InitialTime)/SecsPerDay;
end;

initialization
  InitialTime := Now;
  InitialClock := Clocks;
end.
