unit AlEdit;
{ Add Alignment to a TEdit Control }
{.$DEFINE SETEDITRECT} { Required so that multiline behaviour never shows }
{ For CLX_, Alignment is available anyway }
{ TODO: Implement alignment for LCL_ (may not be possible with GTK) }
{ TODO: Rename as TAlignedEdit }

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     {$IFNDEF CLX_} Graphics, Controls, stdctrls,
     {$ELSE} QGraphics, QControls, QStdCtrls, {$ENDIF}
     SysUtils, Classes;

type

  { TAlignedEdit }

  TAlignedEdit = class(TEdit)
  private
    FnVisibleDigits: Integer;
    {$IFNDEF CLX_}
    FAlignment: TAlignment;
    procedure SetAlignment(Value: TAlignment);
    {$ENDIF}
    {$IFDEF VCL_}
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
      {$IFDEF VER80}
      procedure WMSize(var Message: TWMSize); message WM_SIZE;
      {$ENDIF}
    {$ENDIF}
    function GetnVisibleDigits: Integer;
  protected
    {$IFDEF VCL_}
    procedure CreateParams(var Params: TCreateParams); override;
    {$ENDIF}
    {$IFDEF CLX_}
    procedure FontChanged; override;
    {$ENDIF}
    {TODO : FontChanged in LCL}
    procedure KeyPress(var Key: Char); override;
    {$IFNDEF VER80}
    procedure Resize; override;
    {$ENDIF}
  public
    constructor Create(AOwner: TComponent); override;
    property nVisibleDigits: Integer read GetnVisibleDigits;
  published
    {$IFNDEF CLX_}
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    {$ENDIF}
  end;
  
implementation

uses {$IFDEF CLX_} QForms, {$ENDIF} {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF} UCompat;

{ TAlignedEdit }

{$IFNDEF CLX_}
procedure TAlignedEdit.SetAlignment(Value: TAlignment);
begin
  if Value <> FAlignment then begin
    FAlignment := Value;
    {$IFDEF MSWINDOWS}
    if HandleAllocated then
      RecreateWnd{$IFDEF FPC}(Self){$ENDIF};
    { Note that under GTK/LCL_ this looses any text present }
    {$ENDIF}
  end;
end;
{$ENDIF}

function TAlignedEdit.GetnVisibleDigits: Integer;
var
  {$IFNDEF CLX_}
  DC: HDC;
  SaveFont: HFont;
  ZeroSize: TSize;
  {$ELSE}
  SaveFont: TFont;
  OwnerForm: TWidgetControl;
  Canvas: TCanvas;
  Width10: Integer;
  {$ENDIF}
begin
  if FnVisibleDigits < 0 then begin
    {$IFNDEF CLX_}
    { Get Text Metrics of current font }
    DC := GetDC(0);
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextExtentPoint(DC, AddressOfFirstChar('0123456789'), 10, ZeroSize);
    FnVisibleDigits := MulDiv(ClientWidth, 10, ZeroSize.cx);
    SelectObject(DC, SaveFont);
    ReleaseDC(0, DC);
    {$ELSE}
    { Locate a form with a canvas we can use }
    OwnerForm := Self.Parent;
    while Assigned(OwnerForm) and not (OwnerForm is TCustomForm) do
      OwnerForm := OwnerForm.Parent;
    if OwnerForm <> nil then begin
      Canvas := TCustomForm(OwnerForm).Canvas;
      SaveFont := TFont.Create;
      SaveFont.Assign(Canvas.Font);
      Canvas.Font := Font;
      Width10 := Canvas.TextWidth('0123456789');
      Canvas.Font := SaveFont;
      SaveFont.Free;
    end else
      Width10 := Abs(Font.Height) * 3 div 2; { TODO : Check this Guess }
    FnVisibleDigits := MulDiv(ClientWidth, 10, Width10);
    {$ENDIF}
  end;
  Result := FnVisibleDigits;
end;

procedure TAlignedEdit.KeyPress(var Key: Char);
begin
  if Ord(Key) = VK_RETURN then
    Key := Chr(0) { Always discard return, as it is not meant to be a multiline control }
  else if Key <> Chr(0) then
    inherited KeyPress(Key);
end;

{$IFNDEF VER80}
procedure TAlignedEdit.Resize;
begin
  inherited Resize;
  FnVisibleDigits := -1;
end;
{$ELSE}
procedure TAlignedEdit.WMSize(var Message: TWMSize);
begin
  inherited;
  FnVisibleDigits := -1;
end;
{$ENDIF}

constructor TAlignedEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  AutoSize := False;
  FnVisibleDigits := -1;
  {$IFNDEF CLX_}
  FAlignment := taLeftJustify;
  {$ENDIF}
end;

{$IFDEF VCL_}
procedure TAlignedEdit.CreateParams(var Params: TCreateParams);
const
  Alignments:array[TAlignment] of {$IFDEF DFS_COMPILER_4_UP} DWORD
                                  {$ELSE} Longint {$ENDIF} =(ES_LEFT,ES_RIGHT,ES_CENTER);
begin
  inherited CreateParams(Params);
  { Windows 2k (and possibly others) shows a strange interaction here }
  { If ES_RIGHT set, ES_AUTOVSCROLL must also be set }
  Params.Style := (Params.Style and not ES_AUTOVSCROLL) or ES_MULTILINE or ES_AUTOHSCROLL or Alignments[FAlignment];
  if FAlignment = taRightJustify then
    Params.Style := Params.Style or ES_AUTOVSCROLL;
  // Params.Style := Params.Style and not ES_WANTRETURN; { But this is ineffective }
end;
{$ENDIF}

{$IFDEF VCL_}
procedure TAlignedEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  FnVisibleDigits := -1;
end;
{$ENDIF}
{$IFDEF CLX_}
procedure TAlignedEdit.FontChanged;
begin
  inherited;
  FnVisibleDigits := -1;
end;
{$ENDIF}

end.

