unit Gmarks;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Classes, Graphics,
     {$ELSE} Types, QGraphics, Qt, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     ExpCan;

type

  TPreparePointsEvent = procedure (var p: array of TPoint; np:Integer) of object;


  TMarkParameters = record
    v, h: Integer;
    BlobV, BlobH: Integer;
    IsRealDevice,
    ForceSymmetricalMarks, DotAtCentre:Boolean;
    rgbColour: TColor;
    PreparePoints: TPreparePointsEvent;
  end;

  TMarkType = (mtDiamond, mtPlus, mtBox, mtCross, mtTriangle, mtStar, mtITriangle, mtCircle,
               mtFDiamond, mtFBox, mtFTriangle, mtFITriangle, mtFCircle, mtVbar, mtHbar, mtDot,
               mtXMajor, mtXMinor, mtYMajor, mtYMinor);
const  { Range of "normal" marks }
  FirstMark: TMarkType =  mtDiamond;
  LastMark: TMarkType = mtFCircle;

function QuantOff(d:Integer; plus1:Boolean):Integer;
procedure ClosedFigure(Canvas: TExportableCanvas; x:Integer; y:Integer; const input: array of TPoint; np:Integer;
                       fill :Boolean; var mp: TMarkParameters);
procedure OpenFigure(Canvas: TExportableCanvas; x:Integer; y:Integer; const input: array of TPoint; np:Integer;
                     Circle: Boolean; var mp: TMarkParameters);
procedure Mark(Canvas: TExportableCanvas; x:Integer; y:Integer; MarkType:TMarkType; var mp:TMarkParameters);
procedure Dot(Canvas: TExportableCanvas; x, y:Integer; var mp: TMarkParameters);    (* Plot a dot *)

const
 nMarkTypes = Ord(High(TMarkType))+1;

const
  NullHandle = {$IFDEF MSWINDOWS} 0; {$ENDIF} {$IFDEF UNIX} nil; {$ENDIF}

implementation

uses SysUtils;
{$IFNDEF VER80}
{$T+} // What is this?
{$ENDIF}

function QuantOff(d:Integer; plus1:Boolean):Integer;
begin
  if plus1 and (d <> 0) then
    if d > 0 then
      Result := d + 1
    else
      Result := d - 1
  else
    Result := d;
end;

const MAXnp = 20;

procedure ClosedFigure(Canvas: TExportableCanvas; x:Integer; y:Integer; const input: array of TPoint; np:Integer;
                       fill :Boolean; var mp: TMarkParameters);
var
  out:array[0..MAXnp] of TPoint;
  i:Integer;
begin
  for i := 0 to np-1 do begin
    out[i].x := x+input[i].x*mp.h;
    out[i].y := y+input[i].y*mp.v;
  end;
  if Assigned(mp.PreparePoints) then
    mp.PreparePoints(out, np);
  if fill then begin
    {$IFDEF VER80}
    Polygon(Canvas.Handle, out[0], np);
    {$ELSE}
    Canvas.Polygon(out, np);
    {$ENDIF}
  end else begin
    out[np] := out[0];
    {$IFDEF VER80}
    Polyline(Canvas.Handle, out[0], np+1);
    {$ELSE}
    Canvas.Polyline(out, np+1);
    {$ENDIF}
  end;
end;

const
  BlobBound:array [0..1] of TPoint = ((X:-1; Y:+1), (X:+1; Y:-1));

procedure Dot(Canvas: TExportableCanvas; x, y:Integer; var mp: TMarkParameters);    (* Plot a dot *)
var
  BlobP: TMarkParameters;
  r: TRect;
  OldStyle: TPenStyle;
  OldWidth: Integer;
begin
  OldStyle := Canvas.Pen.Style;
  OldWidth := Canvas.Pen.Width;
  if mp.BlobV = 0 then begin
    r.Left := x;
    r.Top := y;
    if Assigned(mp.PreparePoints) then
      mp.PreparePoints(r.TopLeft, 1);
    {$IFDEF KYLIX1}
    Canvas.Pen.Style := psSolid;
    Canvas.DrawPoint(r.TopLeft);
    {$ELSE}
    Canvas.Pixels[r.Left, r.Top] := Canvas.Pen.Color;
    {$ENDIF}
  end else begin
    BlobP := mp;
    BlobP.v := mp.BlobV;
    BlobP.h := mp.BlobH;
    { QPainter_setPen(DC, QColor(mp.rgbColour)); }
    Canvas.Pen.Style := psSolid;
    { ? Needed ? Canvas.Pen.Mode := pmCopy; }
    Canvas.Pen.Width := 0;
    OpenFigure(Canvas, x, y, BlobBound, 2, True, BlobP);
  end;
  if OldStyle <> Canvas.Pen.Style then
    Canvas.Pen.Style := OldStyle;
  if OldWidth <> Canvas.Pen.Width then
    Canvas.Pen.Width := OldWidth;
end;

const
  box:array [0..3] of TPoint = ((X:-2; Y:-2), (X:+2; Y:-2), (X:+2; Y:+2), (X:-2; Y:+2));
  diamond:array [0..3] of TPoint = ((X: 0; Y:+2), (X:-2; Y: 0), (X: 0; Y:-2), (X:+2; Y: 0));
  triangle:array [0..2] of TPoint =((X: 0; Y:-2), (X:-2; Y: 1), (X: 2; Y: 1));
  itriangle:array [0..2] of TPoint =((X: 0; Y: 2), (X:-2; Y:-1), (X: 2; Y:-1));
  pluscross:array [0..7] of TPoint =((X:-2; Y: 0), (X:+2; Y: 0), (X: 0; Y:-2), (X: 0; Y:+2),
                                     (X:-2; Y:-2), (X:+2; Y:+2), (X:-2; Y:+2), (X:+2; Y:-2));
  xmajor: array [0..1] of TPoint = ((X:0; Y:0), (X:0; Y:2));
  xminor: array [0..1] of TPoint = ((X:0; Y:0), (X:0; Y:1));
  ymajor: array [0..1] of TPoint = ((X:0; Y:0), (X:2; Y:0));
  yminor: array [0..1] of TPoint = ((X:0; Y:0), (X:1; Y:0));

procedure Mark(Canvas: TExportableCanvas; x:Integer; y:Integer; MarkType:TMarkType; var mp: TMarkParameters);
var
  DoDot:Boolean;
  circle:array [0..1] of TPoint;
  OldBrushStyle: TBrushStyle;
  OldBrushColor: TColor;
  procedure MakeCircle;
  begin
    if mp.h > 0 then
      circle[0].x := -2
    else
      circle[0].x := 2;
    if mp.v > 0 then
      circle[0].y := -2
    else
      circle[0].y := 2;
    circle[1].x := -circle[0].x;
    circle[1].y := -circle[0].y;
  end;
begin
  DoDot := false;
  case MarkType of
      mtDot: (* do dot *)
          DoDot := true;
      mtHbar:    (* Horizontal bar (for error bars) *)
          OpenFigure(Canvas, x, y, pluscross, 2, False, mp); (* Changed from wgraph *)
      mtVbar:    (* Vertical bar (for error bars) *)
          OpenFigure(Canvas, x, y, pluscross[2], 2, False, mp); (* Changed from wgraph *)
      mtDiamond: (* do diamond *)
        begin
          ClosedFigure(Canvas, x, y, diamond, 4, false, mp);
          DoDot := mp.DotAtCentre;
        end;
      mtPlus: (* do plus *)
          OpenFigure(Canvas, x, y, pluscross, 4, False, mp); (* Changed from wgraph *)
      mtBox: (* do box *)
        begin
          ClosedFigure(Canvas, x, y, box, 4, false, mp);
          DoDot := mp.DotAtCentre;
        end;
      mtCross: (* do X *)
          OpenFigure(Canvas, x, y, pluscross[4], 4, False, mp); (* Changed from wgraph *)
      mtTriangle: (* do triangle *)
        begin
          ClosedFigure(Canvas, x, y, triangle, 3, false, mp);
          DoDot := mp.DotAtCentre;
        end;
      mtStar: (* do star *)
          OpenFigure(Canvas, x, y, pluscross, 8, False, mp); (* Changed from wgraph *)
      mtITriangle: (* do inverted triangle *)
        begin
          ClosedFigure(Canvas, x, y, itriangle, 3, false, mp);
          DoDot := mp.DotAtCentre;
        end;
      mtCircle: (* Open Circle *)
        begin
          OldBrushColor := Canvas.Brush.Color;
          OldBrushStyle := Canvas.Brush.Style;
          Canvas.Brush.Style := bsClear;
          MakeCircle;
          OpenFigure(Canvas, x, y, Circle, 2, True, mp);
          Canvas.Brush.Style := OldBrushStyle;
          Canvas.Brush.Color := OldBrushColor; { Changing brush Style can also change color }
          DoDot := mp.DotAtCentre;
        end;
      mtFDiamond: (* Filled diamond *)
          ClosedFigure(Canvas, x, y, diamond, 4, True, mp);
      mtFBox: (* Filled box *)
          ClosedFigure(Canvas, x, y, box, 4, True, mp);
      mtFTriangle: (* Filled triangle *)
          ClosedFigure(Canvas, x, y, triangle, 3, True, mp);
      mtFITriangle: (* Filled inverted triangle *)
          ClosedFigure(Canvas, x, y, itriangle, 3, True, mp);
      mtFCircle: (* Filled Circle *)
        begin
          MakeCircle;
          OpenFigure(Canvas, x, y, Circle, 2, True, mp);
        end;
      { Axis marks }
      mtXMinor: OpenFigure(Canvas, x, y, xminor, 2, False, mp);
      mtXMajor: OpenFigure(Canvas, x, y, xmajor, 2, False, mp);
      mtYMinor: OpenFigure(Canvas, x, y, yminor, 2, False, mp);
      mtYMajor: OpenFigure(Canvas, x, y, ymajor, 2, False, mp);
  end;
  if DoDot then
    Dot(Canvas, x, y, mp);
end;

{$R-}
procedure OpenFigure(Canvas: TExportableCanvas; x:Integer; y:Integer; const input: array of TPoint; np:Integer;
                     Circle: Boolean; var mp: TMarkParameters);
var
  out:array[0..MAXnp] of TPoint;
  i:Integer;
  p:TPoint;
begin
  if 2*np > MAXnp then raise Exception.CreateFmt('OpenFigure (%d)', [np]);
  if mp.ForceSymmetricalMarks and not Circle then begin
    p.x := x; p.y := y;
    for i := 0 to np-1 do begin
      out[2*i].x := x + input[i].x*mp.h;
      out[2*i].y := y + input[i].y*mp.v;
      out[2*i+1] := p;
    end;
    np := np * 2;
  end else
    for i := 0 to np-1 do begin
      { As last pixel not set, need to extend lines that are not continued }
      out[i].x := x + QuantOff(input[i].x*mp.h, (i and 1) <> 0);
      out[i].y := y + QuantOff(input[i].y*mp.v, (i and 1) <> 0);
    end;
  if Assigned(mp.PreparePoints) then
    mp.PreparePoints(out, np);
  if Circle then begin
    {$IFDEF MSWINDOWS}
    Canvas.Ellipse(out[0].x, out[0].y, out[1].x, out[1].y);
    {$ENDIF}
    {$IFDEF UNIX}
    if out[0].y > out[1].y then begin
      i := out[0].y;
      out[0].y := out[1].y;
      out[1].y := i;
    end;
    Canvas.Ellipse(out[0].x, out[0].y+1, out[1].x, out[1].y+1); { +1 corrects for Windows/Kylix difference }
    {$ENDIF}
  end else
    for i := 0 to np div 2 -1 do begin
      {$IFDEF VER80}
      Polyline(Canvas.Handle, out[2*i], 2);
      {$ELSE}
      { Copy required as can't do fortran style slice on dynamic arrays }
      out[0] := out[2*i];
      out[1] := out[2*i+1];
      Canvas.Polyline(out, 2);
      {$ENDIF}
    end;
end;

end.
