unit Graphs;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface
{ TODO: Default context menu for graphs }
uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs, StdCtrls,
                 ExtCtrls, Menus, ClipBrd, Printers,
  {$ELSE} QGraphics, QControls, QForms, QDialogs, QStdCtrls, QExtCtrls, QMenus, QClipbrd, QPrinters, Types, Qt, {$ENDIF}
  {$IFDEF LCL_} LCLType, {$ENDIF}
  SysUtils, Classes, ExpCan, Huge, Prevwdlg, PrintPre, Gstyle, Gmarks, ScrlCtrl;

{$IFNDEF NONWINLCL}
 {$DEFINE METAFILES}
{$ENDIF}

const
  Polysize = 200;
  MaxPens = 16;
  DefPixels = 25000;
  AxisPen = MaxPens;
  MaxPlot = 32767; { To avoid problems in Win95/8 in 32 bit programs, limit
                     graphics calls to 16 bit }

type
  TGraphB = class;

  TGraphOption = (grLandScape, grUpsideDown, grScaledFont, grPensHaveWidth, grColouredPens, grPensHaveStyle,
                  grIsotropic, grForceSymmetricalMarks, goThumbTracking, grNoAreaSelection);

  TGraphOptions = set of TGraphOption;

  TReplotEvent = procedure (Sender:TObject; G:TGraphB) of object;

  TGraphPrintPreview = class;

  TDCType = (dcMF, dcClip, dcExport, dcScreen, dcScreenBuffer, dcNone, dcPreview, dcPrinter);

  { TGraphB }

  TGraphB = class(TScrollableControl)
  private
    FLineStyles, FDefaultLineStyles: array [0..MaxPens] of TLineStyle;
    FMarkStyles, FDefaultMarkStyles: array [0..MaxPens-1] of TMarkStyle;
    CLineStyle: TLineStyle;
    MarkLineStyle: TLineStyle;
    FErrorBarSize: Single;
    FOnReplot:TReplotEvent;
    FOnInvalidate:TNotifyEvent;
    FGOptions: TGraphOptions;
    FXmm, FYmm: Double;           { Size of plot in mm on Printer }
    FXPixels, FYPixels: Integer;  { Current plot range in pixels; value depends on DCType, and invalid for dcNone }
    FXOffset, FYOffset: Integer;  { Location (in device co-ordinates) of top left of plot }
    FAvailableHeight: Integer;    { Height argument to ShowOnCanvas - used for font scaling }
    FPrintPreview:TGraphPrintPreview; { Private Print Preview Component, for simple printing }
    FPrOnInvalidate:TNotifyEvent; { For Use by print preview }
    FLastMouse: TPoint;
    FPlotFailed: Boolean;         { If last call of FOnReplot raised an exception }
    FAreaSelected:Boolean;
    CurrentPenSize: Double;
    CurrentMarkSize: Single;
    FTextSize: Integer;
    FSavedAngle, FSavedLinestyle, FSavedTextSize:Integer;
    FJustify, FSavedJustify: TAlignment;
    FStartPlottingCount: Integer; { Allow nesting of start_plotting / stop_plotting calls }
    FOnPaint: TNotifyEvent;
    FScreenBuffer: TBitmap;   { For double buffering; note that this buffer is only for
                              make up of plot between Start/Finish pairs, and is drawn and discarded
                              by Finish. (Could keep, and duplicate TGraphW logic) }
    {$IFDEF UNIX}
    FDoubleBuffered: Boolean;
    {$ENDIF}
    procedure SetXmm(v:Double);
    procedure SetYmm(v:Double);
    procedure SetSelectedArea(r:TRect);
    procedure SetAreaSelected(b: Boolean);
    function GetLogicalSelectedArea: TRect;
    procedure SetLogicalSelectedArea(ARect: TRect);
    procedure MakePrintPreview;
    function GetLineStyles(i:Integer):TLineStyle;
    procedure SetLineStyles(i:Integer; Style:TLineStyle);
    function IsNonDefaultStyle(i:Integer):Boolean;
    function GetMarkStyle(i:Integer):TMarkStyle;
    procedure SetMarkStyle(i:Integer; Style:TMarkStyle);
    function IsNonDefaultMarkStyle(i:Integer):Boolean;
    procedure plot_blob(x:Integer; y:Integer);
    procedure SetCurrentMarkSize(Size: Single);
    function GetTickSize:TPoint;
    procedure SetMetafileScaling;
    {$IFDEF METAFILES}
    function GetPicture:{$IFDEF MSWINDOWS} {$IFDEF WINLCL} HMETAFILE {$ELSE} TMetafile {$ENDIF} {$ENDIF} {$IFDEF UNIX} TDrawing {$ENDIF};
    {$ENDIF}
    procedure CLineStyleChanged(Sender:TObject);
  protected
    {$IFDEF METAFILES}
    { WIN32 uses old style metafiles, as these seem to be compatible with more applications }
    hMF: {$IFDEF MSWINDOWS} {$IFDEF VER80} TMetafile; {$ELSE} HMETAFILE; {$ENDIF}{$ENDIF}
         {$IFDEF UNIX} TDrawing; {$ENDIF}
    {$IFDEF MSWINDOWS}
    MetafileCanvas: TCanvas;
    MetaFileWidth, MetaFileHeight: Integer;
    {$ENDIF}
    {$ENDIF}
    FCurrentCanvas: TExportableCanvas;
    DCType: TDCType;
    InReplot:Boolean;             (* If OnReplot event call *)
    XSCALE:Double;                (* User selected plot scale *)
    YSCALE:Double;
    oldXSCALE, oldYSCALE: Double;      { Previous User selected plot scale }
    oldHPos, oldVPos: Double;          { Previous Scroll Bar Position (before last zoom) }
    Selecting:Boolean;
    FSelectedArea:TRect;
    DPIX, DPIY: Integer;     { pixels per inch - may be out by a scale factor for metafiles }
    YPerX:Double;            (* Multipy to convert an x displacement to a y displacement with the same physical distance *)

    {$IFNDEF CLX_}
    GraphicsLogFont: TLogFont;          { Current font description }
    UnrotatedGraphicsLogFont: TLogFont; { Current font description with rotation set to zero }
    GraphicsFont: TFont;
    UnrotatedGraphicsFont: TFont;
    {$IFDEF MSWINDOWS}
    text_flags:Word;         { Current text justification }
    {$ENDIF}
    {$ENDIF}
    CurFontHeight: Integer;  (* Height in screen or printer pixels *)
    CurFont10Width: Integer; (* 10 * average char width *)
    CurFontLeading, CurFontInternalLeading, CurFontAscent: Integer;

    FTextAngle:Integer;      (* rotation of chars/(10*degrees) *)
    RealTextAngle: Integer;  { FTextAngle modified for land_sea, upside down flags }
    s_TextAngle:Double;      (* sin and cos of TextAngle *)
    c_TextAngle:Double;
    FontSize:Double;         { Base character size in current pixels; may be negative! }

    (* Line Info *)
    mp: TMarkParameters;
    absX:Integer;        (* Current pen position *)
    absY:Integer;
    DRAW_FLAG:Boolean;   (* Started drawing line flag *)
    SETDOT:Boolean;      (* Dot next at current Point *)
    SYMBOL:Boolean;      (* If set, plot a character at each Point *)
    SymbolMark:Integer;  (* Mark to plot at each Point *)
    curseg: array [0..Polysize-1] of TPoint; (* Holds current line segment *)
    nextseg:Integer;
    FLinetype:Integer;               { Current pen number - Was current_linestyle }
    PenSize:Integer;                 (* Current Pen Size *)
    penstyle:Integer;                (* Current Pen Style *)
    pattern: array[0..MAXPATLENGTH-1] of Integer;(* Current Pattern *)
    pattern_length:Integer;          (* Number of segments *)
    penstep:Integer;                 (* where in current pattern? *)
    segleft:Double;                  (* length left in current part pattern *)

    procedure InitGraphics;
    procedure cancel_graphics_DC;
    procedure setup_graphics_DC;
    procedure InitFont;
    procedure start_plotting;        (* Call to ensure DC present *)
    procedure stop_plotting;         (* Call when done with DC *)
    procedure flush_polyline;
    procedure end_draw;
    procedure LoadMP;
    procedure MT(x:Integer; y:Integer);
    procedure LT(x:Integer; y:Integer);
    procedure PreparePoints(var p: array of TPoint; np:Integer);
    procedure set_pensize(p:Double);
    procedure set_linestyle(ALinestyle:Integer; linescale:Double);
    procedure NeedPen(AStyle: TLineStyle);
    procedure NeedLinePen;
    procedure WIN_change_font;
    procedure FontHasChanged(Sender:TObject);
{ NOLONGER:    procedure set_window_orientation(DCHandle:HDCPainter); }
    procedure SetScaling(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint);
    procedure Paint; override;
    procedure MakeScreenImage(ScreenCanvas: TCanvas);
    procedure SetOnReplot(e:TReplotEvent);
    procedure SetGOptions(opt:TGraphOptions);
    procedure HintDisplacement(Displacement:Integer; var v, h:Integer; AllowZero:Boolean);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure UpdateSelection( X, Y: Integer; NewPoint:Boolean);
    procedure DrawSelectedArea;
    procedure ChangeZoom(newXSCALE, newYSCALE, newHscroll, newVscroll: Double); virtual;
    procedure ChangeScrollPos(Hpos, Vpos: double); virtual;
    procedure Resize; override;
    procedure UpdateScrollBars; override;
    function HScrollFraction: Double; { Scroll Bar Positions as fraction of plot space }
    function VScrollFraction: Double;
    procedure ShowOnCanvas(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint); virtual;
    procedure SetLastMouse(X, Y: Integer); virtual;
    procedure DoOptionChanged(Sender: TObject);
    procedure Loaded; override;
    procedure SetCurrentCanvas(ACanvas: TExportableCanvas);
    procedure SetRealCanvas(ACanvas: TCanvas);
    procedure SetLinetype(ipen:Integer);
    procedure SetLineStyle(AStyle: TLineStyle);
    function GetLineStyle:TLineStyle;
  protected
    procedure DoCopyToClipboard(Sender: TObject);
    procedure DoChooseFont(Sender: TObject);
  public
    property Canvas; { Re-export }
    property LineStyles[i:Integer]:TLineStyle read GetLineStyles write SetLineStyles;
    property MarkStyles[i:Integer]:TMarkStyle read GetMarkStyle write SetMarkStyle;
    property ErrorBarSize: Single read FErrorBarSize write FErrorBarSize;
    property LastMouse: TPoint read FLastMouse;
    property CurrentCanvas:TExportableCanvas read FCurrentCanvas;
    property AreaSelected: Boolean read FAreaSelected write SetAreaSelected;
    property TickSize:TPoint read GetTickSize;
    { May want to make the following private or protected }
    function ClientY:Integer;
    function ClientX:Integer;
    procedure SetScreenScaling;
    procedure PenInvalid;
    procedure LogToDev(var x, y: Integer); { Convert from logical to device pixels;
                                             mainly sign conversion } {$IFNDEF VER80} overload; {$ENDIF}
    procedure DevToLog(var x, y: Integer); { Convert from device to logical pixels;
                                             mainly sign conversion } {$IFNDEF VER80} overload; {$ENDIF}

    {$IFNDEF VER80}
    procedure LogToDev(var Point: TPoint); overload;
    procedure DevToLog(var Point: TPoint); overload;
    {$ENDIF}

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Start;
    procedure Finish;
    procedure Suspend; {\ To allow display of partly completed metafile }
    procedure Resume;  {/ Ignored unless OnReplot not set }
    {$IFDEF METAFILES}
    procedure StartMetafile;
    procedure StopMetafile;
    {$ENDIF}
    procedure SetToDefaultStyle(Pen: Integer; AStyle: TLineStyle);
    procedure LineStyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
    procedure Vector(x:Integer; y:Integer);
    procedure Move(x:Integer; y:Integer);
    procedure Box(Style: TBrushStyle; Density: Double; const Rect: TRect);
    procedure ClearBox(const Rect: TRect);
    procedure DrawGraphic(Rect: TRect; Graphic: TGraphic);
    procedure SetTextSize(s:Integer); (* Scale up text size by s/100 *)
    procedure TextI(x:Integer; y:Integer; str:string);
    procedure XorTextI(x:Integer; y:Integer; str:string);
    function  TextExtentF(str:string; Fraction: Single {$IFDEF NONWINLCL}; var sz: TSize {$ENDIF}):TPoint;
    function  TextExtent(str:string):TPoint;
    procedure Mark(x:Integer; y:Integer; number:Integer);
    procedure MarkT(x:Integer; y:Integer; AStyle: TMarkStyle);
    procedure MarkE(x:Integer; y:Integer; AType: TMarkType);
    procedure Angle(ang:Integer); (* In 1/10ths of degrees *)
    procedure Justify(mode:TAlignment);
    procedure Clear;
    property  SelectedArea:TRect read FSelectedArea write SetSelectedArea;
    property  LogicalSelectedArea:TRect read GetLogicalSelectedArea write SetLogicalSelectedArea;
    procedure CancelSelection;
    procedure Invalidate; override;
    procedure ZoomSelected(Anisotropic:Boolean);
    procedure ResetZoom;
    procedure PreviousZoom; virtual;
    procedure ZoomOut;
    function land_sea:Boolean;
    procedure ShowPrintPreview(Sender:TObject; ACanvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                               APage: Integer; OnPrinter:Boolean);
    procedure Setmm(x, y: Double);
    procedure ExportTo(AStream: TStream; CanvasClass: TStreamCanvasClass); virtual;
    { Current pixel range; value depends on DCType, and invalid for dcNone }
    property XPixels:Integer read FXPixels;
    property YPixels:Integer read FYPixels;
{    function ScreenAvCharWidth:Integer; TODO: Use?
    function ScreenCellHeight:Integer; }
    function IsRealDevice:Boolean;
    function IsClipboard:Boolean;
    function IsPrinter:Boolean;
    function IsScreen:Boolean;
    function AvCharWidth:Double;
    function CellHeight:Double;
    function PenDialog(Pen:Integer):Boolean;
    function MarkDialog(AMark:Integer):Boolean;
    procedure PrinterSetup;
    procedure OptionChanged; virtual; { Options changed, but basic plot has not }
    procedure PlotChanged; virtual;   { Plot has changed; forces call of OnReplot }
    {$IFDEF METAFILES}
    property Picture: {$IFDEF MSWINDOWS} {$IFDEF WINLCL} HMETAFILE {$ELSE} TMetafile {$ENDIF} {$ENDIF} {$IFDEF UNIX} TDrawing {$ENDIF} read GetPicture;
    {$ENDIF}
    property PlotFailed:Boolean read FPlotFailed;
    property TextSize:Integer { Replot time - Text Size scale factor - % } read FTextSize write SetTextSize;
    property TextAngle:Integer read FTextAngle write Angle;
    property LineType:Integer read FLinetype write SetLineType; { Pen number: -1 for custom }
    property LineStyle:TLineStyle read GetLineStyle write SetLineStyle;

  published

    procedure PrintPreview;
    procedure Print;
    function CanPrint: Boolean; virtual;
    procedure CopyToClipboard; virtual;
    function CanCopyToClipboard: Boolean;

    property Align;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;
    property BorderStyle;
    property Color;
    {$IFDEF MSWINDOWS}
    property Ctl3D;
    property ParentCtl3D;
    property DragCursor;
    {$IFDEF VCL_}
    property Locked;
    {$ENDIF}
    {$ENDIF}
    property DragMode;
    property DoubleBuffered {$IFDEF UNIX} :Boolean read FDoubleBuffered write FDoubleBuffered {$ENDIF} ;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
{$ifdef WIN32}
    property OnStartDrag;
{$endif}

    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;

   (* Max plotting range - in mm*100 on Printer *)
    property Xmm:Double read FXmm write SetXmm;
    property Ymm:Double read FYmm write SetYmm;
    property OnReplot:TReplotEvent read FOnReplot write SetOnReplot;
    Property GOptions:TGraphOptions read FGOptions write SetGOptions
             default [grLandScape, grForceSymmetricalMarks];
    property OnInvalidate:TNotifyEvent read FOnInvalidate write FOnInvalidate;

   {$IFNDEF LCL_}
    property LineStyle0:TLineStyle index 0 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle1:TLineStyle index 1 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle2:TLineStyle index 2 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle3:TLineStyle index 3 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle4:TLineStyle index 4 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle5:TLineStyle index 5 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle6:TLineStyle index 6 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle7:TLineStyle index 7 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;
    property LineStyle8:TLineStyle index 8 read GetLineStyles write SetLineStyles stored IsNonDefaultStyle;

    property MarkStyle0:TMarkStyle index 0 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle1:TMarkStyle index 1 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle2:TMarkStyle index 2 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle3:TMarkStyle index 3 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle4:TMarkStyle index 4 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle5:TMarkStyle index 5 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle6:TMarkStyle index 6 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle7:TMarkStyle index 7 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    property MarkStyle8:TMarkStyle index 8 read GetMarkStyle write SetMarkStyle stored IsNonDefaultMarkStyle;
    {$ENDIF}

  end;

  TGraphPrintPreview = class(TPrintPreview)
  private
    FGraph: TGraphB;
  protected
    procedure SetGraph(G:TGraphB);
    procedure DoShowPreview(Sender:TObject; Canvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                            APage: Integer; OnPrinter:Boolean); override;
  public
    procedure SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double); override;
  published
    property Graph:TGraphB read FGraph write SetGraph;
  end;

  TGraphD = class;

  RecordType = (tyEnd, tyMap, tyPspace, tyMapLin, tyPoint, tyJoin, tyClipPoint, tyClipJoin,
                tyTextJ, tyTitlen, tyAxis, tyScales, tyAxes, tyTextAngle, tyStyle, tySymbol, tyRectangle,
                tyPolyStart, tyPolyPoint, tyPolyEnd, tyStyleT, tySymbolT, tyErrorBarX, tyErrorBarY, tyNoYaxis,
                tyXaxLab, tyYaxLab, tyRectangleEx, tyDefScale, tyGraphic);

  TReplotDEvent = procedure (Sender:TObject; G:TGraphD) of object;

  {$IFNDEF DYNAMIC_ARRAYS}
  PolygonPoints = array [0..8101] of TPoint;
  {$ENDIF}

  TDPoint = record
    X: Double;
    Y: Double;
  end;

  TDRect = record
    case Integer of
      0: (Left, Top, Right, Bottom: Double);
      1: (TopLeft, BottomRight: TDPoint);
  end;

  TGraphD = class(TGraphB)
  private
    FRecordCalls:Boolean;  (* Set to save calls for better replots *)
    FOnDReplot:TReplotDEvent;
    Recording:Boolean;
    RStream:TStream;
    FLastMouseX, FLastMouseY: Double;
    XofSave, XsSave, XminSave, YofSave, YsSave, YminSave: Double;
    FPolygonPoints: {$IFNDEF DYNAMIC_ARRAYS} ^PolygonPoints {$ELSE} array of TPoint {$ENDIF};
    NPolygonPoints: Integer;
    CMarkStyle: TMarkStyle;
    FXaxPos, FYaxPos: TDPoint;
    FZoomArea: TDRect;
    FPreviousZoomArea: TDRect;
    procedure SetRecordCalls(ARecording:Boolean);
    procedure DoGraphReplot(Sender:TObject; G:TGraphB);
    function GetSelectedMathArea: TDrect;
    procedure SetSelectedMathArea(ARect: TDRect);
    procedure SetZoomArea(const Value: TDRect);
    procedure SetAxisStyle(const Value: TLineStyle);
    function IsNotAxisDefault: Boolean;
    function GetAxisStyle: TLineStyle;
  protected
    FXmin:double;    (* User co-ordinate limits *)
    FXmax:double;
    FYmin:double;
    FYmax:double;
    Xlast:double;   (* Last argument to Point/Join *)
    Ylast:double;
    XLastPoly, YLastPoly: Double; (* Last argument to PolyPoint *)
    Xp1:double;     (* Fraction of available area to use for user co-ordinate limits *)
    Xp2:double;
    Yp1:double;
    Yp2:double;
    Xof:double;     (* For conversion to TGraphB co-ordinates *)
    Yof:double;
    Xs:double;
    Ys:double;
    ClipBoardStream: TStream;
    ClipFile: TextFile;
    procedure RecomputeScale;
    procedure Titlen(n:Integer; str:string);    (* Helper function *)
    procedure clppnt(var x1:double; var y1:double; x2:double; y2:double; var out:Integer);
    procedure Clamp(var x1:double; var y1:double; AOutCode:Integer);
    procedure PlayBack(s:TStream);
    procedure SetOnReplot(e:TReplotDEvent);
    procedure PutReals(t:RecordType; d:array of Double);
    procedure PutType(t:RecordType);
    procedure PutString(s:string);
    procedure SetLastMouse(X, Y: Integer); override;
    procedure ShowOnCanvas(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint); override;
    procedure ChangeZoom(newXSCALE, newYSCALE, newHscroll, newVscroll: Double); override;
    procedure ChangeScrollPos(Hpos, Vpos: double); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ExportTo(AStream: TStream; CanvasClass: TStreamCanvasClass); override;
    function ProcX(x:double):Integer;
    function ProcY(y:double):Integer;
    function PixelSize(x: Boolean): Double;
    function CloseToX(x1, x2: Double):Boolean;
    function CloseToY(y1, y2: Double):Boolean;
    function RoundToDisplay(value: Double; xvalue: Boolean): Double;
    procedure Map(xl:double; xu:double; yl:double; yu:double);
    procedure MapArg(var xl, xu, yl, yu:double);
    procedure Pspace(xl:double; xu:double; yl:double; yu:double);
    procedure PspaceArg(var xl, xu, yl, yu:double);
    procedure DefaultScale;
    procedure Start;
    procedure Finish;
    procedure Point(x:double; y:double);
    procedure Symbol(x:double; y:double; n:Integer);
    procedure SymbolT(x:double; y:double; AStyle: TMarkStyle);
    procedure ErrorBarX(x1, x2, y:double);
    procedure ErrorBarY(x, y1, y2:double);
    procedure Join(x:double; y:double);
    procedure ClipPoint(x:double; y:double);
    procedure ClipJoin(x1:double; y1:double);
    procedure Rectangle(x1, x2, y1, y2:double);
    procedure RectangleEx(AStyle: TBrushStyle; Density: Double; x1, x2, y1, y2:double);
    procedure DrawGraphic(x1, x2, y1, y2:Double; Graphic: TGraphic);
    procedure PolyStart;
    procedure PolyPoint(x1:double; y1:double);
    procedure PolyEnd;
    { Character Sizes }
    procedure Chlen(var Acx:double; var Acy:double);
    function  cx: Double;
    function  cy: Double;
    procedure Text(x:double; y:double; s:string);
    procedure TextJ(x:double; y:double; mode:TAlignment; s:string);
    procedure TextAngle(ang:double);
    procedure Style(AStyle:Integer);
    procedure StyleT(AStyle:TLineStyle);
    procedure StyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
    procedure Title1(title:string);
    procedure Title2(title:string);
    procedure Axis(xl1:double; xu1:double; yl:double; xaxis:Boolean; s:string);
    procedure ScaledAxis(xle, xue, xl1:double; xu1:double; yl:double; xaxis:Boolean; s:string);
    procedure NoYaxis;
    procedure Scales(xaxis:string; yaxis:string);
    procedure Axes(xaxis:string; yaxis:string);
    procedure XaxLab(xaxis:string); { Must be called *after* call to axis/axes/scales }
    procedure YaxLab(yaxis:string);
    function  PlotLength:Double;
    procedure MapLin(xl:double; xu:double; yl:double; yu:double);
    function  OutCode(x:double; y:double):Integer;
    procedure PlotChanged; override;
    procedure LPtoXY(lp: TPoint; var x, y:Double);
    procedure XYtoLP(var lp: TPoint; x, y:Double);
    procedure PreviousZoom; override;
    property LastMouseX: Double read FLastMouseX;
    property LastMouseY: Double read FLastMouseY;
    property SelectedMathArea: TDRect read GetSelectedMathArea write SetSelectedMathArea;
    property ZoomArea: TDRect read FZoomArea write SetZoomArea;
    property PreviousZoomArea: TDRect read FPreviousZoomArea write FPreviousZoomArea;
    property Xmin: Double read FXmin;    { User co-ordinate limits }
    property Xmax: Double read FXmax;
    property Ymin: Double read FYmin;
    property Ymax: Double read FYmax;
  published
    function CanPrint: Boolean; override;
    procedure CopyToClipboard; override;
    property RecordCalls:Boolean read FRecordCalls write SetRecordCalls;
    property OnReplot:TReplotDEvent read FOnDReplot write SetOnReplot;
    property AxisStyle:TLineStyle read GetAxisStyle write SetAxisStyle stored IsNotAxisDefault;
end;

procedure AxisInterval(range:double; var nticks:Integer; var ds:double; var se:double; var ne:Integer; var nd:Integer);
procedure DrawFocusRectHelper(Canvas: TCanvas; r: TRect);
function SmallX:Integer;
function SmallY:Integer;

const
  default_pen_colour: array [0..Maxpens-1] of TColor = (
    clBlack,
    clMaroon,
    clGreen,
    clYellow,
    clNavy,
    clPurple,
    clTeal,
    clRed,
    clLime,
    clBlue,
    clFuchsia,
    clGray,
    clDkGray,
    clOlive,
    clAqua,
    clWhite
   );

implementation

uses GraphPen, GPenDlg, GMarkDlg, MathSup, {$IFNDEF VER80} Math, {$ENDIF}
     {$IFDEF LCL_}
       LCLIntf,
       {$IFNDEF MSWINDOWS} GTKDef, {$IFNDEF GTK2} gdk, {$ELSE} gdk2, {$ENDIF} {$ENDIF}
     {$ENDIF}
     ExportG, TextAssign, HugeClip;


{$IFNDEF VER80}
{$IFNDEF FPC} { TODO: Remove when FPC bug fixed ? }
{$T+}
{$ENDIF}
{$ENDIF}

{$IFNDEF KYLIX}
function SmallX:Integer;
begin
  Result := {$IFDEF MSWINDOWS} GetSystemMetrics(SM_CXDOUBLECLK) {$ELSE} 3 {$ENDIF};
end;

function SmallY:Integer;
begin
  Result := {$IFDEF MSWINDOWS} GetSystemMetrics(SM_CYDOUBLECLK) {$ELSE} 3 {$ENDIF};
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
{$IFDEF VER80}
const
  NullMF = nil;
procedure DiscardMF(var hMF: TMetafile);
begin
  FreeAndNil(hMF);
end;
{$ELSE}
const
  NullMF = 0;
procedure DiscardMF(var hMF: HMETAFILE);
begin
  if hMF <> 0 then
    DeleteMetafile(hMF);
  hMF := 0;
end;
{$ENDIF}
{$ENDIF}
{$IFDEF CLX_}
procedure DiscardMF(var hMF: TDrawing);
begin
  FreeAndNil(hMF);
end;
const
  NullMF = nil;
{$ENDIF}
{$IFDEF KYLIX}
const
  SmallX = 3;
  SmallY = 3;
{$ENDIF}

function Min(a, b:Integer):Integer;
begin
  if a < b then
    Result := a
  else
    Result := b;
end;

procedure TGraphB.SetToDefaultStyle(Pen: Integer; AStyle: TLineStyle);
begin
  AStyle.Assign(FDefaultLineStyles[Pen]);
end;

procedure TGraphB.CLineStyleChanged(Sender:TObject);
begin
  FLinetype := -1;
end;

constructor TGraphB.Create(AOwner:TComponent);
var
  i:Integer;
begin
  inherited Create(AOwner);
  VertScrollBar.ScrollJump := 10;
  HorzScrollBar.ScrollJump := 10;
  Width := 100;
  Height := 100;
  CLineStyle := TLineStyle.Create;
  CLineStyle.OnChange := {$IFDEF FPC}@{$ENDIF}CLineStyleChanged;
  MarkLineStyle := TLineStyle.Create;
  MarkLineStyle.Width := 0;
  MarkLineStyle.Pattern := psSolid;
  for i := 0 to MaxPens do begin
    FLineStyles[i] := TLineStyle.Create;
    FDefaultLineStyles[i] := TLineStyle.Create;
    if i < MaxPens then begin
      FMarkStyles[i] := TMarkStyle.Create;
      FDefaultMarkStyles[i] := TMarkStyle.Create;
    end
  end;
  FErrorBarSize := 2;
  SymbolMark := 3;
  FGOptions := [grLandScape, grForceSymmetricalMarks];
  {$IFDEF MSWINDOWS}
  Font.Name := 'Arial';
  {$ENDIF}
  {$IFDEF UNIX}
  Font.Name := 'Helvetica';
  {$ENDIF}
  {$IFDEF LCL_}
  Font.Height := 12;
  {$ENDIF}
  ParentFont := false;
  Font.OnChange := {$IFDEF FPC}@{$ENDIF}FontHasChanged;
  Brush.OnChange := {$IFDEF FPC}@{$ENDIF}FontHasChanged;

  FCurrentCanvas := nil;
  DCType := dcNone;
  FStartPlottingCount := 1;
  LoadMP;
  FTextAngle := 0;
  {$IFNDEF CLX_}
  GraphicsFont := TFont.Create;
  UnrotatedGraphicsFont := TFont.Create;
  {$ENDIF}
  {$IFDEF METAFILES}
  hMF := NullMF;
  {$ENDIF}
  mp.PreparePoints := {$IFDEF FPC}@{$ENDIF}PreparePoints;
  for i := 0 to MaxPens-1 do begin
    with FLineStyles[i] do begin
      Width := 0;
      Colour := default_pen_colour[i];
      Pattern := TLinePattern(i mod MAXSTYLE);
    end;
    FLineStyles[i].OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
    FMarkStyles[i].Colour := default_pen_colour[i];
    FMarkStyles[i].MarkType := TMarkType(i mod nMarkTypes);
    FMarkStyles[i].DotAtCentre := True;
    FMarkStyles[i].OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
  end;
  with FLineStyles[AxisPen] do begin
    Width := 0;
    Colour := clBlack;
    Pattern := psSolid;
    OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
  end;
  XSCALE := 1;
  YSCALE := 1;   { User selected plot scale }
  FXmm := 250;
  FYmm := 170;
  ResetZoom;
  SetScaling(Canvas, 0, 0, Width, Height); { Sensible defaults for scaling }
end;

procedure TGraphB.SetCurrentCanvas(ACanvas: TExportableCanvas);
begin
  FreeAndNil(FCurrentCanvas);
  FCurrentCanvas := ACanvas;
end;

procedure TGraphB.SetRealCanvas(ACanvas: TCanvas);
begin
  SetCurrentCanvas(TCanvasCanvas.Create(ACanvas));
end;

function TGraphB.PenDialog(Pen:Integer):Boolean;
var
 Dialog:TGraphPenDialog;
begin
  Dialog := TGraphPenDialog.Create(Application);
  try
    Dialog.Graph := Self;
    Dialog.Pen := Pen;
    Result := Dialog.Execute;
    if Result then
      Dialog.Apply;
  finally
    Dialog.Free;
  end;
end;

function TGraphB.MarkDialog(AMark:Integer):Boolean;
var
 Dialog:TGraphMarkDialog;
begin
  Dialog := TGraphMarkDialog.Create(Application);
  try
    Dialog.Graph := Self;
    Dialog.Mark := AMark;
    Result := Dialog.Execute;
    if Result then
      Dialog.Apply;
  finally
    Dialog.Free;
  end;
end;

procedure TGraphB.DoOptionChanged(Sender:Tobject);
begin
  OptionChanged;
end;

procedure TGraphB.Loaded;
var
  i: Integer;
begin
  inherited Loaded;
  for i := 0 to MaxPens do begin
    FDefaultLineStyles[i].Assign(FLineStyles[i]);
    if i < MaxPens then
      FDefaultMarkStyles[i].Assign(FMarkStyles[i]);
  end;
  if not (csDesigning in ComponentState) and not Assigned(PopupMenu) then begin
    PopupMenu := TPopupMenu.Create(Self);
    PopupMenu.Items.Add(TMenuItem.Create(PopupMenu));
    with PopupMenu.Items[0] do begin
      Caption := 'Copy to clipboard';
      Hint := 'Use Paste Special to paste picture or underlying data (if available) in the target application';
      OnClick := {$IFDEF FPC}@{$ENDIF}DoCopyToClipboard;
    end;
    PopupMenu.Items.Add(TMenuItem.Create(PopupMenu));
    with PopupMenu.Items[1] do begin
      Caption := 'Font...';
      Hint := 'Choose font for graph text';
      OnClick := {$IFDEF FPC}@{$ENDIF}DoChooseFont;
    end;
  end;
end;

function TGraphB.GetLineStyles(i:Integer):TLineStyle;
begin
  Result := FLineStyles[i];
end;

function TGraphB.IsNonDefaultStyle(i:Integer):Boolean;
begin
  Result := (FLineStyles[i].Width <> 0)
            or (FLineStyles[i].Colour <> default_pen_colour[i])
            or (Ord(FLineStyles[i].Pattern) <> (i mod MAXSTYLE));
end;

procedure TGraphB.SetLineStyles(i:Integer; Style:TLineStyle);
begin
  FLineStyles[i].Assign(Style);
end;

function TGraphB.GetMarkStyle(i:Integer):TMarkStyle;
begin
  Result := FMarkStyles[i];
end;

function TGraphB.IsNonDefaultMarkStyle(i:Integer):Boolean;
begin
  Result := (FMarkStyles[i].Size <> 2)
            or (FMarkStyles[i].Colour <> default_pen_colour[i])
            or (Ord(FMarkStyles[i].MarkType) <> (i mod nMarkTypes))
            or not FMarkStyles[i].DotAtCentre;
end;

procedure TGraphB.SetMarkStyle(i:Integer; Style:TMarkStyle);
begin
  FMarkStyles[i].Assign(Style);
end;

procedure TGraphB.Setmm(x, y: Double);
begin
  if (x <> Xmm) or (y <> Ymm) then begin
    FXmm := x;
    FYmm := y;
    OptionChanged;
  end;
end;

procedure TGraphB.SetXmm(v:Double);
begin
  Setmm(v, FYmm);
end;

procedure TGraphB.SetYmm(v:Double);
begin
  Setmm(FXmm, v);
end;

procedure TGraphB.SetGOptions(opt:TGraphOptions);
begin
  if opt <> FGOptions then begin
    FGOptions := opt;
    OptionChanged;
  end;
end;

function TGraphB.IsRealDevice:Boolean;
begin
  Result := DCType >= dcScreen;
end;

function TGraphB.IsClipboard:Boolean;
begin
  Result := DCType = dcClip;
end;

function TGraphB.IsPrinter:Boolean;
begin
  Result := DCType >= dcPreview;
end;

function TGraphB.IsScreen:Boolean;
begin
  Result := (DCType = dcScreen) or (DCType = dcScreenBuffer) or (DCType = dcNone);
end;

procedure TGraphB.start_plotting;        (* Call to ensure DC present *)
begin
  if DCType = dcNone then begin
    FStartPlottingCount := 0;
    DCType := dcScreen;
    SetRealCanvas(Canvas);
    WIN_change_font;
  end;
  Inc(FStartPlottingCount);
end;

procedure TGraphB.stop_plotting;         (* Call when done with DC *)
begin
  Dec(FStartPlottingCount);
  if FStartPlottingCount = 0 then begin
    if DCType <> dcScreen then
      raise Exception.Create('stop_plotting mismatch');
    DCType := dcNone;
    SetCurrentCanvas(nil);
    FStartPlottingCount := 1; { Other DCTypes will not use this mechanism }
  end;
end;

function TGraphB.land_sea:Boolean;
begin
  Result := grLandscape in GOptions;
end;

function TGraphB.ClientY:Integer;
begin
  if land_sea then
    Result := PaintHeight
  else
    Result := PaintWidth;
end;

function TGraphB.ClientX:Integer;
begin
  if not land_sea then
    Result := PaintHeight
  else
    Result := PaintWidth;
end;

procedure TGraphB.InitFont;
var
  ph: Integer;
begin
  CurFontHeight := 0;
  {$IFNDEF CLX_}
  GetObject(Font.Handle, sizeof(GraphicsLogFont), @GraphicsLogFont);
  GraphicsLogFont.lfOutPrecision := OUT_TT_ONLY_PRECIS;
  GraphicsLogFont.lfClipPrecision := CLIP_LH_ANGLES;
  {$ELSE}
  if CurrentCanvas <> nil then
    CurrentCanvas.Font.Assign(Font)
  else
    Canvas.Font.Assign(Font);
  {$ENDIF}
  if (grScaledFont in GOptions) and (IsPrinter or IsClipboard or (DCType = dcExport)) then
    FontSize := Font.Size / 72 * 25.4 / Ymm * YPixels
  else begin
    ph := PaintHeight;
    if ph = 0 then
      ph := Height; { In case paint rect hasn't been calculated yet }
    if ph = 0 then
      ph := FAvailableHeight;
    if ph = 0 then
      ph := 1;
    FontSize := Font.Height / ph * FAvailableHeight;
    if IsScreen then
      FontSize := FontSize / YSCALE;
  end;
  {$IFNDEF CLX_}
  if Abs(FontSize) < 32000 then
    GraphicsLogFont.lfHeight := Round(FontSize);
  {$ENDIF}
  {$IFDEF VCL_}
  text_flags := TA_LEFT;
  {$ENDIF}
  FTextAngle := 0;
  FJustify := taLeftJustify;
  FTextSize := 1;
  {$IFDEF CLX_}
  WIN_change_font;
  {$ENDIF}
end;

procedure TGraphB.InitGraphics;
begin
  InitFont;
  DRAW_FLAG := False;
  FLinetype := 0;
  setup_graphics_DC;
  LoadMP;
end;

procedure TGraphB.Resize;
begin
  inherited Resize;
  Invalidate;
end;

{$IFDEF METAFILES}
procedure TGraphB.StartMetafile;
begin
  DiscardMF(hMF);
  SetMetafileScaling;
  {$IFDEF MSWINDOWS}
  {$IFDEF VER80}
  hMF := TMetafile.Create;
  {$ENDIF}
  MetafileCanvas := TCanvas.Create;
  MetafileCanvas.Handle := CreateMetafile(nil);
  SetWindowExtEx(MetafileCanvas.Handle, FXPixels, FYPixels, nil);
  SetRealCanvas(MetafileCanvas);
  {$ENDIF}
  {$IFDEF UNIX}
  hMF := TDrawing.Create;
  hMF.Height := FYPixels;
  hMF.Width := FXPixels;
  hMF.Canvas.Start(True);
  SetRealCanvas(hMF.Canvas);
  {$ENDIF}
end;

procedure TGraphB.StopMetafile;
begin
  { Finish Metafile }
  {$IFDEF MSWINDOWS}
  {$IFDEF VER80}
  hMF.Handle := CloseMetaFile(MetafileCanvas.Handle);
  hMF.Width := FXPixels;
  hMF.Height := FYPixels;
  {$ELSE}
  hMF := CloseMetaFile(MetafileCanvas.Handle);
  {$ENDIF}
  {$ENDIF}
  {$IFDEF UNIX}
  hMF.Canvas.Stop;
  {$ENDIF}
  SetCurrentCanvas(nil);
end;
{$ENDIF}

procedure TGraphB.Start;
var
  ACanvas: TCanvas;
begin
  if not InReplot then begin
    if CurrentCanvas = nil then
      if Assigned(FOnReplot) then begin
        { Assume direct screen access required }
        FreeAndNil(FScreenBuffer);
        if DoubleBuffered then begin
          FScreenBuffer := TBitmap.Create;
          FScreenBuffer.Height := PaintHeight;
          FScreenBuffer.Width := PaintWidth;
          DCType := dcScreenBuffer;
          ACanvas := FScreenBuffer.Canvas;
          SetRealCanvas(ACanvas);
          SetScreenScaling;
          { Bitmap only holds visible part of screen }
          FXOffset := HorzScrollBar.Position;
          FYOffset := VertScrollBar.Position;
        end else begin
          DCType := DCScreen;
          ACanvas := Canvas;
          SetRealCanvas(ACanvas);
          SetScreenScaling;
        end;
        ACanvas.Brush.Color := Color;
        ACanvas.FillRect(Canvas.ClipRect);
      {$IFDEF METAFILES}
      end else begin
        StartMetafile;
        DCType := dcMF;
      {$ENDIF}
      end;
    InitGraphics;
  end;
end;

procedure TGraphB.cancel_graphics_DC;
begin
  DCType := dcNone;
  { Leave scaling and other things in place for mouse }
  start_plotting;
  SetScreenScaling;
  InitGraphics;
  stop_plotting;
end;

procedure TGraphB.Finish;
var
  NeedInvalidate: Boolean;
begin
  if not InReplot then begin
    NeedInvalidate := True;
    if CurrentCanvas <> nil then begin
      end_draw;
      if CurrentCanvas.RealCanvas <> Canvas then
       {$IFDEF METAFILES}
        if DCType = dcMF then begin
          StopMetafile;
        end else
        {$ENDIF}
        if DCType = dcScreenBuffer then begin
          Canvas.Draw(PaintRect.Left - HorzScrollBar.Position,
                      PaintRect.Top - VertScrollBar.Position,
                      FScreenBuffer);
          FreeAndNil(FScreenBuffer);
          NeedInvalidate := False;
        end else
          raise Exception.Create('Help!');
      SetCurrentCanvas(nil);
      cancel_graphics_DC;
    end;
    if NeedInvalidate then
      Invalidate;
    FTextAngle := 0;
  end
end;

procedure TGraphB.Suspend;
begin
  if not InReplot and not Assigned(FOnReplot) then begin
    FSavedAngle := TextAngle;
    FSavedTextSize := TextSize;
    FSavedLinestyle := FLinetype;
    FSavedJustify := FJustify;
    Finish;
  end;
end;

procedure TGraphB.Resume;
{$IFDEF MSWINDOWS}
var
  LastMF: {$IFDEF VER80} TMetafile; {$ELSE} HMETAFILE; {$ENDIF}
{$ENDIF}
begin
  if not InReplot and not Assigned(FOnReplot) then begin
    {$IFDEF METAFILES}
    if hMF <> NullMF then begin
      {$IFDEF MSWINDOWS}
        LastMF := hMF;
        hMF := NullMF;
        StartMetafile;
        {$IFDEF VER80}
          CurrentCanvas.Draw(0, 0, LastMF);
        {$ELSE}
          PlayMetafile(CurrentCanvas.RealCanvas.Handle, LastMF);
        {$ENDIF}
        DiscardMF(LastMF);
      {$ENDIF}
      {$IFDEF UNIX}
        SetMetafileScaling;
        hMF.Canvas.Start(True);
        SetRealCanvas(hMF.Canvas);
      {$ENDIF}
    end;
    {$ENDIF}
    Start;
    TextSize := FSavedTextSize;
    Angle(FSavedAngle);
    Justify(FSavedJustify);
    if FSavedLinestyle >= 0 then
      Linetype := FSavedLinestyle
    else
      FLinetype := -1;
  end;
end;

procedure TGraphB.HintDisplacement(Displacement:Integer; var v, h:Integer; AllowZero:Boolean);
var
  t: Integer;
begin
  v := Abs(Displacement);
  if not AllowZero and (v = 0) then v := 1;
  h := MulDiv(v, DPIX, DPIY);
  if not land_sea then begin
    t := v;
    v := h;
    h := t;
  end;
end;

procedure TGraphB.PenInvalid;
begin
  CurrentPenSize := -1;
  penstyle := MAXSTYLE+10;
end;

procedure TGraphB.setup_graphics_DC;
begin
  PenInvalid;
  WIN_change_font;
end;

procedure TGraphB.flush_polyline;
begin
  if nextseg > 1 then begin
    start_plotting;
    NeedLinePen;
    {$IFDEF VER80}
    Polyline(CurrentCanvas.Handle, curseg, nextseg);
    {$ELSE}
    CurrentCanvas.Polyline(curseg, nextseg);
    {$ENDIF}
    stop_plotting;
  end;
end;

procedure TGraphB.end_draw;
begin
  if DRAW_FLAG and (nextseg > 1) then begin
    flush_polyline;
    nextseg := 0;
  end;
  DRAW_FLAG := False;
  segleft := pattern[0];
  if penstep >=0 then penstep := 0;
end;

procedure TGraphB.LT(x:Integer; y:Integer);
begin
  if (nextseg > High(curseg)) then begin
    flush_polyline;
    curseg[0] := curseg[nextseg-1];
    nextseg := 1;
  end;
  LogToDev(x,y);
  curseg[nextseg].x := x;
  curseg[nextseg].y := y;
  Inc(nextseg);
end;

procedure TGraphB.set_pensize(p:Double);
begin
  CurrentPenSize := p;
  if Font.Size = 0 then
    p := FontSize
  else
    p := Abs(p / Font.Size * FontSize);
  PenSize := Round(p);
  if not IsRealDevice and (p < Abs(FontSize) / 20) then
    HintDisplacement(Round(Abs(FontSize) / 20), mp.BlobV, mp.BlobH, False)
  else
    HintDisplacement(Round(p/2), mp.BlobV, mp.BlobH, IsRealDevice);
  if CurrentCanvas <> nil then begin
{    QPainter_setPen(DCHandle, QColor(mp.rgbColour));
     QPainter_setBrush(DCHandle, QColor(mp.rgbColour)); }
     with CurrentCanvas.Pen do begin
       Color := mp.rgbColour;
       Mode := pmCopy;
       Style := {$IFDEF CLX_}QGraphics.{$ELSE}Graphics.{$ENDIF}psSolid;
       Width := PenSize;
     end;
     with CurrentCanvas.Brush do begin
       Color := mp.rgbColour;
       Style := bsSolid;
     end;
  end;
end;

(*
 * select a line style/type and scale/pitch
 *)
procedure TGraphB.set_linestyle(ALinestyle:Integer; linescale:Double);
var
  i: Integer;
  p:TPat;
begin
  penstyle := ALinestyle;
  if (penstyle <= 0) or (penstyle >= MAXSTYLE) then begin
    (* Simple solid line or dots at end of segments *)
    SETDOT := ALinestyle = MAXSTYLE;
    if not SETDOT then penstyle := 0;
    SYMBOL := false;
    penstep := -1;
  end else begin
    p := pats[penstyle];
    pattern_length := p.length;
    linescale := linescale / p.scale;
    for i := 0 to pattern_length-1 do
      pattern[i] := Round(p.segments[i] * linescale);
    SETDOT := False;
    penstep := 0;
    segleft := p.segments[0];
  end;
end;

procedure TGraphB.NeedPen(AStyle: TLineStyle);
var
  newsize: Double;
  NewColour: TColor;
  NewPenStyle: Integer;
begin
  if grColouredPens in GOptions then
    NewColour := ColorToRGB(AStyle.Colour)
  else NewColour := 0;
  if grPensHaveWidth in GOptions then
    NewSize := AStyle.Width
  else NewSize := 0;
  if grPensHaveStyle in GOptions then
    NewPenStyle := Ord(AStyle.Pattern)
  else
    NewPenStyle := 0;
  if penstyle <> NewPenStyle then
    set_linestyle(NewPenStyle, abs(FontSize*2));
  if (NewColour <> mp.rgbColour) or
     (NewSize <> CurrentPenSize) then begin
    mp.rgbColour := NewColour;
    set_pensize(NewSize);
  end;
end;

procedure TGraphB.NeedLinePen;
begin
  NeedPen(Linestyle);
end;

procedure  TGraphB.SetLinetype(ipen:Integer);
begin
  end_draw;
  if (ipen > MAXPENS-1) or (ipen < 0) then ipen := ipen mod MAXPENS;
  if ipen < 0 then ipen := ipen + MAXPENS;
  ClineStyle.Assign(Linestyles[ipen]);
  FLinetype := ipen;
end;

procedure TGraphB.SetLineStyle(AStyle: TLineStyle);
begin
  end_draw;
  ClineStyle.Assign(Astyle);
  FLinetype := -1;
end;

function TGraphB.GetLineStyle:TLineStyle;
begin
  if FLinetype < 0 then
    Result := ClineStyle
  else
    Result := Linestyles[FLinetype];
end;

procedure TGraphB.LineStyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
begin
  end_draw;
  ClineStyle.Colour := Colour;
  ClineStyle.Width := AWidth;
  ClineStyle.Pattern := APattern;
  FLinetype := -1;
end;

procedure TGraphB.LoadMP;
begin
  mp.IsRealDevice := IsRealDevice;
  mp.ForceSymmetricalMarks := grForceSymmetricalMarks in GOptions;
end;

procedure TGraphB.plot_blob(x:Integer; y:Integer);    { Plot a dot }
begin
  Dot(CurrentCanvas, x, y, mp);
end;

procedure  TGraphB.Vector(x:Integer; y:Integer);
var
  length, x1, y1, f:Double;
begin
  start_plotting;
  NeedLinePen;
  if (SETDOT) then begin
    if not DRAW_FLAG then plot_blob(absX, absY);
    plot_blob(x, y);
  end else begin
    if penstep<0 then
      LT(x, y)
    else begin
      length := sqrt(1.*(x-absX)*(x-absX)+1.*(y-absY)*(y-absY));
      x1 := absX;
      y1 := absY;
      while length > 0 do
        if length > segleft then begin
          if segleft > 0 then begin
            f := segleft / length;
            x1 := x1 + f*(x-x1);
            y1 := y1 + f*(y-y1);
            if (penstep and 1) <> 0 then
              MT(Round(x1), Round(y1))
            else
              LT(Round(x1), Round(y1));
            length := length - segleft;
          end else if pattern[penstep] < 0 then
            plot_blob(absX, absY); (* First dot in line *)
          repeat    (* onto next segment of pattern *)
            Inc(penstep);
            if penstep >= pattern_length then
              penstep := 0;
            segleft := pattern[penstep];
            if segleft < 0 then
              plot_blob(Round(x1), Round(y1));
          until segleft > 0;
        end else begin
          if (penstep and 1) <> 0 then
            MT(x, y)
          else
            LT(x, y);
          segleft := segleft - length;
          length := -1;
        end;
    end;
    if SYMBOL then Mark(x, y, SymbolMark);
  end;
  absX := x;
  absY := y;
  DRAW_FLAG := TRUE;
  stop_plotting;
end;

procedure TGraphB.MT(x:Integer; y:Integer);
begin
  flush_polyline;
  nextseg := 0;
  LT(x, y);
end;

procedure TGraphB.Move(x:Integer; y:Integer);
begin
  end_draw;
  absX := x;
  absY := y;
  if (SYMBOL) then begin
    start_plotting;
    Mark(x, y, SymbolMark);
    stop_plotting;
  end else
    MT(absX, absY);
end;

procedure TGraphB.Box(Style: TBrushStyle; Density: Double; const Rect: TRect);
var
  p: TRect;
  function ScaleColour(Value: Integer): Integer;
  begin
    { Interpolate colour:
       - Density = 1 => full colour => value unchanged
       - Density = 0 => white => 255 }
    Result := Round((Value-255)*Density+255);
    { Sanity check - probably never happens }
    if Result < 0 then
      Result := 0
    else if Result > 255 then
      Result := 255;
  end;
begin
  start_plotting;
  NeedLinePen;
  if (Density >= 0) and (Density < 1) then
    CurrentCanvas.Brush.Color := RGB(ScaleColour(GetRValue(mp.rgbColour)),
                                     ScaleColour(GetGValue(mp.rgbColour)),
                                     ScaleColour(GetBValue(mp.rgbColour)));
  p := Rect;
  LogToDev(p.Left, p.Top);
  LogToDev(p.Right, p.Bottom);
  if Style <> bsSolid then begin
    CurrentCanvas.Brush.Style := Style;
    {$IFDEF MSWINDOWS}
    { TRANSPARENT seems to be ignored, so set sensible background colour }
    if CurrentCanvas.RealCanvas <> nil then
      SetBkColor(CurrentCanvas.RealCanvas.Handle, ColorToRGB(Color));
    {$ENDIF}
  end;
  CurrentCanvas.FillRect(p);
  { Restore Brush }
  CurrentCanvas.Brush.Color := mp.rgbColour;
  CurrentCanvas.Brush.Style := bsSolid;
  stop_plotting;
end;

procedure TGraphB.ClearBox(const Rect: TRect);
begin
  CurrentCanvas.Brush.Color := Color;
  CurrentCanvas.FillRect(Rect);
  { Restore Brush }
  CurrentCanvas.Brush.Color := mp.rgbColour;
end;

procedure TGraphB.DrawGraphic(Rect: TRect; Graphic: TGraphic);
var
  t: Integer;
begin
  start_plotting;
  LogToDev(Rect.Left, Rect.Top);
  LogToDev(Rect.Right, Rect.Bottom);
  t := Rect.Top;
  Rect.Top := Rect.Bottom;
  Rect.Bottom := t;
  {$IFDEF OFFLINUX}
  if Graphic.Transparent and (Graphic is TBitmap) and Assigned(CurrentCanvas.RealCanvas) then
    StretchBlt(CurrentCanvas.RealCanvas.Handle, Rect.Left, Rect.Top, Rect.Right-Rect.Left, Rect.Bottom-Rect.Top,
               TBitmap(Graphic).Canvas.Handle, 0, 0, Graphic.Width, Graphic.Height, SRCINVERT)
  else
  {$ENDIF}
    {$IFNDEF MSWINDOWS}
    CurrentCanvas.StretchDraw(Rect, Graphic);
    {$ELSE} // It should be possible to get this to work?
//    if Graphic.Transparent and (Graphic is TBitmap) and (CurrentCanvas.RealCanvas <> nil) then begin
//      StretchBlt(CurrentCanvas.RealCanvas.Handle, Rect.Left, Rect.Top, Rect.Right-Rect.Left, Rect.Bottom-Rect.Top,
//                 TBitmap(Graphic).Canvas.Handle, 0, 0, Graphic.Width, Graphic.Height, SRCAND)
//    end else
      CurrentCanvas.StretchDraw(Rect, Graphic);
    {$ENDIF}
  stop_plotting;
end;

procedure TGraphB.FontHasChanged(Sender:TObject);
begin
  CurFontHeight := 0;
  if CurrentCanvas <> nil then
    WIN_change_font;
  OptionChanged;
end;

function TGraphB.AvCharWidth:Double;
begin
  Result := CellHeight; { This Sets the CurFont10Width and other variables }
  Result := Result * CurFont10Width / 10 /(CurFontHeight + CurFontLeading) / YPerX;
  {$IFDEF FIX_CHARACTER_BOX_SIZE_ON_ZOOM}
  if IsRealDevice then
    if land_sea then
      Result := Result / YSCALE * XSCALE
    else
      Result := Result / XSCALE * YSCALE;
  {$ENDIF}
end;

{ TODO: Use?
function TGraphB.ScreenAvCharWidth:Integer;
begin
  Result := Round(AvCharWidth/ScaleFactor);
end;

function TGraphB.ScreenCellHeight:Integer;
begin
  Result := Round(CellHeight/ScaleFactor);
end;
}

function TGraphB.CellHeight:Double;
var
  sz: TSize;
  Cnvs: TCanvas;
  {$IFDEF CLX_}
  tm: QFontMetricsH;
  {$ELSE}
  tm: TTextMetric;
  {$ENDIF}
  {$IFDEF VCL_}
  DC: HDC;
  {$ENDIF}
begin
  if DCType = dcNone then InitFont;
  {$IFDEF VCL_}
  if CurFontHeight = 0 then begin
    if IsPrinter then begin
      Cnvs := CurrentCanvas.RealCanvas;
      if RealTextAngle <> 0 then
        Cnvs.Font.Assign(UnrotatedGraphicsFont);
    end else begin
      Cnvs := TCanvas.Create;
      DC := GetDC(0);
      Cnvs.Handle := DC;
      Cnvs.Font.Assign(Font);
    end;
    {$IFDEF VER80}
    sz.cx := Cnvs.TextWidth('0123456789');
    sz.cy := Cnvs.TextHeight('0123456789');
    {$ELSE}
    sz := Cnvs.TextExtent('0123456789');
    {$ENDIF}
    GetTextMetrics(Cnvs.Handle, tm);
    CurFontHeight := tm.tmHeight;
    CurFontLeading := tm.tmExternalLeading;
    CurFontInternalLeading := tm.tmInternalLeading;
    CurFontAscent := tm.tmAscent;
    if CurFontHeight = 0 then { Driver bug, or just very small font - get a number from somewhere }
      CurFontHeight := sz.cy;
    if CurFontHeight = 0 then
      CurFontHeight := 1;
    CurFont10Width := sz.cx;
    if not IsPrinter then begin
      Cnvs.Free;
      ReleaseDC(0,DC);
    end;
  end;
  if IsPrinter then
    Result := CurFontHeight + CurFontLeading
  else if FontSize < 0 then
    Result := -FontSize * (CurFontHeight + CurFontLeading) / (CurFontHeight - CurFontInternalLeading)
  else
    Result := FontSize * (CurFontHeight + CurFontLeading) / CurFontHeight;
  {$ENDIF}
  {$IFDEF LCL_}
  if (DCType = dcNone) or (CurrentCanvas.RealCanvas = nil) then
    Cnvs := Canvas
  else
    Cnvs := CurrentCanvas.RealCanvas;
  sz := Cnvs.TextExtent('0123456789'); { Need to unrotate font? }
  CurFont10Width := sz.cx;
  if IsPrinter then begin
    { Printer does not have a proper handle }
    Result := sz.cy;
    { TODO: Refine the values below }
    CurFontHeight := sz.cy;
    CurFontLeading := 0;
    CurFontInternalLeading := 0;
    CurFontAscent := sz.cy;
  end else begin
    GetTextMetrics(Cnvs.Handle, tm);
    CurFontHeight := tm.tmHeight;
    CurFontLeading := tm.tmExternalLeading;
    CurFontInternalLeading := tm.tmInternalLeading;
    CurFontAscent := tm.tmAscent;
    if CurFontHeight = 0 then { Driver bug, or just very small font - get a number from somewhere }
      CurFontHeight := sz.cy;
    if CurFontHeight = 0 then
      CurFontHeight := 1;
    if FontSize < 0 then
      Result := -FontSize * (CurFontHeight + CurFontLeading) / (CurFontHeight - CurFontInternalLeading)
    else
      Result := FontSize * (CurFontHeight + CurFontLeading) / CurFontHeight;
   end;
  {$ENDIF}
  {$IFDEF CLX_}
  if (DCType = dcNone) or (CurrentCanvas.RealCanvas = nil) then
    Cnvs := Canvas
  else
    Cnvs := CurrentCanvas.RealCanvas;
  sz := Cnvs.TextExtent('0123456789');
  CurFont10Width := sz.cx;
  tm := QFontMetrics_create(Cnvs.Font.Handle);
  CurFontHeight := Max(1, QFontMetrics_height(tm));
  CurFontLeading := QFontMetrics_leading(tm);
  QFontMetrics_destroy(tm);
{  if IsPrinter then }
    Result := CurFontHeight + CurFontLeading;
{  else
    Result := Abs(FontSize) * (CurFontHeight + CurFontLeading) / CurFontHeight; }
  {$ENDIF}
{$IFDEF FIX_CHARACTER_BOX_SIZE_ON_ZOOM}
  if IsScreen then
    if land_sea then
      Result := Result * YSCALE
    else
      Result := Result * YPerX * XSCALE
  else
{$ENDIF}
    if not land_sea then
      Result := Result * YPerX;
end;

procedure TGraphB.WIN_change_font; (* Update font from GraphicsLogFont;
                                     Angle of rotation from TextAngle *)
var
 Fnt: TFont;
 {$IFDEF VCL_}
 procedure STA(Mode: Uint);
 begin
   if not Assigned(CurrentCanvas) then
     SetTextAlign(Canvas.Handle, Mode)
   else
     SetTextAlign(CurrentCanvas.RealCanvas.Handle, Mode);
 end;
 {$ENDIF}
begin
  if land_sea then
    if grUpsideDown in GOptions then
      RealTextAngle := TextAngle + 1800
    else
      RealTextAngle := TextAngle
  else begin
    if grUpsideDown in GOptions then
      RealTextAngle := TextAngle + 900
    else
      RealTextAngle := TextAngle + 2700
  end;
  if not Assigned(CurrentCanvas) then
    Fnt := Canvas.Font
  else
    Fnt := CurrentCanvas.Font;
  {$IFNDEF CLX_}
  if IsRealDevice and (RealTextAngle <> 0) then begin
    UnrotatedGraphicsLogFont := GraphicsLogFont;
    UnrotatedGraphicsLogFont.lfEscapement := 0;
    UnrotatedGraphicsLogFont.lfOrientation := 0;
    {$IFDEF VCL_}
    UnrotatedGraphicsFont.PixelsPerInch := Fnt.PixelsPerInch;
    UnrotatedGraphicsFont.Handle := CreateFontIndirect(UnrotatedGraphicsLogFont);
    {$ENDIF}
    {$IFDEF LCL_}
    UnrotatedGraphicsFont.Handle := CreateFontIndirect(UnrotatedGraphicsLogFont);
    {$ENDIF}
  end;
  GraphicsLogFont.lfOrientation := RealTextAngle;
  GraphicsLogFont.lfEscapement := RealTextAngle;
  {$IFDEF VCL_}
  if DCType <> dcExport then
    if IsRealDevice and (RealTextAngle <> 0) then
      STA(TA_LEFT or TA_TOP or TA_NOUPDATECP)
    else
      STA(text_flags or TA_BASELINE or TA_NOUPDATECP);
  GraphicsFont.PixelsPerInch := Fnt.PixelsPerInch;
  GraphicsFont.Handle := CreateFontIndirect(GraphicsLogFont);
  Fnt.Assign(GraphicsFont);
  {$ENDIF}
  {$IFDEF LCL_}
  GraphicsFont.Handle := CreateFontIndirect(GraphicsLogFont);
  Fnt.Assign(GraphicsFont);
  {$ENDIF}
  { SetTextColor(DCHandle, ColorToRGB(Font.Color));
    SelectObject(DCHandle, graphics_font); }
  {$ELSE}
  Fnt.Height := Round(FontSize);
  {$ENDIF}
  Fnt.Color := Font.Color;
  if RealTextAngle = 0 then begin
    s_TextAngle := 0;
    c_TextAngle := 1;
  end else begin
    s_TextAngle := sin(RealTextAngle*0.1*Pi/180);
    c_TextAngle := cos(RealTextAngle*0.1*Pi/180);
  end;
  CurrentMarkSize := -1 ;
end;

procedure  TGraphB.SetTextSize(s:Integer); (* Scale up text size by s/100 *)
begin
  if s <> TextSize then begin
    FTextSize := s;
    {$IFNDEF CLX_}
    GraphicsLogFont.lfHeight := Round((FontSize*s)/100);
    {$ENDIF}
    WIN_change_font;
  end;
end;

(*
procedure TGraphB.set_window_orientation(DCHandle:HDCPainter);
  procedure SetWindow(x, y, w, h: Integer);
  begin
    {$IFDEF MSWINDOWS}
    SetWindowExtEx(DCHandle, w, h, Nil);
    SetWindowOrgEx(DCHandle, x, y, Nil);
    {$ENDIF}
    {$IFDEF LINUX}
    QPainter_setWindow(DCHandle.Handle, x, y, w, h);
    {$ENDIF}
  end;
begin
  if land_sea then
    if grUpsideDown in GOptions then
      SetWindow(XPixels, 0, -XPixels, YPixels)
    else
      SetWindow(0, YPixels, XPixels, -YPixels)
  else
    if grUpsideDown in GOptions then
      SetWindow(YPixels, XPixels, -YPixels, -XPixels)
    else
      SetWindow(0, 0, YPixels, XPixels);
end;
*)

procedure TGraphB.LogToDev(var x, y: Integer); { Convert from logical to device pixels; }
var
  t: Integer;
begin
  if land_sea then
    if grUpsideDown in GOptions then
      x := XPixels - x
    else
      y := YPixels - y
  else begin
    t := x;
    if grUpsideDown in GOptions then begin
      x := YPixels - y;
      y := XPixels - t;
    end else begin
      x := y;
      y := t;
    end;
  end;
  Inc(x, FXOffset);
  Inc(y, FYOffset);
end;

procedure TGraphB.LogToDev(var Point: TPoint);
begin
  LogToDev(Point.x, Point.y);
end;

procedure TGraphB.DevToLog(var x, y: Integer); { Convert from device to logical pixels; }
var
  t: Integer;
begin
  Dec(x, FXOffset);
  Dec(y, FYOffset);
  if land_sea then
    if grUpsideDown in GOptions then
      x := XPixels - x
    else
      y := YPixels - y
  else begin
    t := x;
    if grUpsideDown in GOptions then begin
      x := YPixels - y;
      y := XPixels - t;
    end else begin
      x := y;
      y := t;
    end;
  end;
end;

procedure TGraphB.DevToLog(var Point: TPoint);
begin
  DevToLog(Point.x, Point.y);
end;

procedure TGraphB.PreparePoints(var p: array of TPoint; np:Integer);
var
  i: Integer;
begin
  for i := 0 to np-1 do begin
    LogToDev(p[i].x, p[i].y);
  end;
end;

function TGraphB.TextExtentF(str:string; Fraction: Single {$IFDEF NONWINLCL}; var sz: TSize {$ENDIF}):TPoint;
var
  {$IFDEF MSWINDOWS}
  tm : TTextMetric;
  {$ELSE}
  {$IFDEF CLX_}
  sz : TSize;
  {$ENDIF}
  {$IFDEF LCL_}
  lbearing, rbearing, descent: LongInt;
  {$ENDIF}
  {$ENDIF}
  v,h: Single;
  xf, yf: Double;
  Cnvs: TCanvas;
begin
  Cnvs := CurrentCanvas.RealCanvas;
  if Cnvs = nil then
    Cnvs := Canvas;
  {$IFDEF MSWINDOWS}
  h := Fraction;
  if Fraction <> 0 then begin
    if IsRealDevice and (RealTextAngle <> 0) then
      Cnvs.Font.Assign(UnrotatedGraphicsFont);
    (* Determine size using rotation = 0, as extent at any other rotation is suspect *)
    h := h*Cnvs.TextWidth(str);
    GetTextMetrics(Cnvs.Handle, tm);
    if IsRealDevice and (RealTextAngle <> 0) then
      Cnvs.Font.Assign(GraphicsFont);
  end else
    GetTextMetrics(Cnvs.Handle, tm);
  v := tm.tmAscent;   (* Assume baseline - add tm.tmDescent for bottom *)
  {$ELSE}
  {$IFDEF LCL_}
  // if CurrentCanvas.RealCanvas <> nil then begin
  // gdk_text_extents(PGdiObject(GraphicsFont.Handle)^.GDIFontObject, PChar(str), Length(Str), @lbearing, @rBearing, @sz.cx, @sz.cy, @descent);
  //  sz := Cnvs.TextExtent(str)
  // end else
  sz := Cnvs.TextExtent(str);
  h := Fraction * sz.cx;
  v := sz.cy;
  {$ENDIF}
  {$IFDEF CLX_}
  sz := Cnvs.TextExtent(str);
  h := Fraction * sz.cx;
  v := sz.cy;
  {$ENDIF}
  {$ENDIF}
  { TODO: Check following for LINUX }
  if land_sea then begin
    xf := v/YPerX*s_TextAngle + h*c_TextAngle;
    yf := h*YPerX*s_TextAngle - v*c_TextAngle;
  end else begin
    xf := v*s_TextAngle + h/YPerX*c_TextAngle;
    yf := h*s_TextAngle - v*YPerX*c_TextAngle;
  end;
  if Abs(xf) >= MaxPlot then
    Result.x := MaxPlot
  else
    Result.x := Round(xf);
  if Abs(yf) >= MaxPlot then
    Result.y := MaxPlot
  else
    Result.y := Round(yf);
end;

function TGraphB.TextExtent(str:string):TPoint;
{$IFDEF NONWINLCL}
var
  sz: TSize;
{$ENDIF}
begin
  Result := TextExtentF(str, 1 {$IFDEF NONWINLCL}, sz{$ENDIF});
end;

procedure TGraphB.XorTextI(x, y: Integer; str: string);
var
  Extent: TPoint;
  Bmp: TBitmap;
  CopyMode: TCopyMode;
begin
  Bmp := nil;
  CopyMode := cmSrcCopy;
  if not IsRealDevice or (str = '') then
    TextI(x, y, str)
  else
    try
      start_plotting;
      CopyMode := Canvas.CopyMode;
      LogToDev(x, y);
      Extent := TextExtent(str);
      if (Extent.X <> 0) and (Extent.Y <> 0) then begin
        Bmp := TBitmap.Create;
        with Bmp do begin
          {$IFNDEF UNIX}
          Monochrome := True; { Seem to be problems creating monochrome bitmaps? }
          {$ENDIF}
          Height := Abs(Extent.Y);
          Width := Abs(Extent.X);
          with Canvas do Begin
            Brush.Color := clBlack;
            FillRect(Rect(0, 0, Height, Width));
            Font := CurrentCanvas.RealCanvas.Font;
            Font.Color := clWhite;
            TextOut(0, 0, str);
          end;
        end;
        Canvas.CopyMode := cmSrcInvert;
        Canvas.Draw(x, y-Bmp.Height, Bmp);
      end;
    finally
      Canvas.CopyMode := CopyMode;
      Bmp.Free;
      stop_plotting;
    end;
end;

procedure TGraphB.TextI(x:Integer; y:Integer; str:string);
var
{$IFDEF MSWINDOWS}
  CanPlot:Boolean;
  h: Single;
  xo, yo: Integer;
  Extent: TPoint;
{$ELSE}
{$IFDEF LCL_}
  h: Single;
  Extent: TPoint;
  sz: TSize;
  i: Integer;
  lbearing, rbearing, descent: LongInt;
  buf: string[2];
  //ws: WideString;
  //ps: PChar;
  Cinc: Integer;
{$ENDIF}
{$ENDIF}
{$IFDEF CLX_}
  Flags: Integer;
  FontMetrics: QFontMetricsH;
  Descent, x0, y0: Integer;
{$ENDIF}
begin
  start_plotting;
  CurrentCanvas.Brush.Style := bsClear;
  if {$IFDEF LCL_} not IsPrinter and {$ENDIF} (CurrentCanvas.RealCanvas <> nil) then begin
    {$IFDEF MSWINDOWS}
    CanPlot := True;
    LogToDev(x, y);
    if IsRealDevice and (RealTextAngle <> 0) then begin
      case text_flags and (TA_LEFT or TA_RIGHT or TA_CENTER) of
        TA_RIGHT:  h := 1;
        TA_CENTER: h := 0.5;
        TA_LEFT:   h := 0;
      else
        h := 0;
      end;
      Extent := TextExtentF(str, h);
      CanPlot := (Extent.x <> MaxPlot) and (Extent.y <> MaxPlot);
      if CanPlot then begin
        xo := x;
        Dec(x, Extent.x);
        yo := y;
        Inc(y, Extent.y);
        {$IFNDEF VER80}
        { Keep graphics calls to 16 bits, as Win95/8 can give overflow
          problems otherwise }
        CanPlot := (Abs(x) < MaxPlot) and (Abs(x) < MaxPlot)
                   and (Abs(xo) < MaxPlot) and (Abs(yo) < MaxPlot);
        {$ELSE}
        { This odd sequence checks for 16 bit overflow }
        if (x xor xo) < 0 then CanPlot := False;
        if (y xor yo) < 0 then CanPlot := False;
        {$ENDIF}
      end;
    end;
    if CanPlot then
      if land_sea then
        CurrentCanvas.RealCanvas.TextOut(x, y, str)
      else
        CurrentCanvas.RealCanvas.TextOut(y, x, str);
    {$ELSE}
    {$IFDEF LCL_}
    LogToDev(x, y);
    case FJustify of
        taLeftJustify:  h := 0;
        taRightJustify: h := 1;
        taCenter:       h := 0.5;
    end;
    //if TDeviceContext(CurrentCanvas.RealCanvas.Handle).DCTextMetric.IsDoubleByteChar then begin
      //ws := Str;
      //ps := PChar(ws);
      //Cinc := 2;
    //end else begin
      //ps := PChar(Str);
      //Cinc := 1;
    //end;
    if RealTextAngle = 0 then begin
      Extent := TextExtentF(str, h, sz);
      x := x - Extent.x;
      //with TDeviceContext(CurrentCanvas.RealCanvas.Handle) do
        //if land_sea then
          //gdk_draw_text(Drawable, PGdiObject(GraphicsFont.Handle)^.GDIFontObject, GC, x, y, ps, Length(Str))
        //else
          //gdk_draw_text(Drawable, PGdiObject(GraphicsFont.Handle)^.GDIFontObject, GC, y, x, ps, Length(Str));
      if land_sea then
        ExtUTF8Out(CurrentCanvas.RealCanvas.Handle, x, y-sz.cy, 0, nil, PChar(Str), Length(Str), nil)
      else
        ExtUTF8Out(CurrentCanvas.RealCanvas.Handle, y, x-sz.cy, 0, nil, PChar(Str), Length(Str), nil)
    end else begin
      {$IFNDEF GTK2}
      { Rotated text has to be done one character at a time under GTK1 at least }
      { TODO: Angles other than 0 and 90 need work (180 is particularly odd) }
      CurrentCanvas.RealCanvas.Font.Assign(UnrotatedGraphicsFont);
      if h > 0 then begin
        // gdk_text_extents(PGdiObject(UnrotatedGraphicsFont.Handle)^.GDIFontObject, PChar(Str), Length(Str), @lbearing, @rBearing, @sz.cx, @sz.cy, @descent);
        // WriteLn(RealTextAngle, ' ', lbearing, ' ', rBearing, ' ', sz.cx, ' ', sz.cy, ' ', descent);
        GetTextExtentPoint(CurrentCanvas.RealCanvas.Handle, PChar(Str), Length(Str), sz);
        x := x - Round(h*sz.cx*c_TextAngle);
        y := y + Round(h*sz.cx*s_TextAngle);
      end;
      CurrentCanvas.RealCanvas.Font.Assign(GraphicsFont);
      buf := #0#0;
      if TDeviceContext(CurrentCanvas.RealCanvas.Handle).DCTextMetric.IsDoubleByteChar then
        Cinc := 2
      else
        Cinc := 1;
      for i := 1 to Length(str) do begin
        buf[Cinc] := Str[i];
        gdk_text_extents(PGdiObject(UnrotatedGraphicsFont.Handle)^.GDIFontObject, @buf[1], Cinc, @lbearing, @rBearing, @sz.cx, @sz.cy, @descent);
        with TDeviceContext(CurrentCanvas.RealCanvas.Handle) do
          if land_sea then
            gdk_draw_text(Drawable, PGdiObject(GraphicsFont.Handle)^.GDIFontObject, GC, x { - Round(sz.cy*s_TextAngle)}, y {- Round(sz.cx*s_TextAngle)}, @buf[1], Cinc)
          else
            gdk_draw_text(Drawable, PGdiObject(GraphicsFont.Handle)^.GDIFontObject, GC, y, x, @buf[1], Cinc);
        x := x + Round(sz.cx*c_TextAngle);
        y := y - Round(sz.cx*s_TextAngle);
      end;
      {$ELSE}
      { TODO: try under gtk2; may just work without special handling using Pango }
      if land_sea then
        ExtUTF8Out(CurrentCanvas.RealCanvas.Handle, x, y-sz.cy, 0, nil, PChar(Str), Length(Str), nil)
      else
        ExtUTF8Out(CurrentCanvas.RealCanvas.Handle, y, x-sz.cy, 0, nil, PChar(Str), Length(Str), nil)
      {$ENDIF}
    end;
    {$ENDIF}
    {$IFDEF CLX_}
    Descent := 0;
    FontMetrics := nil;
    try
      FontMetrics := QFontMetrics_create(CurrentCanvas.RealCanvas.Font.Handle);
      Descent := QFontMetrics_descent(FontMetrics);
    finally
      QFontMetrics_destroy(FontMetrics);
    end;
    Dec(y, Descent+2);  // TODO: Is 2 good in general?
    // TODO: Update font
    LogToDev(x,y);
    Flags := Integer(AlignmentFlags_AlignBottom) or Integer(AlignmentFlags_DontClip);
    case FJustify of
        taLeftJustify:  Flags := Flags or Integer(AlignmentFlags_AlignLeft);
        taRightJustify: Flags := Flags or Integer(AlignmentFlags_AlignRight);
        taCenter:       Flags := Flags or Integer(AlignmentFlags_AlignHCenter);
    end;
    if RealTextAngle <> 0 then begin
      x0 := x; y0 := y;
      x := Round(x0*c_TextAngle - y0*s_TextAngle);
      y := Round(y0*c_TextAngle + x0*s_TextAngle);
      QPainter_rotate(CurrentCanvas.RealCanvas.Handle, -RealTextAngle/10);
    end;
    CurrentCanvas.RealCanvas.TextRect(Rect(x, y, x, y), x, y, str, Flags);
    if RealTextAngle <> 0 then
      QPainter_rotate(CurrentCanvas.RealCanvas.Handle, RealTextAngle/10);
    {$ENDIF}
    {$ENDIF}
  end else begin
    LogToDev(x,y);
    CurrentCanvas.RotatedTextOut(x, y, str, RealTextAngle, FJustify);
  end;
  PenInvalid;
  stop_plotting;
end;

procedure TGraphB.Mark(x:Integer; y:Integer; number:Integer);
begin
  number := number mod MaxPens;
  if number < 0 then Inc(number, MaxPens);
  MarkT(x, y, MarkStyles[number]);
end;

function TGraphB.GetTickSize:TPoint;
begin
  if DCType = dcNone then InitFont;
  HintDisplacement(Round(Abs(FontSize) / 3), Result.y, Result.x, False);
end;

procedure TGraphB.SetCurrentMarkSize(Size: Single);
var
  f: Double;
begin
  if CurrentMarkSize <> Size then begin
    CurrentMarkSize := Size;
    if Font.Size = 0 then
      f := Size
    else
      f := Size / Font.Size * FontSize;
    HintDisplacement(Round(Abs(f)), mp.v, mp.h, False);
  end;
end;

procedure TGraphB.MarkT(x:Integer; y:Integer; AStyle: TMarkStyle);
begin
  start_plotting;
  MarkLineStyle.Colour := AStyle.Colour;
  mp.DotAtCentre := AStyle.DotAtCentre;
  NeedPen(MarkLineStyle);
  SetCurrentMarkSize(AStyle.Size);
  Gmarks.Mark(CurrentCanvas, x, y, AStyle.MarkType, mp);
  stop_plotting;
end;

procedure TGraphB.MarkE(x:Integer; y:Integer; AType: TMarkType);
begin
  start_plotting;
  NeedLinePen;
  SetCurrentMarkSize(ErrorBarSize);
  Gmarks.Mark(CurrentCanvas, x, y, AType, mp);
  stop_plotting;
end;

procedure TGraphB.Angle(ang:Integer);
begin
  FTextAngle := ang;
  WIN_change_font;
end;

procedure TGraphB.Justify(mode:TAlignment);
{$IFDEF MSWINDOWS}
var
  textm:Word;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  case mode of
      taLeftJustify:  textm := TA_LEFT;
      taRightJustify: textm := TA_RIGHT;
      taCenter:       textm := TA_CENTER;
  else
    raise Exception.CreateFmt('TGraphB.Justify (%d)', [Ord(mode)]);
  end;
  if (textm <> text_flags) and (CurrentCanvas <> nil) and not (IsRealDevice and (RealTextAngle <> 0))
     and (CurrentCanvas.RealCanvas <> nil) then
    SetTextAlign(CurrentCanvas.RealCanvas.Handle, textm or TA_BASELINE or TA_NOUPDATECP);
  text_flags := textm;
  {$ENDIF}
  { TODO: LINUX: May need to flag change font? }
  FJustify := mode;
end;

procedure  TGraphB.Clear;
begin
  if CurrentCanvas = nil then begin
    {$IFDEF METAFILES}
    DiscardMF(hMF);
    {$ENDIF}
    Invalidate;
  end;
end;

procedure TGraphB.SetMetafileScaling;
var
  w, h: Integer;
begin
  w := Round(Xmm*100);
  h := Round(Ymm*100);
  if not (grIsotropic in GOptions) then
    { Adjust pixel ratio to follow Window size }
    if w/PaintWidth > h/PaintHeight then
      w := Round(h/PaintHeight*PaintWidth)
    else
      h := Round(w/PaintWidth*PaintHeight);
  SetScaling(nil, 0, 0, w, h);
  {$IFDEF MSWINDOWS}
  MetaFileWidth := FXPixels;
  MetaFileHeight := FYPixels;
  {$ENDIF}
end;

procedure TGraphB.SetScaling(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint);
var
  {$IFDEF CLX_}
  PaintDevice: QPaintDeviceH;
  PaintDeviceMetrics: QPaintDeviceMetricsH;
  {$ENDIF}
  t: Longint;
begin
  FAvailableHeight := AHeight;
  if (ACanvas <> nil) and (IsRealDevice or (ACanvas = Canvas)) then begin
    {$IFNDEF CLX_}
    if (ACanvas <> Canvas) or HandleAllocated then begin
      DPIX := GetDeviceCaps(ACanvas.Handle, LOGPIXELSX);
      DPIY := GetDeviceCaps(ACanvas.Handle, LOGPIXELSY);
    end else
    {$ELSE}
    PaintDevice := QPainter_device(ACanvas.Handle);
    if PaintDevice <> nil then begin
      PaintDeviceMetrics := QPaintDeviceMetrics_create(PaintDevice);
      DPIX := QPaintDeviceMetrics_logicalDpiX(PaintDeviceMetrics);
      DPIY := QPaintDeviceMetrics_logicalDpiY(PaintDeviceMetrics);
      QPaintDeviceMetrics_destroy(PaintDeviceMetrics);
    { Interestingly, on my iiyama monitor (?by dirctly querying monitor?), the two
      values returned as above, and by xdpyinfo are slightly different (81 + 86).
      This leads to horrible looking marks, so assume the two values are different
      only if the difference is significant }
      if (DPIX - DPIY) / (DPIX + DPIY) < 0.3 then begin
        DPIY := (DPIX+DPIY) div 2;
        DPIX := DPIY;
      end;
    end else
    {$ENDIF}
    begin
      { Assume on screen }
      DPIX := Screen.PixelsPerInch;
      DPIY := Screen.PixelsPerInch;
    end;
    YPerX := DPIY / DPIX;
    if not land_sea then begin
      YPerX := 1 / YPerX;
      t := DPIX;
      DPIX := DPIY;
      DPIY := t;
    end
  end else begin
    YPerX := 1;
    { Use standard for enhanced metafiles - 0.01 mm }
    DPIX := 2540;
    DPIY := 2540;
  end;
  if land_sea then begin
    FXPixels := AWidth;
    FYPixels := AHeight;
  end else begin
    FXPixels := AHeight;
    FYPixels := AWidth;
  end;
  if grIsotropic in GOptions then begin
    { Shrink one axis to maintain Xmm/Ymm ratio }
    if FXPixels * YPerX / FYPixels > Xmm/Ymm then
      FXPixels := Round(FYPixels * Xmm/Ymm / YPerX)
    else
      FYPixels := Round(FXPixels * Ymm/Xmm * YPerX);
  end;
  FXOffset := X;
  FYOffset := Y;
end;

procedure TGraphB.ShowOnCanvas(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint);
  {$IFDEF MSWINDOWS}
var
  SavedDC: Integer;
  ViewOrg: TPoint;
  {$ENDIF}
  procedure ResetDC;
  begin
    InReplot := False;
    if Assigned(ACanvas) then
      {$IFDEF MSWINDOWS}
      RestoreDC(ACanvas.Handle, SavedDC);
      {$ENDIF}
      {$IFDEF CLX_}
      QPainter_restore(ACanvas.Handle);
      {$ENDIF}
    if CurrentCanvas <> nil then
      cancel_graphics_DC;
    DCType := dcNone;
    LoadMP;
  end;
begin
  if Assigned(ACanvas) then
    {$IFDEF MSWINDOWS}
    SavedDC := SaveDC(ACanvas.Handle);
    {$ENDIF}
    {$IFDEF CLX_}
    QPainter_save(ACanvas.Handle); // TODO: ?Required
    {$ENDIF}
  if Assigned(FOnReplot) then begin
    if Assigned(ACanvas) then
      SetRealCanvas(ACanvas);
    try
      if Assigned(ACanvas) then begin
        ACanvas.Refresh;
        ACanvas.Pen.Width := 1; { Refresh sets pen width to zero, which can not normally
                                  be chosen under CLX, giving consistency problems in plots }
        ACanvas.Brush.Style := bsSolid;
      end;
      SetScaling(ACanvas, X, Y, AWidth, AHeight);
      InitGraphics;
      InReplot := True;
      if not PlotFailed then FOnReplot(Self, Self);
      end_draw;
      InReplot := False;
    except
      ResetDC;
      FPlotFailed := True;
      raise;
    end;
  end
  {$IFDEF METAFILES}
  else begin
    SetScaling(ACanvas, X, Y, AWidth, AHeight);
    {$IFDEF MSWINDOWS}
    {$IFDEF VER80}
    if hMF <> nil then
      ACanvas.StretchDraw(Bounds(X, Y, AWidth, AHeight), hMF);
    {$ELSE}
    if hMF <> 0 then begin
      { Stretch draw logic from Delphi 1 }
      SetMapMode(ACanvas.Handle, MM_ANISOTROPIC);
      SetWindowExtEx(ACanvas.Handle, MetaFileWidth, MetaFileHeight, nil);
      SetViewportExtEx(ACanvas.Handle, AWidth, AHeight, nil);
      GetViewPortOrgEx(ACanvas.Handle, ViewOrg);
      SetViewportOrgEx(ACanvas.Handle, ViewOrg.X + X, ViewOrg.Y + Y, nil);
      PlayMetafile(ACanvas.Handle, hMF);
    end;
    {$ENDIF}
    {$ELSE}
    if hMF <> nil then
      ACanvas.StretchDraw(Bounds(X, Y, AWidth, AHeight), hMF)
    {$ENDIF}
  end
  {$ENDIF}
  ;
  ResetDC;
end;

procedure TGraphB.OptionChanged;
begin
  if CurrentCanvas = nil then begin
    LoadMP;
    if not Assigned(FOnReplot) then
      Clear
    else
      Invalidate;
  end;
end;

procedure TGraphB.PlotChanged;
begin
  if not Assigned(FOnReplot) then
    Clear
  else
    Invalidate;
  FPlotFailed := False;
end;

procedure TGraphB.MakeScreenImage(ScreenCanvas: TCanvas);
begin
  if (PaintWidth > 4) and (PaintHeight > 4) then begin
    ShowOnCanvas(ScreenCanvas,
                 PaintRect.Left - HorzScrollBar.Position,
                 PaintRect.Top - VertScrollBar.Position,
                 Round(PaintWidth*XSCALE),
                 Round(PaintHeight*YSCALE));
  end;
end;

procedure TGraphB.SetScreenScaling;
begin
  SetScaling(Canvas, PaintRect.Left - HorzScrollBar.Position,
             PaintRect.Top - VertScrollBar.Position,
             Round(PaintWidth*XSCALE), Round(PaintHeight*YSCALE));
end;

procedure TGraphB.SetLastMouse(X, Y: Integer);
begin
  FLastMouse.x := X;
  FLastMouse.y := Y;
  DevToLog(FLastMouse.x, FLastMouse.y);
end;

procedure TGraphB.Paint;
{$IFDEF UNIX}
var
  Buffered: Boolean;
  SaveHandle: {$IFDEF CLX_}QPainterH{$ENDIF} {$IFDEF LCL_} HDC {$ENDIF} ;
{$ENDIF}
begin
  DCType := dcScreen;
  {$IFDEF UNIX}
  Buffered := False;
  if DoubleBuffered and not (csDesigning in ComponentState) then begin
    if FScreenBuffer <> nil then Exit; { Construction in progress - ignore paint }
    FScreenBuffer := TBitmap.Create;
    FScreenBuffer.Height := Height;
    FScreenBuffer.Width := Width;
    DCType := dcScreenBuffer;
    {$IFDEF CLX_}
    FScreenBuffer.Canvas.Start;
    {$ENDIF}
    SaveHandle := Canvas.Handle;
    Canvas.Handle := FScreenBuffer.Canvas.Handle;
    Canvas.Refresh;
    Buffered := True;
  end;
  try
  {$ENDIF}
  Canvas.Brush.Color := Color;
  inherited Paint;
  {$IFDEF MSWINDOWS}
  Canvas.FillRect(Canvas.ClipRect);
  IntersectClipRect(Canvas.Handle, PaintRect.Left, PaintRect.Top, PaintRect.Right, PaintRect.Bottom);
  {$ENDIF}
  MakeScreenImage(Canvas);
  DrawSelectedArea;
  if Assigned(FOnPaint) then
   FOnPaint(Self);
  {$IFDEF UNIX}
  finally
    if Buffered then begin
      Canvas.Handle := SaveHandle;
      Canvas.Draw(0, 0, FScreenBuffer);
      {$IFDEF CLX_}
      FScreenBuffer.Canvas.Stop;
      {$ENDIF}
      FreeAndNil(FScreenBuffer);
    end;
  end;
  {$ENDIF}
end;

procedure TGraphB.ShowPrintPreview(Sender:TObject; ACanvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                                   APage: Integer; OnPrinter:Boolean);
begin
  if OnPrinter then
    DCType := dcPrinter
  else
    DCType := dcPreview;
  FCurrentCanvas := nil;
  ShowOnCanvas(ACanvas.RealCanvas(), X, Y, AWidth, AHeight)
end;

procedure TGraphB.MakePrintPreview;
begin
  if not Assigned(FPrintPreview) then begin
    FPrintPreview := TGraphPrintPreview.Create(Application);
    FPrintPreview.Graph := Self;
    FPrintPreview.Orientation := Printer.Orientation;
    FPrintPreview.SetMargins(Xmm, Ymm, FPrintPreview.XMargin, FPrintPreview.YMargin);
  end;
end;

procedure TGraphB.PrintPreview;
begin
  MakePrintPreview;
  FPrintPreview.Show;
end;

procedure TGraphB.Print;
begin
  MakePrintPreview;
  FPrintPreview.Print;
end;

procedure TGraphB.PrinterSetup;
begin
  MakePrintPreview;
  FPrintPreview.PrinterSetup;
end;

destructor TGraphB.Destroy;
var
  i: Integer;
begin
  SetCurrentCanvas(nil);
  {$IFDEF METAFILES}
  DiscardMF(hMF);
  {$ENDIF}
  FreeAndNil(FScreenBuffer);
  for i := 0 to MaxPens do begin
    FLineStyles[i].Free;
    FLineStyles[i] := nil;
    FDefaultLineStyles[i].Free;
    FDefaultLineStyles[i] := nil;
    if i < MaxPens then begin
      FMarkStyles[i].Free;
      FMarkStyles[i] := nil;
      FDefaultMarkStyles[i].Free;
      FDefaultMarkStyles[i] := nil;
    end;
  end;
  CLineStyle.Free;
  MarkLineStyle.Free;
  {$IFNDEF CLX_}
  FreeAndNil(UnrotatedGraphicsFont);
  FreeAndNil(GraphicsFont);
  {$ENDIF}
  inherited Destroy;
end;

procedure TGraphB.SetOnReplot(e:TReplotEvent);
begin
  if @FOnReplot <> @e then begin
    FOnReplot := e;
    PlotChanged;
  end;
end;

function TGraphB.CanCopyToClipboard: Boolean;
begin
  if Assigned(FOnReplot) then
    Result := not PlotFailed
  {$IFDEF METAFILES}
  else
    Result := hMF <> NullMF
  {$ENDIF}
    ;
end;

function TGraphB.CanPrint: Boolean;
begin
  Result := CanCopyToClipboard;
end;

procedure TGraphB.SetAreaSelected(b: Boolean);
begin
  if b <> AreaSelected then
    if b then
      SetSelectedArea(FSelectedArea)
    else begin
      DrawSelectedArea;
      FAreaSelected := False;
    end;
end;

procedure TGraphB.SetSelectedArea(r:TRect);
begin
  DrawSelectedArea;
  FSelectedArea := r;
  FAreaSelected := ( abs(FSelectedArea.Right - FSelectedArea.Left) > 0 )
                  or (abs(FSelectedArea.Bottom - FSelectedArea.Top) > 0 );
  DrawSelectedArea;
end;

function TGraphB.GetLogicalSelectedArea: TRect;
begin
  Result := SelectedArea;
  DevToLog(Result.Left, Result.Top);
  DevToLog(Result.Right, Result.Bottom);
end;

procedure TGraphB.SetLogicalSelectedArea(ARect: TRect);
begin
  LogToDev(ARect.Left, ARect.Top);
  LogToDev(ARect.Right, ARect.Bottom);
  SelectedArea := ARect;
end;

procedure DrawFocusRectHelper(Canvas: TCanvas; r: TRect);
var
  t: Longint;
begin
  if r.Left > r.Right then begin
    t := r.Left;
    r.Left := r.Right;
    r.Right := t;
  end;
  if r.Top > r.Bottom then begin
    t := r.Top;
    r.Top := r.Bottom;
    r.Bottom := t;
  end;
  Canvas.Brush.Style := bsSolid; { These two are required under MSWINDOWS, else no focus rect appears }
  Canvas.Brush.Color := clWhite;
  {$IFNDEF LCL_}
  Canvas.DrawFocusRect(r);
  {$ELSE}
    {$IFDEF MSWINDOWS}
      DrawFocusRect(Canvas.Handle, r);
    {$ELSE}
      Canvas.Pen.Style := Graphics.psDot;
      gdk_gc_set_function(TDeviceContext(Canvas.Handle).GC, GDK_INVERT);
      Canvas.Frame(r);
      gdk_gc_set_function(TDeviceContext(Canvas.Handle).GC, GDK_COPY);
      Canvas.Pen.Style := Graphics.psSolid;
    {$ENDIF}
  {$ENDIF}
end;

procedure TGraphB.DrawSelectedArea;
var
  r:TRect;
begin
  if not (grNoAreaSelection in GOptions) and AreaSelected then begin
    r := FSelectedArea;
    OffsetRect(r, PaintRect.Left, PaintRect.Top);
    DrawFocusRectHelper(Canvas, r);
  end else
    FAreaSelected := False;
end;

procedure TGraphB.CancelSelection;
begin
  DrawSelectedArea;
  Selecting := False;
  FAreaSelected := False;
end;

procedure TGraphB.Invalidate;
begin
  FAreaSelected := False;
  if HandleAllocated then
    inherited Invalidate;
  if Assigned(FPrOnInvalidate) then FPrOnInvalidate(Self);
  if Assigned(FOnInvalidate) then FOnInvalidate(Self);
end;

procedure TGraphB.UpdateSelection( X, Y: Integer; NewPoint:Boolean);
begin
  DrawSelectedArea;
  if NewPoint and Selecting then begin
    FSelectedArea.Right := X - PaintRect.Left;
    FSelectedArea.Bottom := Y - PaintRect.Top;
    { Make sure rectangle shown includes current cursor point }
    if FSelectedArea.Right > FSelectedArea.Left then
      Inc(FSelectedArea.Right);
    if FSelectedArea.Bottom > FSelectedArea.Top then
      Inc(FSelectedArea.Bottom);
    FAreaSelected := ( abs(FSelectedArea.Right - FSelectedArea.Left) > SmallX )
                    or (abs(FSelectedArea.Bottom - FSelectedArea.Top) > SmallY );
    DrawSelectedArea;
  end;
end;

procedure TGraphB.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  p: TMouseEvent;
begin
  if (Button = mbLeft) or (Button = mbRight) then begin
    CancelSelection;
    Selecting := True;
    FSelectedArea.Left := X - PaintRect.Left;
    FSelectedArea.Top := Y - PaintRect.Top;
  end;
  p := OnMouseDown;
  if Assigned(p) then begin
    SetLastMouse(X, Y);
    OnMouseDown(Self, Button, Shift, LastMouse.x, LastMouse.y);
  end;
end;

procedure TGraphB.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  p: TMouseMoveEvent;
begin
  if Selecting then
    UpdateSelection(X, Y, True);
  p := OnMouseMove;
  if Assigned(p) then begin
    SetLastMouse(X, Y);
    OnMouseMove(Self, Shift, LastMouse.x, LastMouse.y);
  end;
end;

procedure TGraphB.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  p: TMouseEvent;
begin
  if ((Button = mbLeft) or (Button = mbRight)) and Selecting then begin
    UpdateSelection(X, Y, True);
    Selecting := False;
  end;
  p := OnMouseUp;
  if Assigned(p) then begin
    SetLastMouse(X, Y);
    OnMouseUp(Self, Button, Shift, LastMouse.x, LastMouse.y);
  end;
end;

procedure TGraphB.UpdateScrollBars;
begin
  if not( Assigned(VertScrollBar) and Assigned(HorzScrollBar) ) then
    Exit;
  HorzScrollBar.Range := Round(XSCALE*PaintWidth) - PaintWidth;
  HorzScrollBar.PageSize := PaintWidth div 2;
  VertScrollBar.Range := Round(YSCALE*PaintHeight) - PaintHeight;
  VertScrollBar.PageSize := PaintHeight div 2;
  inherited UpdateScrollBars;
end;

{ Scroll Bar Positions as fraction of plot space }
function TGraphB.HScrollFraction: Double;
begin
  if HorzScrollBar.Range > 0 then
    Result := HorzScrollBar.Position/(XSCALE*PaintWidth)
  else
    Result := 0;
end;

function TGraphB.VScrollFraction: Double;
begin
  if VertScrollBar.Range > 0 then
    Result := VertScrollBar.Position/(YSCALE*PaintHeight)
  else
    Result := 0;
end;

procedure TGraphB.ChangeScrollPos(Hpos, Vpos: double);
begin
  HorzScrollBar.Position := Round(Hpos*XSCALE*PaintWidth);
  VertScrollBar.Position := Round(Vpos*YSCALE*PaintHeight);
end;

procedure TGraphB.ChangeZoom(newXSCALE, newYSCALE, newHscroll, newVscroll: Double);
begin
  if (newXSCALE*PaintWidth > 32700) or (newYSCALE*PaintHeight > 32700) then
    ShowMessage('Magnification too high')
  else begin
    oldXSCALE := XSCALE;
    oldYSCALE := YSCALE;
    oldHPos := HScrollFraction;
    oldVPos := VScrollFraction;

    XSCALE := newXSCALE;
    YSCALE := newYSCALE;
    UpdateScrollBars;
    ChangeScrollPos(newHscroll, newVscroll);
  end;
  {$IFDEF METAFILES}
  if hMF = NullMF then
    PlotChanged
  else
  {$ENDIF}
    Invalidate;
end;

procedure TGraphB.ResetZoom;
begin
  ChangeZoom(1, 1, 0, 0);
end;

procedure TGraphB.PreviousZoom;
begin
  ChangeZoom(oldXSCALE, oldYSCALE, oldHPos, oldVPos);
end;

procedure TGraphB.ZoomOut;
var
  magx,magy, h, v: Double;
begin
  if AreaSelected then begin
    magx := Abs(SelectedArea.Left-SelectedArea.Right)/PaintWidth;
    magy := Abs(SelectedArea.Top-SelectedArea.Bottom)/PaintHeight;
    if grIsotropic in GOptions then
      if magx < magy then
        magx := magy
      else
        magy := magx;
    h := HScrollFraction - Min(SelectedArea.left,SelectedArea.right)/(XSCALE*PaintWidth*magx);
    v := VScrollFraction - Min(SelectedArea.top,SelectedArea.bottom)/(YSCALE*PaintHeight*magy);
    ChangeZoom(magx*XSCALE, magy*YSCALE, h, v);
  end else
    ChangeZoom(XSCALE/2, YSCALE/2, HScrollFraction - 1/(2*XSCALE), VScrollFraction - 1/(2*YSCALE));
end;

procedure TGraphB.ZoomSelected(Anisotropic:Boolean);
var
  Magx, Magy, AMagx, AMagy, h, v:Double;
  function Adjust(Mag: Double): Double;
  begin
    { Adjust mag to correct for problems }
    if Mag > 2 then
      Result := 0.9
    else if Mag > 1.05 then
      Result := 0.95
    else
      Result := 1;
    Result := Mag * Result;
  end;
begin
  if AreaSelected then begin
    magx := PaintWidth/Abs(SelectedArea.Left-SelectedArea.Right);
    AMagx := Adjust(magx);
    magy := PaintHeight/Abs(SelectedArea.Top-SelectedArea.Bottom);
    AMagy := Adjust(magy);
    if not Anisotropic then
      if AMagx > AMagy then
        AMagx := AMagy
      else
        AMagy := AMagx;
    h := HScrollFraction - (1-AMagx/Magx)/2/AMagx/XSCALE
         + Min(SelectedArea.left,SelectedArea.right)/(XSCALE*PaintWidth);
    v := VScrollFraction - (1-AMagy/Magy)/2/AMagy/YSCALE
         + Min(SelectedArea.top,SelectedArea.bottom)/(YSCALE*PaintHeight);
    ChangeZoom(AMagx*XSCALE, AMagy*YSCALE, h, v);
  end else
    ChangeZoom(2*XSCALE, 2*YSCALE, HScrollFraction + 1/(4*XSCALE), VScrollFraction + 1/(4*YSCALE));
end;

procedure TGraphB.ExportTo(AStream: TStream; CanvasClass: TStreamCanvasClass);
var
  CustomStreamCanvas: TCustomStreamCanvas;
begin
  if not Assigned(FOnReplot) or InReplot then
    raise Exception.Create('OnReplot must be set for ExportTo');
  CustomStreamCanvas := CanvasClass.Create(AStream, Xmm, Ymm);
  try
    SetScaling(nil, 0, 0, Round(Xmm/25.4*CustomStreamCanvas.DPI),
                          Round(Ymm/25.4*CustomStreamCanvas.DPI));
    DPIX := CustomStreamCanvas.DPI;
    DPIY := CustomStreamCanvas.DPI;
    SetCurrentCanvas(CustomStreamCanvas);
    { TODO : this stuff appears in several places }
    DCType := dcExport;
    InitGraphics;
    InReplot := True;
    FOnReplot(Self, Self);
    end_draw;
    cancel_graphics_DC;
  finally
    InReplot := False;
    SetCurrentCanvas(nil); { Will free CustomStreamCanvas }
  end;
end;

{$IFDEF METAFILES}
function TGraphB.GetPicture: {$IFDEF MSWINDOWS} {$IFDEF WINLCL} HMETAFILE {$ELSE} TMetafile {$ENDIF} {$ENDIF} {$IFDEF UNIX} TDrawing {$ENDIF};
{$IFDEF WIN32}
var
  nBytes: Cardinal;
  buffer: string;
{$ENDIF}
begin
  if Assigned(FOnReplot) then begin
    try
      StartMetafile;
      DCType := dcClip;
      InitGraphics;
      InReplot := True;
      FOnReplot(Self, Self);
      end_draw;
      InReplot := False;
      StopMetafile;
      cancel_graphics_DC;
    except
      InReplot := False;
      {$IFDEF MSWINDOWS}
      if (MetafileCanvas <> nil) and (MetafileCanvas.Handle <> 0) then
        CloseMetaFile(MetafileCanvas.Handle);
      FreeAndNil(MetafileCanvas);
      {$ENDIF}
      DiscardMF(hMF);
      cancel_graphics_DC;
      raise;
    end;
    {$IFDEF MSWINDOWS}
      {$IFDEF LCL}
        Result := hMF;
        hMF := 0; { Hand over ownership }
      {$ELSE}
        Result := TMetafile.Create;
        {$IFDEF VER80}
          Result.Handle := hMF;
        {$ELSE}
          nBytes := GetMetaFileBitsEx(hMF, 0, nil);
          SetLength(buffer, nBytes);
          GetMetaFileBitsEx(hMF, nBytes, PChar(buffer));
          Result.Handle := SetWinMetaFileBits(nBytes, PChar(buffer), HDC(nil), PMetafilePict(nil)^);
          DiscardMF(hMF);
        {$ENDIF}
        hMF := 0; { Hand over ownership }
      {$ENDIF}
    {$ENDIF}
    {$IFDEF UNIX}
      Result := hMF;
      hMF := nil; { Hand over ownership }
    {$ENDIF}
  end else begin
    {$IFDEF MSWINDOWS}
      {$IFDEF LCL}
        Result := CopyEnhMetafile(hMF, nil);
      {$ELSE}
        Result := TMetafile.Create;
        Result.Handle := CopyEnhMetafile(hMF, nil);
      {$ENDIF}
    {$ENDIF}
    {$IFDEF UNIX}
    Result := TDrawing.Create;
    Result.Assign(hMF);
    {$ENDIF}
  end;
end;
{$ENDIF}

procedure TGraphB.CopyToClipboard;
{$IFDEF MSWINDOWS}
var
  hGMem:THandle;
  AFormat: Word;
  lpMFP:^TMetaFilePict;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  if Assigned(FOnReplot) then
    try
      StartMetafile;
      DCType := dcClip;
      InitGraphics;
      InReplot := True;
      FOnReplot(Self, Self);
      end_draw;
      InReplot := False;
      {$IFDEF VER80}
      hMF.Width := MetafileWidth;
      hMF.Height := MetafileHeight;
      { Do not set hMF.Inch := DPIX; as this affects width and height }
      {$ELSE}
      hMF := CloseMetaFile(MetafileCanvas.Handle);
      {$ENDIF}
      FreeAndNil(MetafileCanvas);
      cancel_graphics_DC;
    except
      InReplot := False;
      if (MetafileCanvas <> nil) and (MetafileCanvas.Handle <> 0) then
        CloseMetaFile(MetafileCanvas.Handle);
      FreeAndNil(MetafileCanvas);
      DiscardMF(hMF);
      cancel_graphics_DC;
      raise;
    end;
  if hMF <> NullMF then begin
    hGMem := GlobalAlloc(GHND, SizeOf(TMetaFilePict));
    if hGMem = 0 then raise Exception.Create('No Handle');
    try
      lpMFP := GlobalLock(hGMem);
      try
        lpMFP^.mm := MM_ANISOTROPIC;
        { grIsotropic just determines whether the aspect ratio of the picture
          reflects the window size or the Xmm/Ymm settings }
        lpMFP^.xExt := MetafileWidth;
        lpMFP^.yExt := MetafileHeight;
        lpMFP^.hMF := CopyMetafile(hMF{$IFDEF VER80}.Handle{$ENDIF}, nil);
      finally
        GlobalUnlock(hGMem);
      end;
    except
      GlobalFree(hGMem);
      raise;
    end;
    AFormat := CF_METAFILEPICT;
    {$IFDEF LCL}
      { TODO: Clipboard.SetAsHandle }
    {$ELSE}
      Clipboard.SetAsHandle(AFormat, hGMem);
    {$ENDIF}
  end;
  {$ENDIF}
  {$IFNDEF METAFILES}
  // TODO:
  {$ENDIF}
  {$IFDEF UNIX}
  // TODO:
  {$ENDIF}
end;

procedure TGraphB.DoCopyToClipboard(Sender: TObject);
begin
  CopyToClipboard;
end;

procedure TGraphB.DoChooseFont(Sender: TObject);
var
  FontDialog: TFontDialog;
begin
  FontDialog := TFontDialog.Create(Self);
  try
    FontDialog.Font := Font;
    if FontDialog.Execute then
      Font := FontDialog.Font;
    OptionChanged;
  finally
    FontDialog.Free;
  end;
end;

constructor TGraphD.Create(AOwner:TComponent);
begin
  Xp2 := 1;
  Yp2 := 1;
  inherited Create(AOwner);
  CMarkStyle := TMarkStyle.Create;
  SetRecordCalls(True);
end;

destructor TGraphD.Destroy;
begin
  RStream.Free;
  {$IFNDEF DYNAMIC_ARRAYS}
  if FPolygonPoints <> nil then
    Dispose(FPolygonPoints);
  FPolygonPoints := nil;
  {$ENDIF}
  CMarkStyle.Free;
  inherited Destroy;
end;

procedure  TGraphD.DoGraphReplot(Sender:TObject; G:TGraphB);
begin
  Recording := false;
  DefaultScale;
  if Assigned(FOnDReplot) then
      FOnDReplot(Self, Self)
  else
      PlayBack(Rstream);
  { Need to save scaling factors for mouse event scaling to work }
  if IsScreen then begin
    XofSave := Xof;
    XsSave := Xs;
    XminSave := Xmin;
    YofSave := Yof;
    YsSave := Ys;
    YminSave := Ymin;
  end else begin
    Xof := XofSave;
    Xs := XsSave;
    FXmin := XminSave;
    Yof := YofSave;
    Ys := YsSave;
    FYmin := YminSave;
  end
end;

procedure TGraphD.ShowOnCanvas(ACanvas: TCanvas; X, Y, AWidth, AHeight:Longint);
var
  RecPos: Longint;
begin
  if Recording then begin
    RecPos := RStream.Position;
    try
      Recording := False;
      PutType(tyEnd);
      inherited ShowOnCanvas(ACanvas, X, Y, AWidth, AHeight);
    finally
      Recording := True;
      RStream.Position := RecPos;
    end
  end else
    inherited ShowOnCanvas(ACanvas, X, Y, AWidth, AHeight);
end;

procedure TGraphD.PlotChanged;
begin
  if RStream <> nil then begin
    RStream.Position := 0;
    PutType(tyEnd);
  end;
  inherited PlotChanged;
end;

procedure TGraphD.PreviousZoom;
begin
  ZoomArea := PreviousZoomArea;
end;

procedure TGraphD.ChangeScrollPos(Hpos, Vpos: double);
begin
  with FZoomArea do begin
    Left := (Hpos-Xp1)/(Xp2-Xp1);
    Right := (Hpos+1/XSCALE-Xp1)/(Xp2-Xp1);
    Top := (Vpos-Yp1)/(Yp2-Yp1);
    Bottom := (Vpos+1/YSCALE-Yp1)/(Yp2-Yp1);
  end;
  inherited ChangeScrollPos(Hpos, Vpos);
end;

procedure TGraphD.ChangeZoom(newXSCALE, newYSCALE, newHscroll, newVscroll: Double);
var
  RStreamSave: TStream;
  Xp1p, Xp2p, Yp1p, Yp2p: Double; { New pspace }
begin
  { Do not clear recorded calls on change of zoom }
  RStreamSave := RStream;
  RStream := nil;
  try
    PreviousZoomArea := ZoomArea;
    { Assume pspace stuff stays constant pixel size, so scales inversely with XSCALE/YSCALE }
    Xp1p := Xp1 * XSCALE/newXSCALE;
    Xp2p := 1 - (1-Xp2) * XSCALE/newXSCALE;
    Yp1p := Yp1 * YSCALE/newYSCALE;
    Yp2p := 1 - (1-Yp2) * YSCALE/newYSCALE;
    newHscroll := (newHscroll-Xp1)*(Xp2p-Xp1p)/(Xp2-Xp1) + Xp1p;
    newVscroll := (newVscroll-Yp1)*(Yp2p-Yp1p)/(Yp2-Yp1) + Yp1p;
    inherited ChangeZoom(newXSCALE, newYSCALE, newHscroll, newVscroll);
    { Assume pspace stuff stays constant pixel size, so scales inversely with XSCALE/YSCALE }
    Xp1 := Xp1p;
    Xp2 := Xp2p;
    Yp1 := Yp1p;
    Yp2 := Yp2p;
  finally
    RStream := RStreamSave;
  end;
end;

procedure TGraphD.SetZoomArea(const Value: TDRect);
var
  Previous: TDRect;
begin
  Previous := ZoomArea;
  with Value do
    if (Right > Left) and (Bottom > Top) then
      ChangeZoom(1/(Right-Left)/(Xp2-Xp1), 1/(Bottom-Top)/(Yp2-Yp1),
                 Left*(Xp2-Xp1) + Xp1, Top*(Yp2-Yp1) + Yp1)
    else
      ResetZoom;
  PreviousZoomArea := Previous;
end;

procedure  TGraphD.SetOnReplot(e:TReplotDEvent);
begin
  FOnDReplot := e;
  SetRecordCalls(FRecordCalls);
  PlotChanged;
end;

procedure  TGraphD.SetRecordCalls(ARecording:Boolean);
begin
  FRecordCalls := ARecording;
  if ARecording or Assigned(FOnDReplot) then
      inherited OnReplot := {$IFDEF FPC}@{$ENDIF}DoGraphReplot
  else
      inherited OnReplot := Nil;
  if not ARecording then begin
    RStream.Free;
    RStream := nil;
  end;
end;

function TGraphD.CanPrint: Boolean;
begin
  if RecordCalls then
    Result := Assigned(FOnDReplot) or ( (RStream <> nil) and (RStream.Size > 1) )
  else
    Result := inherited CanPrint;
end;

procedure TGraphD.CopyToClipboard;
begin
  if RecordCalls and (Assigned(FOnDReplot) or ( (RStream <> nil) and (RStream.Size > 1)) ) then begin
    ClipBoardStream := THugeMemoryStream.Create;
    Clipboard.Open;
    try
      AssignStream(ClipFile,ClipBoardStream);
      ReWrite(ClipFile);
      inherited;
    finally
      Close(ClipFile);
      THugeMemoryStream(ClipBoardStream).CopyToClipboard(CF_TEXT);
      ClipBoardStream := nil;
      Clipboard.Close;
    end;
  end else
    inherited;
end;

procedure TGraphD.ExportTo(AStream: TStream; CanvasClass: TStreamCanvasClass);
begin
  if CanvasClass = TTextCanvas then begin
    { Special case - treat text file as wanting real number output }
    ClipBoardStream := AStream;
    try
      AssignStream(ClipFile, ClipBoardStream);
      ReWrite(ClipFile);
      inherited ExportTo(nil, TNullCanvas)
    finally
      Close(ClipFile);
      ClipBoardStream := nil;
    end;
  end else
    inherited;
end;

procedure TGraphD.PutReals(t:RecordType; d:array of Double);
begin
  PutType(t);
  RStream.WriteBuffer(d, SizeOf(d));
end;

procedure TGraphD.PutType(t:RecordType);
begin
  RStream.WriteBuffer(t, SizeOf(t));
end;

procedure TGraphD.PutString(s:string);
var
  b: Byte;
begin
{$IFNDEF VER80}
  if Length(s) > 255 then
    b := 255
  else
    b := Length(s);
  RStream.WriteBuffer(b, 1);
  if b > 0 then
    RStream.WriteBuffer(s[1], b);
{$ELSE}
  RStream.WriteBuffer(s, Length(s)+1);
{$ENDIF}
end;

procedure TGraphD.PlayBack(s:TStream);
var
  t: RecordType;
  d:array [0..4] of Double;
  Graphic: TGraphic;
  mode: TAlignment;
  buffer, buffer2:string;
  n:Integer;
  xaxis:Boolean;
  AWidth: Single;
  Colour: TColor;
  APattern: TLinePattern;
  Size: Single;
  MarkType: TMarkType;
  DotAtCentre: Boolean;
  BrushStyle: TBrushStyle;
  AStyle: TLineStyle;
  procedure ReadString(var str:string);
  var
    b:Byte;
  begin
    s.ReadBuffer(b, 1);
    SetLength(str, b);
    if b > 0 then s.ReadBuffer(str[1], b);
  end;
begin
   if s = nil then Exit;
   s.Position := 0;
   AStyle := nil;
   try
     while (s.Read(t, SizeOf(t))=SizeOf(t)) and (t <> tyEnd ) do
      case t of
        tyMap:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            Map(d[0], d[1], d[2], d[3]);
          end;
        tyPspace:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            Pspace(d[0], d[1], d[2], d[3]);
          end;
        tyMapLin:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            MapLin(d[0], d[1], d[2], d[3]);
          end;
        tyDefScale: DefaultScale;
        tyPoint:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            Point(d[0], d[1]);
          end;
        tySymbol:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            s.ReadBuffer(n, SizeOf(Integer));
            Symbol(d[0], d[1], n);
          end;
        tyJoin:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            Join(d[0], d[1]);
          end;
        tyClipPoint:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            ClipPoint(d[0], d[1]);
          end;
        tyClipJoin:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            ClipJoin(d[0], d[1]);
          end;
        tyTextJ:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            s.ReadBuffer(mode, SizeOf(TAlignment));
            ReadString(Buffer);
            Justify(mode);
            TextI(ProcX(d[0]), ProcY(d[1]), buffer);
          end;
        tyTitlen:
          begin
            s.ReadBuffer(n, SizeOf(Integer));
            ReadString(Buffer);
            Titlen(n, buffer);
          end;
        tyAxis:
          begin
            s.ReadBuffer(d, 3*SizeOf(Double));
            s.ReadBuffer(xaxis, SizeOf(Boolean));
            ReadString(Buffer);
            Axis(d[0], d[1], d[2], xaxis, buffer);
          end;
        tyScales:
          begin
            ReadString(Buffer);
            ReadString(Buffer2);
            Scales(buffer, buffer2);
          end;
        tyXaxLab:
          begin
            ReadString(Buffer);
            XaxLab(buffer);
          end;
        tyYaxLab:
          begin
            ReadString(Buffer);
            YaxLab(buffer);
          end;
        tyAxes:
          begin
            ReadString(Buffer);
            ReadString(Buffer2);
            Axes(buffer, buffer2);
          end;
        tyTextAngle:
          begin
            s.ReadBuffer(d, SizeOf(Double));
            TextAngle(d[0]);
          end;
        tyStyle:
          begin
            s.ReadBuffer(n, SizeOf(Integer));
            Style(n);
          end;
        tyRectangleEx:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            s.ReadBuffer(BrushStyle, SizeOf(BrushStyle));
            RectangleEx(BrushStyle, d[0], d[1], d[2], d[3], d[4]);
          end;
        tyRectangle:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            Rectangle(d[0], d[1], d[2], d[3]);
          end;
        tyGraphic:
          begin
            s.ReadBuffer(d, 4*SizeOf(Double));
            s.ReadBuffer(Graphic, Sizeof(Graphic));
            DrawGraphic(d[0], d[1], d[2], d[3], Graphic);
          end;
        tyPolyStart: PolyStart;
        tyPolyEnd:   PolyEnd;
        tyPolyPoint:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            PolyPoint(d[0], d[1]);
          end;
        tyStyleT:
          begin
            s.ReadBuffer(AWidth, SizeOf(AWidth));
            s.ReadBuffer(Colour, SizeOf(Colour));
            s.ReadBuffer(APattern, SizeOf(APattern));
            if not Assigned(AStyle) then
              AStyle := TLineStyle.Create;
            AStyle.Width := AWidth;
            AStyle.Colour := Colour;
            AStyle.Pattern := APattern;
            StyleT(AStyle);
          end;
        tySymbolT:
          begin
            s.ReadBuffer(d, 2*SizeOf(Double));
            s.ReadBuffer(Size, SizeOf(Size));
            s.ReadBuffer(Colour, SizeOf(Colour));
            s.ReadBuffer(MarkType, SizeOf(MarkType));
            s.ReadBuffer(DotAtCentre, SizeOf(DotAtCentre));
            CMarkStyle.Size := Size;
            CMarkStyle.Colour := Colour;
            CMarkStyle.MarkType := MarkType;
            CMarkStyle.DotAtCentre := DotAtCentre;
            SymbolT(d[0], d[1], CMarkStyle);
          end;
        tyErrorBarX:
          begin
            s.ReadBuffer(d, 3*SizeOf(Double));
            ErrorBarX(d[0], d[1], d[2]);
          end;
        tyErrorBarY:
          begin
            s.ReadBuffer(d, 3*SizeOf(Double));
            ErrorBarY(d[0], d[1], d[2]);
          end;
        tyNoYaxis: NoYaxis;
      else
        raise Exception.Createfmt('TGraphD:Bad Record %d', [Ord(t)]);
      end;
   finally
     AStyle.Free;
   end;
end;


function TGraphD.ProcX(x:double):Integer;
begin
  x := (x-Xmin)*Xs + Xof;
  if abs(x) > MaxPlot then
    if x < 0 then
      Result := -MaxPlot
    else
      Result := MaxPlot
  else
    Result := Round(x);
end;

function TGraphD.ProcY(y:double):Integer;
begin
  y := (y-Ymin)*Ys + Yof;
  if abs(y) > MaxPlot then
    if y < 0 then
      Result := -MaxPlot
    else
      Result := MaxPlot
  else
    Result := Round(y);
end;

procedure TGraphD.LPtoXY(lp: TPoint; var x, y:Double);
begin
  if Xs <> 0 then
    x := (lp.x-Xof)/Xs + Xmin;
  if Ys <> 0 then
    y := (lp.y-Yof)/Ys + Ymin;
end;

procedure TGraphD.XYtoLP(var lp: TPoint; x, y:Double);
begin
  lp.x := ProcX(x);
  lp.y := ProcY(y);
end;

procedure TGraphD.RecomputeScale;
begin
  Xof := Xp1*Xpixels;
  Yof := Yp1*Ypixels;
  if Xmax <> Xmin then Xs := (Xp2-Xp1)*Xpixels/(Xmax-Xmin);
  if Ymax <> Ymin then Ys := (Yp2-Yp1)*Ypixels/(Ymax-Ymin);
end;

procedure TGraphD.Map(xl:double; xu:double; yl:double; yu:double);
begin
  if Recording then
    PutReals(tyMap, [xl, xu, yl, yu]);
  FXmin := xl;
  FYmin := yl;
  FXmax := xu;
  FYmax := yu;
  RecomputeScale;
end;

procedure TGraphD.MapArg(var xl, xu, yl, yu:double);
begin
  xl := Xmin;
  yl := Ymin;
  xu := Xmax;
  yu := Ymax;
end;

procedure TGraphD.Pspace(xl:double; xu:double; yl:double; yu:double);
begin
  if Recording then
    PutReals(tyPspace, [xl, xu, yl, yu]);
  if xl <> xu then begin
    Xp1 := xl;
    Xp2 := xu;
  end else
    DefaultScale;
  if yl <> yu then begin
    Yp1 := yl;
    Yp2 := yu;
  end else
    DefaultScale;
  RecomputeScale;
end;

procedure TGraphD.NoYaxis;
begin
  if Recording then
    PutType(tyNoYaxis);
  Xp1 := 1 - Xp2;
  RecomputeScale;
end;


procedure TGraphD.PspaceArg(var xl, xu, yl, yu:double);
begin
  xl := Xp1;
  yl := Yp1;
  xu := Xp2;
  yu := Yp2;
end;

function TGraphD.PlotLength:Double; { Plot length in mm on printer }
begin
  Result := (Xp2-Xp1) * Xmm;
end;

procedure TGraphD.MapLin(xl:double; xu:double; yl:double; yu:double);
var
  t, fac, dy, dx: Double;
begin
  if Recording then
    PutReals(tyMapLin, [xl, xu, yl, yu]);
  { MapLin ensures that at least the range of x) and (y specified is
    displayed) and sets map such that a displacement in an arbitrary
    direction will cause the same movement to be displayed. }
  if (Xp2 = Xp1) or (xu = xl) or (Yp2 = Yp1) then Exit;
  t := abs( (Yp2 - Yp1) * YPixels / YPerX
            / (Xp2 - Xp1) / XPixels );
  if xu <> xl then
    fac := abs( (yu-yl) / (xu-xl) )
  else
    fac := t+1; { Force too much y }
  if fac > t then begin
    { too much y - extend x }
    dx := (abs(yu-yl)/t - abs(xu-xl))/2;
    if xl > xu then dx := -dx;
    xl := xl - dx;
    xu := xu + dx;
  end else if fac < t then begin
    { too much x - extend y }
    dy := (t*abs(xu-xl) - abs(yu-yl))/2;
    if yl > yu then dy := -dy;
    yl := yl - dy;
    yu := yu + dy;
  end;
  FXmin := xl;
  FYmin := yl;
  FXmax := xu;
  FYmax := yu;
  RecomputeScale;
end;

procedure  TGraphD.DefaultScale;
var
  xc: Double;
begin
  if Recording then
    PutType(tyDefScale);
  FXmin := 0;
  FYmin := 0;
  FXmax := 1;
  FYmax := 1;
  xc := AvCharWidth/XPixels;
  Xp2 := 1-3.5*xc;
  Xp1 := xc*8;
  if Xp1 >= Xp2 then begin
    { Plot too small, or font too large }
    Xp1 := 0.1;
    Xp2 := 0.9;
  end;
  Yp1 := 2*CellHeight/YPixels;
  Yp2 := 1-Yp1;
  if Yp1 >= Yp2 then begin
    { Plot too small, or font too large }
    Yp1 := 0.1;
    Yp2 := 0.9;
  end;
  RecomputeScale;
end;

procedure  TGraphD.Point(x:double; y:double);
begin
  if Assigned(ClipBoardStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y);
  end;
  if Recording then
    PutReals(tyPoint, [x, y])
  else begin
      Xlast := x;  Ylast := y;
      Move(ProcX(x), ProcY(y));
  end;
end;

procedure TGraphD.Symbol(x:double; y:double; n:Integer);
begin
  if Assigned(ClipBoardStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y, #9, n);
  end;
  if Recording then begin
    PutReals(tySymbol, [x, y]);
    RStream.WriteBuffer(n, SizeOf(n));
  end else if OutCode(x,y) = 0 then
    Mark(ProcX(x), ProcY(y), n);
end;

procedure TGraphD.SymbolT(x:double; y:double; AStyle: TMarkStyle);
var
  Size: Single;
  Colour: TColor;
  MarkType: TMarkType;
  DotAtCentre: Boolean;
begin
  if Assigned(ClipBoardStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y, #9, AStyle.Colour, #9, Ord(AStyle.MarkType)+100*Ord(AStyle.DotAtCentre), #9, AStyle.Size);
  end;
  if Recording then begin
    PutReals(tySymbolT, [x, y]);
    Size := AStyle.Size;
    Colour := AStyle.Colour;
    MarkType := AStyle.MarkType;
    DotAtCentre := AStyle.DotAtCentre;
    RStream.WriteBuffer(Size, SizeOf(Size));
    RStream.WriteBuffer(Colour, SizeOf(Colour));
    RStream.WriteBuffer(MarkType, SizeOf(MarkType));
    RStream.WriteBuffer(DotAtCentre, SizeOf(DotAtCentre));
  end else if OutCode(x,y) = 0 then
    MarkT(ProcX(x), ProcY(y), AStyle);
end;

procedure TGraphD.ErrorBarX(x1, x2, y:double);
begin
  if Recording then
    PutReals(tyErrorBarX, [x1, x2, y])
  else begin
    ClipPoint(x1, y);
    ClipJoin(x2, y);
    MarkE(ProcX(x1), ProcY(y), mtVBar);
    MarkE(ProcX(x2), ProcY(y), mtVBar);
  end;
end;

procedure TGraphD.ErrorBarY(x, y1, y2:double);
begin
  if Recording then
    PutReals(tyErrorBarY, [x, y1, y2])
  else begin
    ClipPoint(x, y1);
    ClipJoin(x, y2);
    MarkE(ProcX(x), ProcY(y1), mtHBar);
    MarkE(ProcX(x), ProcY(y2), mtHBar);
  end;
end;

procedure TGraphD.Join(x:double; y:double);
begin
  if Assigned(ClipBoardStream) then
    WriteLn(ClipFile, x, #9, y);
  if Recording then
    PutReals(tyJoin, [x, y])
  else begin
    Xlast := x;  Ylast := y;
    Vector(ProcX(x), ProcY(y));
  end;
end;

procedure TGraphD.RectangleEx(AStyle: TBrushStyle; Density: Double; x1, x2, y1, y2:double);
var
  p: Array [0..1] of TPoint;
  function Check(var v1, v2:Double; vmin, vmax:Double):Boolean;
  var
    t:Double;
  begin
    if (v1 < v2) <> (vmin < vmax) then begin
      t := v1;
      v1 := v2;
      v2 := t;
    end;
    if vmin < vmax then
      Result := (v2 >= vmin) and (v1 <= vmax)
    else
      Result := (v2 <= vmin) and (v1 >= vmax);
  end;
begin
  if Recording then begin
    PutReals(tyRectangleEx, [Density, x1, x2, y1, y2]);
    RStream.WriteBuffer(AStyle, SizeOf(AStyle));
  end else if Check(x1, x2, Xmin, Xmax) and Check(y1, y2, Ymin, Ymax) then begin
    p[0].x := ProcX(x1);
    p[1].x := ProcX(x2);
    p[0].y := ProcY(y1);
    p[1].y := ProcY(y2);
    Box(AStyle, Density, TRect(p));
  end;
end;

procedure TGraphD.Rectangle(x1, x2, y1, y2:double);
begin
  if Recording then
    PutReals(tyRectangle, [x1, x2, y1, y2])
  else
    RectangleEx(bsSolid, -1, x1, x2, y1, y2);
end;

procedure TGraphD.DrawGraphic(x1, x2, y1, y2:Double; Graphic: TGraphic);
var
  Rect: TRect;
begin
  if Recording then begin
    PutReals(tyGraphic, [x1, x2, y1, y2]);
    RStream.WriteBuffer(Graphic, SizeOf(Graphic));
  end else begin
    Rect.Left := ProcX(x1);
    Rect.Right := ProcX(x2);
    Rect.Top := ProcY(y1);
    Rect.Bottom := ProcY(y2);
    inherited DrawGraphic(Rect, Graphic);
  end;
end;

function TGraphD.OutCode(x:double; y:double):Integer;
begin
    (* Return bitmap indicating where x, y is with respect
       to the region specified by map:
       Result has bits set as follows:
       Bit 0: (1) to left
           1: (2) above
           2: (4) to right
           3: (8) below
     *)
    Result := 0;
    if Xmin < Xmax then begin
        if x < Xmin then Result := 1;
        if x > Xmax then Result := Result + 4;
    end else begin
        if x < Xmax then Result := 1;
        if x > Xmin then Result := Result + 4;
    end;
    if Ymin < Ymax then begin
        if y < Ymin then Result := Result + 8;
        if y > Ymax then Result := Result + 2;
    end else begin
        if y < Ymax then Result := Result + 8;
        if y > Ymin then Result := Result + 2;
    end;
end;

procedure TGraphD.clppnt(var x1:double; var y1:double; x2:double; y2:double; var out:Integer);
begin
    { Internal routine to clip start point of line (x1, y1)
      NB: no need to test for divide by zero here;
      eliminated by pre-testing }
    if (out and 5) <> 0 then
      if ((out and 1) <> 0) xor (Xmin > Xmax) then begin
          y1 := y1 + (y2-y1)*(Xmin-x1)/(x2-x1);
          x1 := Xmin;
      end else { if (out and 4) <> 0 then } begin
          y1 := y1 + (y2-y1)*(Xmax-x1)/(x2-x1);
          x1 := Xmax;
      end
    else if (out and 10) <> 0 then
      if ((out and 2) <> 0) xor (Ymin > Ymax) then begin
          x1 := x1 + (x2-x1)*(Ymax-y1)/(y2-y1);
          y1 := Ymax;
      end else { if (out and 8) <> 0 then } begin
          x1 := x1 + (x2-x1)*(Ymin-y1)/(y2-y1);
          y1 := Ymin;
      end
    else
      Exit;
    out := OutCode(x1, y1);
end;

procedure TGraphD.Clamp(var x1:double; var y1:double; AOutCode :Integer);
begin
    (* Internal routine to force point to boundary *)
    if (AOutCode and 1) <> 0 then
        x1 := Xmin
    else if (AOutCode and 4) <> 0 then
        x1 := Xmax;
    if (AOutCode and 2) <> 0 then
        y1 := Ymax
    else if (AOutCode and 8) <> 0 then
        y1 := Ymin;
end;

procedure TGraphD.ClipPoint(x:double; y:double);
begin
  if Recording then
    PutReals(tyClipPoint, [x, y])
  else begin
    Xlast := x;  Ylast := y;
    if OutCode(x,y) = 0 then
      Move(ProcX(x), ProcY(y));
  end;
end;

procedure  TGraphD.ClipJoin(x1:double; y1:double);
var
  x, y: Double;
  orgout1, out1, out2: Integer;
begin
    if Recording then begin
      PutReals(tyClipJoin, [x1, y1]);
      Exit;
    end;
    (* Join previous Point to x, y; clips line to edge of map area
       Clip adapted from R Wilton, Programmers Guide to PC) and (PS/2 Video Systems,  p 217
     *)
    if IsNanOrInf(XLast) or IsNanOrInf(YLast) then
      ClipPoint(x1, y1)
    else begin
      x := x1;
      y := y1;
      orgout1 := OutCode(Xlast, Ylast);
      out1 := orgout1;
      out2 := OutCode(x, y);
      while (out1 and out2) = 0 do begin
        (* Some of line visible *)
        if (out1 or out2) <> 0 then begin
            (* Clipping Required *)
            if out1 <> 0 then begin
              clppnt(Xlast, Ylast, x, y, out1);
              Continue;
            end;
            if out2 <> 0 then clppnt(x, y, Xlast, Ylast, out2);
            Continue;
        end;
        if orgout1 <> 0 then Point(Xlast, Ylast);
        Join(x, y);
        Break;
      end;
      Xlast := x1;
      Ylast := y1;
    end;
end;

procedure TGraphD.PolyStart;
begin
  if Assigned(ClipBoardStream) then
    WriteLn(ClipFile);
  if Recording then
    PutType(tyPolyStart)
  else begin
    {$IFNDEF DYNAMIC_ARRAYS}
    if FPolygonPoints = nil then
      New(FPolygonPoints);
    {$ENDIF}
    if NPolygonPoints <> 0 then
      raise Exception.CreateFmt('PolyStart with %d points unshown', [NPolygonPoints]);
  end;
end;

procedure TGraphD.PolyPoint(x1:double; y1:double);
var
  x, y: Double;
  orgout1, out1, out2: Integer;
  procedure AddPoint(x, y: Double);
  var
    ix, iy: Integer;
  begin
    ix := ProcX(x);
    iy := ProcY(y);
    if (NPolygonPoints = 0)
       or (ix <> FPolygonPoints{$IFNDEF DYNAMIC_ARRAYS}^{$ENDIF}[NPolygonPoints-1].x)
       or (iy <> FPolygonPoints{$IFNDEF DYNAMIC_ARRAYS}^{$ENDIF}[NPolygonPoints-1].y) then begin
      LogToDev(ix, iy); { TODO : Shift this into TGraphB }
      {$IFDEF DYNAMIC_ARRAYS}
      if High(FPolygonPoints) <= NPolygonPoints then
        SetLength(FPolygonPoints, NPolygonPoints+300);
      {$ENDIF}
      FPolygonPoints{$IFNDEF DYNAMIC_ARRAYS}^{$ENDIF}[NPolygonPoints].x := ix;
      FPolygonPoints{$IFNDEF DYNAMIC_ARRAYS}^{$ENDIF}[NPolygonPoints].y := iy;
      Inc(NPolygonPoints);
    end;
  end;
begin
  if Assigned(ClipBoardStream) then
    WriteLn(ClipFile, x, #9, y);
  if Recording then begin
    PutReals(tyPolyPoint, [x1, y1]);
    Exit;
  end;
  {$IFNDEF DYNAMIC_ARRAYS}
  if FPolygonPoints = nil then
    raise Exception.Create('PolyPoint without PolyStart');
  if NPolygonPoints > High(FPolygonPoints^) then
    raise Exception.CreateFmt('Too many PolyPoints (%d)', [NPolygonPoints]);
  {$ENDIF}
  x := x1;
  y := y1;
  out2 := OutCode(x1, y1);
  if NPolygonPoints > 0 then begin
    out1 := OutCode(XLastPoly, YLastPoly);
    orgout1 := out1;
    while (out1 and out2) = 0 do begin
      (* Some of line visible *)
      if (out1 or out2) <> 0 then begin
          (* Clipping Required *)
          if out1 <> 0 then begin
            clppnt(XlastPoly, YlastPoly, x, y, out1);
            Continue;
          end;
          if out2 <> 0 then clppnt(x, y, XlastPoly, YlastPoly, out2);
          Continue;
      end;
      if orgout1 <> 0 then AddPoint(XlastPoly, YlastPoly);
      AddPoint(x, y);
      Break;
    end;
  end;
  if (out2 <> 0) or (NPolygonPoints = 0) then begin
    Clamp(x, y, out2);
    AddPoint(x, y);
  end;
  XLastPoly := x1;
  YLastPoly := y1;
end;

procedure TGraphD.PolyEnd;
var
  InPlot:Boolean;
begin
  if Recording then
    PutType(tyPolyEnd)
  else begin
    if {$IFNDEF DYNAMIC_ARRAYS}(FPolygonPoints = nil) or {$ENDIF} (NPolygonPoints = 0) then
      Exit;
    InPlot := CurrentCanvas <> nil;
    if not InPlot then
      start_plotting;
    NeedLinePen;
    CurrentCanvas.Pen.Style := psClear;
    {$IFNDEF DYNAMIC_ARRAYS}
    CurrentCanvas.Polygon(FPolygonPoints^, NPolygonPoints);
    {$ELSE}
    SetLength(FPolygonPoints, NPolygonPoints);
    CurrentCanvas.Polygon(FPolygonPoints, Length(FPolygonPoints));
    {$ENDIF}
    PenInvalid;
    if not InPlot then
      stop_plotting;
    NPolygonPoints := 0;
  end;
end;

function TGraphD.cx: Double;
begin
  Result := AvCharWidth / Xs;
end;

function TGraphD.cy: Double;
begin
  Result := CellHeight / Ys;
end;

procedure TGraphD.Chlen(var Acx:double; var Acy:double);
begin
  Acy := cy;
  Acx := cx;
end;

procedure TGraphD.Start;
begin
  if not InReplot then begin
    SetRecordCalls(FRecordCalls);   (* Ensure Event handler set correctly *)
    if (FRecordCalls) then begin
      DefaultScale; { Make some half way sensible character size values }
      Recording := true;
      if RStream = nil then RStream := THugeMemoryStream.Create;
      RStream.Position := 0;
    end else begin
      inherited Start;
      DefaultScale;
    end;
  end;
end;

procedure TGraphD.Finish;
begin
  if not InReplot then begin
    if Recording then begin
      PutType(tyEnd);
      Recording := false;
      Invalidate;
    end else
      inherited Finish;
  end;
end;

procedure TGraphD.TextAngle(ang:double);
begin
    if Recording then
      PutReals(tyTextAngle, [ang])
    else
      inherited Angle(Round(ang*10));
end;

procedure TGraphD.Style(AStyle:Integer);
begin
    if (Recording) then begin
      PutType(tyStyle);
      RStream.WriteBuffer(AStyle, SizeOf(Integer));
    end else
      Linetype := AStyle;
end;

procedure TGraphD.StyleT(AStyle:TLineStyle);
var
  AWidth: Single;
  Colour: TColor;
  APattern: TLinePattern;
begin
    if (Recording) then begin
      PutType(tyStyleT);
      AWidth := AStyle.Width;
      Colour := AStyle.Colour;
      APattern := AStyle.Pattern;
      RStream.WriteBuffer(AWidth, SizeOf(AWidth));
      RStream.WriteBuffer(Colour, SizeOf(Colour));
      RStream.WriteBuffer(APattern, SizeOf(APattern));
      CLineStyle.Assign(AStyle);
    end else
      Linestyle := AStyle;
end;

procedure TGraphD.StyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
begin
  CLineStyle.Colour := Colour;
  CLineStyle.Width := AWidth;
  CLineStyle.Pattern := APattern;
  StyleT(CLineStyle);
end;

procedure TGraphD.TextJ(x:double; y:double; mode:TAlignment; s:string);
begin
    if Recording then begin
      PutReals(tyTextJ, [x, y]);
      RStream.WriteBuffer(mode, SizeOf(mode));
      PutString(s);
    end else begin
      Justify(mode);
      TextI(ProcX(x), ProcY(y), s);
    end;
end;

procedure TGraphD.Text(x:double; y:double; s:string);
var
  p: Integer;
  SavedClipBoardStream: TStream; { Current text stream file }
begin
  { Expand escape sequences, signified by #1 }
  while Length(s) > 0 do begin
    p := Pos(#1, s);
    if p = 0 then begin
      TextJ(x, y, taLeftJustify, s);
      Break;
    end else if p > 0 then begin
      TextJ(x, y, taLeftJustify, Copy(s, 1, p-1));
      Delete(s,1,p-1);
      x := x + (p-1)*cx;
    end;
    if Length(s) < 3 then Break; { Bad escape sequence }
    CMarkStyle.Size := 2;
    CMarkStyle.Colour := Ord(s[3]);
    CMarkStyle.MarkType := TMarkType(Ord(FirstMark) + (Ord(s[2]) and $7f));
    CMarkStyle.DotAtCentre := Ord(s[2]) < 128;
    SavedClipBoardStream := ClipBoardStream; { Disable text output here }
    try
      ClipBoardStream := nil;
      MarkT(ProcX(x+cx), ProcY(y+cy/2), CMarkStyle);
    finally
      ClipBoardStream := SavedClipBoardStream;
    end;
    Delete(s,1,3);
    x := x + 2.5*cx;
  end;
end;

procedure TGraphD.Titlen(n:Integer; str:string);
var
  Weight: Longint;
begin
    if (Recording) then begin
      PutType(tyTitlen);
      RStream.WriteBuffer(n, SizeOf(n));
      PutString(str);
    end else
      if n = 1 then begin
        {$IFNDEF CLX_}
        Weight := GraphicsLogFont.lfWeight;
        GraphicsLogFont.lfWeight := FW_BOLD;
        {$ELSE}
        Weight := CurrentCanvas.Font.Weight;
        CurrentCanvas.Font.Weight := fwBold;
        {$ENDIF}
        WIN_change_font;
        TextJ((Xmin+Xmax)/2, Ymax+n*cy, taCenter, str);
        {$IFNDEF CLX_}
        GraphicsLogFont.lfWeight := Weight;
        {$ELSE}
        CurrentCanvas.Font.Weight := Weight;
        {$ENDIF}
        WIN_change_font;
      end else
        Text(Xmin+1.5*cx, Ymax+n*cy, str);
end;

procedure TGraphD.Title1(title:string);
begin
  Titlen(1, title);
end;

procedure TGraphD.Title2(title:string);
begin
  Titlen(0, title);
end;

procedure  AxisInterval(
(*
  Given an axis length, range, work out the minor tick Integererval, ds
  The major ticks are assumed to be at Intervals of 5*ds, and there
  will be at least nticks of these. Where exponential notation must
  be used, dividing by se := 10**ne will reduce d to the range 1 - 1000
  This routine will not give good results if the range is much smaller
  than the absolute values.
 *)
    (* Input: *)
    range:double; (* Axis length *)
    (* Input/output *)
    var nticks:Integer; (* Minimum number of major ticks (Forced to be >= 2) *)
    (* Outputs *)
    var ds:double;  (* Minor tick Interval divided by se (so Interval is actually ds*se) *)
    var se:double;  (* Scale factor required to reduce d to where exponential notation is not required *)
    var ne:Integer; (* se := 10**ne, i.e. power of 10 needed *)
    var nd:Integer  (* Number of places after the decimal Point required (>=0) *)
);
var
  d, t, s:Double;
  ll:Integer;
begin
    (* Range (to be scaled) *)
    d := abs(range);
    if d = 0 then d := 1;
    (* First Reduce length (d) to approx engineering notation
       i.e 1 - 1000, though 0.1 - 1 and 1e3 - 1e5 are allowed to stand *)
    ne := 0; (* power of 10 *)
    se := 1; (* scale factor (10**ne) *)
    while True do begin
        if (d >= 1e5) or ( (d >= 1e3) and (ne <> 0) ) then begin
            t := d / 1000;
            if t >= d then Exit; (* Only true if d is a nan *)
            d := t;
            se := se * 1000;
            ne := ne + 3;
        end else if (d < 1e-3) or ( (d < 1 ) and (ne <> 0) ) then begin
            d := d * 1000;
            ne := ne - 3;
            se := se / 1000;
        end else
            Break;
    end;
    (* Now reduce to 1 <= d < 10 *)
    s := 1;  (* scale factor *)
    nd := 0; (* power of 10 *)
    while True do begin
        if d < 1 then begin
            d := d * 10;
            s := s / 10;
            nd := nd + 1;
        end else if d >= 10 then begin
            d := d / 10;
            s := s * 10;
            nd := nd - 1;
        end else
            Break;
    end;
    (* Now decide major tick spacing
       decrease ds in a 10 5 2 1 sequence) and (adjust format
       until a reasonable number is obtained *)
    ds := 1;
    ll := 0; (* Current position in 10 5 2 1 ... sequence *)
    if nticks <= 2 then begin
        (* Fix v small plots *)
        nticks := 2;
        if d/ds > 2 then begin
          ds := ds * 2;
          ll := 2;
        end;
        if d/ds > 2 then begin
          ds := ds * 2.5;
          ll := 1;
        end;
    end;
    while d/ds < nticks do begin
        ll := ll + 1;
        case ll of
          1:     (* convert 1 to 0.5 *)
            begin
              ds := ds/2;
              nd := nd + 1;
            end;
          2:     (* convert 0.5 to 0.2 *)
              ds := ds*0.4;
          3:     (* convert 0.2 to 0.1 *)
              ds := ds/2;
        end;
        ll := ll mod 3;
    end;

    ds := ds * s / 5;    (* Minor tick spacing *)

    (* Calculate mark positions *)
    if nd < 0 then nd := 0;
    if range < 0 then ds := -ds;
end;

procedure TGraphD.SetAxisStyle(const Value: TLineStyle);
begin
  FLineStyles[AxisPen].Assign(Value);
end;

function TGraphD.GetAxisStyle: TLineStyle;
begin
  Result := FLineStyles[AxisPen];
end;

function TGraphD.IsNotAxisDefault: Boolean;
begin
  Result := (AxisStyle.Width <> 0)
            or (AxisStyle.Colour <> clBlack)
            or (AxisStyle.Pattern <> psSolid);
end;

procedure TGraphD.ScaledAxis(
{ Draw an x or y axis with a different scale than implied by the current map();
  For an x axis the parameters are: }
   xle, xue: Double;       { The values to display at xl1 and xl2, respectively }
   xl1:double; xu1:double; { The range the axis should cover }
   yl:double;              { Where to plot the axis }
   xaxis:Boolean;          { False for a yaxis, in which case swap x and y for the desciptions above }
   s:string                { The axis label }
  );
var
  xi, xx, yi, yx: Double;
begin
  if xl1 = xu1 then
    Exit;
  MapArg(xi, xx, yi, yx);
  try
    if xaxis then
      Map(xle+(xue-xle)*(xi-xl1)/(xu1-xl1), xue+(xue-xle)*(xx-xu1)/(xu1-xl1), yi, yx)
    else
      Map(xi, xx, xle+(xue-xle)*(yi-xl1)/(xu1-xl1), xue+(xue-xle)*(yx-xu1)/(xu1-xl1));
    Axis(xle, xue, yl, xaxis, s);
  finally
    Map(xi, xx, yi, yx);
  end;
end;

procedure TGraphD.Axis(
(* Draw an x) or (y axis. For an x axis the parameters are: *)
    xl1:double; (* The start of the axis *)
    xu1:double; (* The end of the axis *)
    yl:double;  (* The y co-ordinate of the axis *)
    xaxis:Boolean; (* false for a y axis *)
    s:string(* The label *)
(* For a y axis, swap x) and (y above) and (xaxis := false *)
);
var
  cxl, cyl, xl, xu, ds, se, t, xc: Double;
  nticks, ne, nd: Integer;
  tt:Longint;
  TickType: TMarkType;
  SavedStyleNumber: Integer;
  SavedLineStyle: TLineStyle;
  SavedClipBoardStream: TStream; { Current text stream file }
begin
  if Recording then begin
    PutReals(tyAxis, [xl1, xu1, yl]);
    RStream.WriteBuffer(xaxis, SizeOf(xaxis));
    PutString(s);
    Exit;
  end;
  SavedStyleNumber := FLinetype;
  SavedLineStyle := nil;
  if SavedStyleNumber < 0 then begin
    SavedLineStyle := TLineStyle.Create;
    SavedLineStyle.Assign(ClineStyle);
  end;
  SavedClipBoardStream := ClipBoardStream;
  ClipBoardStream := nil; { Turn off Text stream for axis }
  try
    LineStyle := AxisStyle;
    (* Determine steps for annotation *)
    Chlen(cxl, cyl);
    xl := xl1;
    xu := xu1;
    if xu < xl then begin
      xl := xu1;
      xu := xl1;
    end;
    (* Draw Axis line *)
    (* Nticks = Approximate number of major ticks wanted *)
    if xaxis then begin
        Point(xl, yl);
        Join(xu, yl);
        nticks := Round((xu-xl)/cxl/16);
    end else begin
        Point(yl, xl);
        Join(yl, xu);
        nticks := Round((xu-xl)/cyl/8);
    end;
    SetCurrentMarkSize(Font.Size / 3);
    (* ds = Minor tick interval *)
    AxisInterval(xu-xl, nticks, ds, se, ne, nd);
    (* Convert to non-exponential range *)
    xl := xl/se;
    xu := xu/se;
    { Give up if range too small }
    if Abs(xl) >= MaxLongint*ds then Exit;
    xc := Trunc(xl/ds)*ds;
    if abs(xc-ds-xl) < ds/1000 then
      xc := xc - ds;      { Allow for point just outside range }
    (* draw ticks *)
    while xc <= (xu + ds/1000) do begin
        if xc >= xl then begin
            t := xc*se;
            tt := Round(abs(xc)/ds);
            if tt mod 5 = 0 then
              (* A major (labelled) tick mark *)
              if xaxis then begin
                TextJ(xc*se, yl-cyl, taCenter, Format('%.*f', [nd, xc]));
                TickType := mtXMajor;
              end else begin
                TextJ(yl-cxl/6, xc*se-cyl/3, taRightJustify, Format('%.*f', [nd, xc]));
                TickType := mtYMajor;
              end
            else
              if xaxis then
                TickType := mtXMinor
              else
                TickType := mtYMinor;
            if xaxis then
              Gmarks.Mark(CurrentCanvas, ProcX(t), ProcY(yl), TickType, mp)
            else
              Gmarks.Mark(CurrentCanvas, ProcX(yl), ProcY(t), TickType, mp);
        end;
        xc := xc + ds;
    end;
    (* write * 10 ^ ? *)
    if ne <> 0 then begin
       if xaxis then
        TextJ(xu1+3.5*cxl, yl-2*cyl, taRightJustify, Format('/1e%d', [ne]))
       else begin
        TextAngle(90);
        TextJ(yl-5.99*cxl, xu1, taRightJustify, Format('/1e%d', [ne]));
        TextAngle(0);
       end;
    end;
    if xaxis then begin
      FXaxPos.X := (xl1+xu1)/2;
      FXaxPos.Y := yl-1.9*cyl;
      XaxLab(s);
    end else begin
      FYaxPos.X := yl-5.99*cxl;
      FYaxPos.Y := (xl1+xu1)/2;
      YaxLab(s);
    end;
  finally
    ClipBoardStream := SavedClipBoardStream;
    if SavedStyleNumber < 0 then begin
      LineStyle := SavedLineStyle;
      SavedLineStyle.Free;
    end else
      LineType := SavedStyleNumber;
  end;
end;

procedure TGraphD.XaxLab(xaxis:string);
begin
  if Length(xaxis) = 0 then Exit;
  if Recording then begin
    PutType(tyXaxLab);
    PutString(xaxis);
  end else
    TextJ(FXaxPos.X, FXaxPos.Y, taCenter, xaxis);
end;

procedure TGraphD.YaxLab(yaxis:string);
begin
  if Length(yaxis) = 0 then Exit;
  if Recording then begin
    PutType(tyYaxLab);
    PutString(yaxis);
  end else begin
    TextAngle(90);
    TextJ(FYaxPos.X, FYaxPos.Y, taCenter, yaxis);
    TextAngle(0);
  end;
end;

procedure TGraphD.Scales(xaxis:string; yaxis:string);
begin
  if Recording then begin
    PutType(tyScales);
    PutString(xaxis);
    PutString(yaxis);
  end else begin
    Axis(Xmin, Xmax, Ymin, true, xaxis);
    Axis(Ymin, Ymax, Xmin, false, yaxis);
  end;
end;

function TGraphD.PixelSize(x: Boolean): Double;
{ The size of one screen pixel }
begin
  if x then begin
    Result := XPixels / PaintWidth / XSCALE;
    if Xs <> 0 then Result := Result / Xs;
  end else begin
    Result := YPixels / PaintHeight / YSCALE;
    if Ys <> 0 then Result := Result / Ys;
  end;
  Result := Abs(Result);
end;

function TGraphD.CloseToX(x1, x2: Double):Boolean;
{ var
  p: array [0..1] of TPoint; }
begin
  if Xs <> 0 then begin
{   p[0].x := 0;
    p[1].x := GetSystemMetrics(SM_CXDOUBLECLK);
    p[0].y := 0;
    p[1].y := 0;
    DPtoLP(Canvas.Handle, p, 2);
    Result := Abs(x1-x2) < Abs((p[1].x-p[0].x)/Xs); }
    Result := Abs(x1-x2) < PixelSize(True)*SmallX;
  end else
    Result := False;
end;

function TGraphD.CloseToY(y1, y2: Double):Boolean;
{ var
  p: array [0..1] of TPoint; }
begin
  if Ys <> 0 then begin
{   p[0].x := 0;
    p[1].x := 0;
    p[0].y := 0;
    p[1].y := GetSystemMetrics(SM_CYDOUBLECLK);
    DPtoLP(Canvas.Handle, p, 2);
    Result := Abs(y1-y2) < Abs((p[1].y-p[0].y)/Ys); }
    Result := Abs(y1-y2) < PixelSize(False)*SmallY;
  end else
    Result := False;
end;

function AxPos(a:double; b:double):Double;
begin
  if ((a > 0) and (b > 0)) or ((a < 0) and (b < 0)) then begin
    Result := a;
    if b < a then Result := b;
  end else
    Result := 0;
end;

procedure TGraphD.Axes(xaxis:string; yaxis:string);
begin
    if Recording then begin
      PutType(tyAxes);
      PutString(xaxis);
      PutString(yaxis);
    end else begin
      Axis(Xmin, Xmax, AxPos(Ymin, Ymax), true, xaxis);
      Axis(Ymin, Ymax, AxPos(Xmin, Xmax), false, yaxis);
    end;
end;

procedure TGraphPrintPreview.SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double);
begin
  inherited SetMargins(AXSize, AYSize, AXMargin, AYMargin);
  if Assigned(FGraph) then begin
    if AXSize > 0 then Graph.Xmm := AXSize;
    if AYSize > 0 then Graph.Ymm := AYSize;
  end;
end;

procedure TGraphPrintPreview.SetGraph(G:TGraphB);
begin
  if Assigned(FGraph) then
    FGraph.FPrOnInvalidate := nil;
  FGraph := G;
  if Assigned(FGraph) then begin
    FGraph.FPrOnInvalidate := {$IFDEF FPC}@{$ENDIF}PInvalidate;
    if XSize > 0 then FGraph.Xmm := XSize;
    if YSize > 0 then FGraph.Ymm := YSize;
  end;
end;

procedure TGraphPrintPreview.DoShowPreview(Sender:TObject; Canvas:TExportableCanvas;
  X, Y, AWidth, AHeight:Longint; APage: Integer; OnPrinter:Boolean);
begin
  if Assigned(FGraph) then
    FGraph.ShowPrintPreview(Sender, Canvas, X, Y, AWidth, AHeight, APage, OnPrinter);
  inherited DoShowPreview(Sender, Canvas, X, Y, AWidth, AHeight, APage, OnPrinter);
end;

function TGraphD.GetSelectedMathArea: TDrect;
var
  r: Trect;
begin
  r := LogicalSelectedArea;
  LPtoXY(r.TopLeft, Result.Left, Result.Top);
  LPtoXY(r.BottomRight, Result.Right, Result.Bottom);
end;

procedure TGraphD.SetSelectedMathArea(ARect: TDrect);
var
  r: Trect;
begin
  XYtoLP(r.TopLeft, ARect.Left, ARect.Top);
  XYtoLP(r.BottomRight, ARect.Right, ARect.Bottom);
  LogicalSelectedArea := r;
end;

function TGraphD.RoundToDisplay(value: Double; xvalue: Boolean): Double;
var
  dRes: Double;
begin
  { Discard significant figures (in decimal representation!)
    to match display resolution }
  dRes := PixelSize(xvalue)/2;
  { Truncate to next lowest power of 10 }
  dRes := Power(10, Floor(Log10(dRes)));
  Result := Round(value/dRes)*dREs;
  if Abs(value-Result) > dRes then
    raise Exception.Create('RoundToDisplay');
end;

procedure TGraphD.SetLastMouse(X, Y: Integer);
begin
  inherited SetLastMouse(X, Y);
  LPtoXY(LastMouse, FLastMouseX, FLastMouseY);
end;

end.






