unit GridEx;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses
  Classes, SysUtils, Controls, StdCtrls, Grids;

type

  {$IFNDEF LCL_}
  TSelectEditorEvent = procedure(Sender: TObject; Col,Row: Integer; var Editor: TWinControl) of object;
  {$ENDIF}
  {$IFDEF LCL_}
  TEditStyle =  (esSimple, esEllipsis, esPickList);
  TOnGetPickListItems = procedure (ACol, ARow: Integer; AItems: TStrings) of object;
  {$ENDIF}

  TGetEditStyleEvent = procedure (Sender: TObject; ACol, ARow: Integer; var Style: TEditStyle) of object;


  TStringGridEx = class(TStringGrid)
  private
    FDropDownEditor: TComboBox;
    FOnGetEditStyle: TGetEditStyleEvent;
    FOnGetPickListItems: TOnGetPickListItems;
  protected
    {$IFDEF LCL_}
    procedure SelectEditor; override;
    function CreateDropDownEditor: TComboBox;
    procedure EditordoSetValue; override;
    {$ENDIF}
    {$IFDEF VCL_}
    function CreateEditor: TInplaceEdit; override;
    function GetEditStyle(ACol, ARow: Longint): TEditStyle; override;
    {$ENDIF}
    procedure DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
  public
    destructor Destroy; override;
    property DropDownEditor: TComboBox read FDropDownEditor;
  published
    property OnGetEditStyle: TGetEditStyleEvent read FOnGetEditStyle write FOnGetEditStyle;
    property OnGetPickListItems: TOnGetPickListItems read FOnGetPickListItems write FOnGetPickListItems;
  end;

implementation

{$IFDEF LCL_}
uses LCLType;

type
  TStringGridCellComboBox = class(TComboBox)
  private
    FGrid: TCustomGrid;
  protected
    procedure Change; override;
    procedure KeyDown(var Key : Word; Shift : TShiftState); override;
    procedure msg_SetValue(var Msg: TGridMessage); message GM_SETVALUE;
    procedure msg_GetValue(var Msg: TGridMessage); message GM_GETVALUE;
    procedure msg_SetGrid(var Msg: TGridMessage); message GM_SETGRID;
  end;

{ TStringGridCellComboBox }

type
  TOpenCustomGrid = class(TCustomGrid)
  end;
  
procedure TStringGridCellComboBox.Change;
begin
  inherited Change;
  if Visible and (FGrid <> nil) and (ItemIndex >= 0) then
     TOpenCustomGrid(FGrid).SetEditText(TOpenCustomGrid(FGrid).Col, TOpenCustomGrid(FGrid).Row,
                                        Items[ItemIndex]);
end;

procedure TStringGridCellComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (FGrid<>nil) and not (Key in [VK_UP, VK_DOWN]) then
    Fgrid.EditorKeyDown(Self, Key, Shift);
  inherited KeyDown(Key, Shift);
end;

procedure TStringGridCellComboBox.msg_SetValue(var Msg: TGridMessage);
var
  i: Integer;
begin
  for i := 0 to Items.Count-1 do
    if CompareText(Msg.Value, Items[i]) = 0 then begin
      ItemIndex := i;
      Exit;
    end;
  ItemIndex := 0;
end;

procedure TStringGridCellComboBox.msg_GetValue(var Msg: TGridMessage);
begin
  if ItemIndex >= 0 then
    Msg.Value := Items[ItemIndex];
end;

procedure TStringGridCellComboBox.msg_SetGrid(var Msg: TGridMessage);
begin
  FGrid:=Msg.Grid;
  Msg.Options:=EO_AUTOSIZE or EO_SELECTALL;
end;
{$ENDIF}

{ TStringGridEx }

{$IFDEF LCL_}
procedure TStringGridEx.SelectEditor;
var
  aEditor: TWinControl;
  Style: TEditStyle;
begin
  Style := esSimple;
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Style);
  if Style = esSimple then
    inherited
  else begin
    if not Assigned(DropDownEditor) then begin
      FDropDownEditor := CreateDropDownEditor;
      FDropDownEditor.Visible := False;
    end;
    aEditor:= DropDownEditor;
    if (goEditing in Options) and Assigned(OnSelectEditor) then
      OnSelectEditor(Self, Col, Row, aEditor);
    if aEditor<>Editor then Editor:=aEditor;
 end;
end;

function TStringGridEx.CreateDropDownEditor: TComboBox;
begin
  Result := TStringGridCellComboBox.Create(Self);
  Result.Style := csDropDownList;
  Result.Parent := Self.Parent;
end;

procedure TStringGridEx.EditordoSetValue;
var
  Msg: TGridMessage;
  List: TStringList;
  i: Integer;
begin
  if Assigned(Editor) and (Editor = DropDownEditor) then begin
    Msg.MsgID:=GM_SETGRID; { Lockout Changes while list being made up }
    Msg.Grid:=nil;
    Editor.Dispatch(Msg);
    List := TStringList.Create;
    try
      { To avoid crashes, only write to dropdown list if it has changed }
      DoGetPicklistItems(Col, Row, List);
      if DropDownEditor.Items.Count <> List.Count then
        DropDownEditor.Items.Assign(List)
      else
        for i := 0 to List.Count-1 do
          DropDownEditor.Items[i] := List[i]
    finally
      List.Free;
    end;
    Msg.Grid:=Self;
    Editor.Dispatch(Msg);
  end;
  inherited EditordoSetValue;
end;
{$ENDIF}

destructor TStringGridEx.Destroy;
begin
  if Assigned(DropDownEditor) then
    FreeAndNil(FDropDownEditor);
  inherited Destroy;
end;

{$IFDEF VCL_}
function TStringGridEx.CreateEditor: TInplaceEdit;
begin
  Result := TInplaceEditList.Create(Self);
  TInplaceEditList(Result).OnGetPickListitems := DoGetPicklistItems;
end;

function TStringGridEx.GetEditStyle(ACol, ARow: Integer): TEditStyle;
begin
  Result := inherited GetEditStyle(ACol, ARow);
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Result);
end;
{$ENDIF}

procedure TStringGridEx.DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
begin
  if Assigned (OnGetPickListItems) then
    OnGetPickListItems(Col, Row, AItems);
end;

end.

