unit LogFind;

interface

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses
  {$IFNDEF CLX_} { WinProcs, WinTypes, } { DsgnIntf, } Dialogs, Forms,
  {$ELSE} QDialogs, QForms, QGraphics, {$ENDIF}
  {$IFDEF LCL_} SrchStrs, {$ENDIF}
  Classes, LogFiDlg;

type

  TLogFindDialog = class(TComponent)
  private
    FCtl3D: Boolean;
    FHelpContext: THelpContext;
    FOnFind: TNotifyEvent;
    FOptions: TFindOptions;
    FFindText: string;
    FLogFindDlg: TLogFindDlg;
    FOwnDialog: Boolean;
    FLocal: Boolean;
    function GetFindText: string;
    procedure SetFindText(const Value: string);
    function GetOptions: TFindOptions;
    procedure SetOptions(Value: TFindOptions);
    procedure DoDestroy(Sender: Tobject);
    function GetLogFindDlg: TLogFindDlg;
    procedure LoadDialog;
  protected
    procedure SetLocal(const Value: Boolean); virtual;
    procedure Loaded; override;
  public
    procedure DoFind(Sender: Tobject);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Execute: Boolean; virtual;
    procedure CloseDialog;
    property LogFindDlg: TLogFindDlg read GetLogFindDlg;
  published
    property Ctl3D: Boolean read FCtl3D write FCtl3D default True;
    property HelpContext: THelpContext read FHelpContext write FHelpContext default 0;
    property FindText: string read GetFindText write SetFindText;
    property Options: TFindOptions read GetOptions write SetOptions default [frDown];
    property OnFind: TNotifyEvent read FOnFind write FOnFind;
    property OwnDialog:Boolean read FOwnDialog write FOwnDialog default False;
    property Local:Boolean read FLocal write SetLocal default False;
  end;


  TLogReplaceDialog = class(TLogFindDialog)
  private
    FOnReplace: TNotifyEvent;
    FReplaceText: string;
    function GetReplaceText: string;
    procedure SetReplaceText(const Value: string);
    procedure DoReplace(Sender: Tobject);
  protected
    procedure SetLocal(const Value: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    function Execute: Boolean; override;
    destructor Destroy; override;
  published
    property ReplaceText: string read GetReplaceText write SetReplaceText;
    property OnReplace: TNotifyEvent read FOnReplace write FOnReplace;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

var
  LogFindDialog: TLogFindDialog;
  LogReplaceDialog: TLogReplaceDialog;

implementation

{$IFDEF SINGLE}
{$IFDEF VER80}
{$R LogDlg.d16}
{$ELSE}
{$R LogDlg.d32}
{$ENDIF}

procedure Register;
begin
  RegisterComponents('CMW', [TLogFindDialog, TLogReplaceDialog]);
end;

{$ENDIF}

{ TLogFindDialog }

constructor TLogFindDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Ctl3D := True;
  FOptions := [frDown];
end;

destructor TLogFindDialog.Destroy;
begin
  if OwnDialog then
    LogFindDlg.Free;
  SetLocal(False); { Force disconnect }
  inherited Destroy;
end;

function TLogFindDialog.GetFindText: string;
begin
  if LogFindDlg <> nil then
    FFindText := LogFindDlg.FindText.Text;
  Result := FFindText;
end;

procedure TLogFindDialog.SetFindText(const Value: string);
begin
  if LogFindDlg <> nil then
    LogFindDlg.FindText.Text := Value;
  FFindText := Value;
end;

function TLogFindDialog.GetOptions: TFindOptions;
  procedure GetOption(Value: Boolean; Option: TFindOption);
  begin
    if Value then
      FOptions := FOptions + [Option]
    else
      FOptions := FOptions - [Option];
  end;
begin
  if LogFindDlg <> nil then
   with LogFindDlg do begin
    GetOption(Direction.ItemIndex = 1, frDown);
    GetOption(not MatchCase.Visible, frHideMatchCase);
    GetOption(not WholeWord.Visible, frHideWholeWord);
    GetOption(not Direction.Visible, frHideUpDown);
    GetOption(MatchCase.Checked, frMatchCase);
    GetOption(not MatchCase.Enabled, frDisableMatchCase);
    GetOption(not Direction.Enabled, frDisableUpDown);
    GetOption(not WholeWord.Enabled, frDisableWholeWord);
    GetOption(WholeWord.Checked, frWholeWord);
    { GetOption(, frShowHelp); }
  end;
  Result := FOptions;
end;

procedure TLogFindDialog.SetOptions(Value: TFindOptions);
begin
  if LogFindDlg <> nil then
   with LogFindDlg do begin
    MatchCase.Enabled := not (frDisableMatchCase in Value);
    Direction.Enabled := not (frDisableUpDown in Value);
    WholeWord.Enabled := not (frDisableWholeWord in Value);
    MatchCase.Visible := not (frHideMatchCase in Value);
    WholeWord.Visible := not (frHideWholeWord in Value);
    Direction.Visible := not (frHideUpDown in Value);
    if frDown in Value then
      Direction.ItemIndex := 1
    else
      Direction.ItemIndex := 0;
    MatchCase.Checked := frMatchCase in Value;
    WholeWord.Checked := frWholeWord in Value;
    { GetOption(, frShowHelp); }
  end;
  FOptions := Value;
end;

procedure TLogFindDialog.DoFind(Sender: Tobject);
begin
  FOptions := FOptions - [frReplace, frReplaceAll] + [frFindNext];
  if Assigned(FOnFind) then FOnFind(Self);
end;

procedure TLogFindDialog.DoDestroy(Sender: Tobject);
begin
  if OwnDialog then
    FLogFindDlg := nil;
end;

function TLogFindDialog.Execute: Boolean;
begin
  LoadDialog;
  LogFindDlg.Show;
  {$IFDEF LCL_}
  with LogFindDlg do begin
    { TODO: Exposes lcl/gtk bug }
    FindText.SelectAll;
    if IsReplace then
      ReplaceText.SelectAll;
  end;
  {$ENDIF}
  Result := False; { Irrelevant }
end;

procedure TLogFindDialog.Loaded;
begin
  inherited Loaded;
  if not OwnDialog and not (csDesigning in ComponentState) and (LogFindDlg = nil) then
    LoadDialog;
  SetLocal(Local); { Force Load }
end;

procedure TLogFindDialog.CloseDialog;
begin
  if LogFindDlg <> nil then
    LogFindDlg.Close;
end;

function TLogFindDialog.GetLogFindDlg: TLogFindDlg;
begin
  if OwnDialog then
    Result := LogFindDlg
  else
    Result := Logfidlg.LogFindDlg;
end;

procedure TLogFindDialog.LoadDialog;
begin
  if LogFindDlg = nil then begin
    if OwnDialog then
      FLogFindDlg := TLogFindDlg.Create(Application)
    else begin
      Logfidlg.LogFindDlg := TLogFindDlg.Create(Application);
    end;
    {$IFDEF VCL_}
    LogFindDlg.Ctl3D := Ctl3D;
    {$ENDIF}
    SetOptions(FOptions);
    SetFindText(FFindText);
    if OwnDialog then
      LogFindDlg.OnDestroy := {$IFDEF FPC}@{$ENDIF}DoDestroy;
    {$IFDEF OLDPOS}
    LogFindDlg.Top := Screen.Height; { Force nice position }
    {$ENDIF}
  end;
  LogFindDlg.IsReplace := (Self is TLogReplaceDialog);
  with LogFindDlg do begin
    OnFindNext := {$IFDEF FPC}@{$ENDIF}DoFind;
    {$IFNDEF OLDPOS}
    if Assigned(Application.MainForm) then begin
      Left:= Application.MainForm.Width + Application.MainForm.Left - Width;
      Top:= Application.MainForm.Height + Application.MainForm.Top;
    end;
    {$ENDIF}
    if Top > Screen.Height - Height div 2 then
      Top := Screen.Height - Height * 12 div 10;
    if Left > Screen.Width - Width div 2 then
      Left := Screen.Width - Width ;
    {$IFNDEF LCL_}
    { TODO: Exposes lcl/gtk bug }
    FindText.SelectAll;
    if IsReplace then
      ReplaceText.SelectAll;
    {$ENDIF}
    ActiveControl := FindText; { Set in xfm/dfm, but doesn't seem to 'stick' }
  end;
end;

procedure TLogFindDialog.SetLocal(const Value: Boolean);
begin
  FLocal := Value;
  if Local then begin
    if LogFindDialog = Self then
      LogFindDialog := nil;
  end else if LogFindDialog = nil then
    LogFindDialog := Self;
end;

{ TLogReplaceDialog }

constructor TLogReplaceDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

function TLogReplaceDialog.Execute: Boolean;
begin
  if LogFindDlg = nil then begin
    Result := inherited Execute;
    SetReplaceText(FReplaceText);
  end else
    Result := inherited Execute;
  LogFindDlg.OnReplace := {$IFDEF FPC}@{$ENDIF}DoReplace;
end;

procedure TLogReplaceDialog.DoReplace(Sender: Tobject);
begin
  FOptions := FOptions - [frReplace, frReplaceAll, frFindNext] + [LogFindDlg.LastOption];
  if Assigned(FOnReplace) then FOnReplace(Self);
end;

function TLogReplaceDialog.GetReplaceText: string;
begin
  if LogFindDlg <> nil then
    FReplaceText := LogFindDlg.ReplaceText.Text;
  Result := FReplaceText;
end;

procedure TLogReplaceDialog.SetReplaceText(const Value: string);
begin
  if LogFindDlg <> nil then
    LogFindDlg.ReplaceText.Text := Value;
  FReplaceText := Value;
end;

destructor TLogReplaceDialog.Destroy;
begin
  if Self = LogReplaceDialog then
    LogReplaceDialog := nil;
  inherited Destroy;
end;

procedure TLogReplaceDialog.SetLocal(const Value: Boolean);
begin
  FLocal := Value;
  if Local then begin
    if LogReplaceDialog = Self then
      LogReplaceDialog := nil;
  end else if LogReplaceDialog = nil then
    LogReplaceDialog := Self;
end;

end.
