unit NGrid;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs, Menus, StdCtrls, Grids,
     {$ELSE} Types, QGraphics, QControls, QForms, QDialogs, QMenus, QStdCtrls, QGrids, QTypes, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Classes, Huge, MathSup, MatrixVec, TextStrm;

type
  TOnChangedEvent = procedure (Sender: TObject; Col, Row: Longint) of object;
  TOnCellEvent = procedure (Sender: TObject; Col, Row: Longint; var v:Double; var Valid:Boolean) of object;
  TChangeCellQueryEvent = procedure (Sender: TObject; Col, Row: Longint; var CanChange: Boolean) of object;
  TOnGetCellLabelEvent = procedure (Sender: TObject; Col, Row: Longint; var s:string; var IsText:Boolean) of object;
  {$IFDEF LCL_}
  TMovedEvent = procedure (Sender: TObject; FromIndex, ToIndex: Longint) of object;
  TDrawCellEvent = TOnDrawCell;
  TSelectCellEvent = TOnSelectCellEvent;
  {$ENDIF}

  TNumberGrid = class(TCustomGrid)
  private
    FMatrix: TMatrix;
    InAreaSelection:(asNone, asCol, asRow, asAll);
    FOnColumnMoved: TMovedEvent;
    FOnDrawCell: TDrawCellEvent;
    {$IFNDEF LCL_}
    FOnGetEditMask: TGetEditEvent;
    {$ENDIF}
    FOnGetEditText: TGetEditEvent;
    FOnRowMoved: TMovedEvent;
    FOnSelectCell: TSelectCellEvent;
    FOnGetCell: TOnCellEvent;
    FOnGetCellLabel: TOnGetCellLabelEvent;
    FOnSetCell: TOnCellEvent;
    FOnSetEditText: TSetEditEvent;
    FOnTopLeftChanged: TNotifyEvent;
    FOnChangeCellQuery: TChangeCellQueryEvent;
    FTitles: TStringList;
    FLbls: TStringList;
    FOnChanged: TOnChangedEvent;
    FStatusWindow: TControl;
    LastCol, LastRow:Longint; { Last Cell for error entry }
    LastString:string;
    SelectingAnchor:TGridCoord;
    SavedOptions:TGridOptions;
    OriginalText:string;
    OriginalModified:Boolean;
    InhibitNextInvalidate: Boolean;
    procedure TrackSelectedArea(X, Y: Integer);
    {$IFDEF VCL_} { TODO: WINLCL also? }
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    {$ENDIF}
    function GetTextN: TCaption;
    procedure SetTextN(const Value: TCaption);
  protected
    FUpdateCount: Integer;
    CurrentZeroWidth: Integer;
    EndPending: Boolean;
    procedure CheckSizes; virtual;
    {$IFNDEF LCL_}
    function CreateEditor: TInplaceEdit; override;
    function GetEditMask(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF}; override;
    {$ELSE}
    procedure SelectEditor; override;
    {$ENDIF}
    procedure ColumnMoved(FromIndex, ToIndex: Longint); {$IFNDEF LCL_} override; {$ENDIF}
    procedure RowMoved(FromIndex, ToIndex: Longint); {$IFNDEF LCL_} override; {$ENDIF}
    {$IFDEF LCL_}
    procedure ColRowMoved(IsColumn: Boolean; FromIndex,ToIndex: Integer); override;
    procedure EditorDoGetValue; override;
    procedure EditorDoSetValue; override;
    procedure FocusCell(ACol, ARow: Longint; UpdateSelection: Boolean);
    procedure DrawFocusRect(aCol,aRow: Integer; ARect: TRect); override;
    procedure DoCopyToClipboard; override;
    procedure DoCutToClipboard; override;
    procedure DoPasteFromClipboard; override;
    {$ENDIF}
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    function GetEditText(ACol, ARow: Longint): {$IFNDEF CLX_} string; {$ELSE} WideString; {$ENDIF} override;
    function CanEditShow: Boolean; override;
    function CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean; override;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
    procedure SetEditText(ACol, ARow: Longint; const s: {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF}); override;
    procedure TopLeftChanged; override;
    procedure SetTitles(l:TStringList);
    function GetTitle(i:Longint):string;
    procedure SetTitle(i:Longint; s:string);
    procedure SetLbls(l:TStringList);
    function GetLbl(i:Longint):string;
    procedure SetLbl(i:Longint; s:string);
    function GetValid(ARow, ACol:Longint):Boolean;
    procedure SetValid(ARow, ACol:Longint; IsValid:Boolean);
    function GetValue(ARow, ACol:Longint):Double;
    procedure SetValue(ARow, ACol:Longint; const v:Double);
    procedure SizeChanged(OldColCount, OldRowCount: Longint); override;
    procedure Changed(ACol, ARow: Longint); virtual;
    procedure ShowStatus(ACol, ARow: Longint);
    procedure SetCell(ARow, ACol: Longint; v:Double; IsValid:Boolean); virtual;
    function GetPointer(ARow, ACol:Longint):Pointer;
    procedure DefaultDrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
    property Item[ARow, ACol: Longint]:Pointer read GetPointer;
    function CanChange(ARow, ACol:Longint):Boolean; virtual;
  public
    Modified:Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean = False);
    procedure Blank;
    function DataExtent:TGridRect;
    function NaturalWidth: Integer;
    procedure SaveAreaToStream(Stream:TStream; Area:TGridRect);
    procedure SaveToStream(Stream:TStream);
    procedure OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint);
    procedure Loaded; override;
    procedure LoadFromStream(Stream:TStream);
    procedure ClearArea(Area:TGridRect);
    procedure ClearColumn(ACol: Longint);
    procedure CancelSelection;
    procedure Clear;
    procedure SelectAll;
    procedure MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetFocus; override;
    function GetCell(ARow, ACol: Longint; var v:Double):Boolean; virtual;
    function GetCellLabel(ARow, ACol: Longint; var s: string):Boolean;
    procedure ExchangeRows(i, j: Longint);
    procedure Update; override;
    property Canvas;
    property Col;
    property ColWidths;
    property EditorMode;
    property GridHeight;
    property GridWidth;
    property LeftCol;
    property Selection;
    property Row;
    property RowHeights;
    {$IFNDEF LCL_}
    property TabStops;
    {$ENDIF}
    property TopRow;

    property Title[i:LongInt]:string read GetTitle write SetTitle;
    property Lbl[i:LongInt]:string read GetLbl write SetLbl;
    property Valid[ARow, ACol:Longint]:Boolean read GetValid write SetValid;
    property Value[ARow, ACol:Longint]:Double read GetValue write SetValue; default;
    procedure Invalidate; override;
    procedure InvalidateCol(ACol: Longint);
    procedure BeginUpdate;
    procedure EndUpdate;
  published
    procedure ClearSelection;
    procedure CopyToClipboard;
    procedure PasteFromClipboard;
    function CanCopyToClipboard: Boolean;
    function CanPasteFromClipboard: Boolean;

    property Align;
    {$IFDEF DFS_COMPILER_4_UP}
    property Anchors;
    {$ENDIF}
    {$IFNDEF LCL_}
    property BorderStyle;
    {$ENDIF}
    property Caption read GetTextN write SetTextN;
    property Color;
    property ColCount;
    {$IFDEF MSWINDOWS}
    property Ctl3D;
    property DragCursor;
    property ParentCtl3D;
    {$ENDIF}
    property DefaultColWidth;
    property DefaultRowHeight;
    property DefaultDrawing;
    property DragMode;
    property Enabled;
    property FixedColor;
    property FixedCols;
    property FixedRows;
    property Font;
    property GridLineWidth;
    property Options;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RowCount;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property VisibleColCount;
    property VisibleRowCount;
    property OnClick;
    property OnColumnMoved: TMovedEvent read FOnColumnMoved write FOnColumnMoved;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawCell: TDrawCellEvent read FOnDrawCell write FOnDrawCell;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    {$IFNDEF LCL_}
    property OnGetEditMask: TGetEditEvent read FOnGetEditMask write FOnGetEditMask;
    {$ENDIF}
    property OnGetEditText: TGetEditEvent read FOnGetEditText write FOnGetEditText;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnRowMoved: TMovedEvent read FOnRowMoved write FOnRowMoved;
    property OnSelectCell: TSelectCellEvent read FOnSelectCell write FOnSelectCell;
    property OnSetEditText: TSetEditEvent read FOnSetEditText write FOnSetEditText;
    property OnTopLeftChanged: TNotifyEvent read FOnTopLeftChanged write FOnTopLeftChanged;
    property OnGetCellLabel: TOnGetCellLabelEvent read FOnGetCellLabel write FOnGetCellLabel;

    property OnChanged: TOnChangedEvent read FOnChanged write FOnChanged;
    property Titles:TStringList read FTitles write SetTitles;
    property Lbls:TStringList read FLbls write SetLbls;
    property StatusWindow: TControl read FStatusWindow write FStatusWindow;
    property OnGetCell: TOnCellEvent read FOnGetCell write FOnGetCell;
    property OnSetCell: TOnCellEvent read FOnSetCell write FOnSetCell;
    property OnChangeCellQuery: TChangeCellQueryEvent read FOnChangeCellQuery write FOnChangeCellQuery;
  end;

{ Helper prcedure for filling grid }
procedure TextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Alignment: TAlignment = taLeftJustify);
procedure ColouredTextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Colour: TColor; Alignment: TAlignment = taLeftJustify);

implementation

uses {$IFNDEF CLX_} ClipBrd, {$ELSE} Qt, QClipBrd, {$ENDIF}
     {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF}
     HugeClip, Math, Vector, WCompat, UCompat;

{$IFDEF SINGLE}
{$IFDEF WIN32}
{$R NGrid.d32}
{$ELSE}
{$R NGrid.d16}
{$ENDIF}
{$ENDIF}

type

  TNGInplaceEdit = class({$IFDEF LCL_}TStringCellEditor{$ELSE}TInplaceEdit{$ENDIF})
  private
    Grid:TNumberGrid;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor CreateWithGrid(AOwner:TNumberGrid);
    property ReadOnly;
  published
    procedure PasteFromClipboard; { Implementation relies on this being a non-virtual function }
  end;


constructor TNumberGrid.Create(AOwner: TComponent);
begin
  FTitles := TStringList.Create;
  FLbls := TStringList.Create;
  inherited Create(AOwner);
  {$IFDEF LCL_}
    ExtendedSelect := True;
  {$ENDIF}
  LastCol := -1;
  CheckSizes;
end;

destructor TNumberGrid.Destroy;
begin
  FMatrix.Free;
  FTitles.Free;
  FLbls.Free;
  inherited Destroy;
end;

procedure TNumberGrid.Loaded;
begin
  inherited Loaded;
  CheckSizes;
end;

procedure TNumberGrid.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TNumberGrid.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then Invalidate;
end;


procedure TNumberGrid.InvalidateCol(ACol: Longint);
begin
  if FUpdateCount = 0 then
    Invalidate;
{ inherited InvalidateCol(ACol) is broken in WIN32 and does not exist in DELPHI 1 }
end;

procedure TNumberGrid.Changed(ACol, ARow: Longint);
begin
  if FUpdateCount = 0 then
    if ACol < 0 then
      if ARow < 0 then
        Invalidate
      else
        InvalidateRow(ARow)
    else if ARow < 0 then
      InvalidateCol(ACol)
    else
      InvalidateCell(ACol, ARow);
  if (ARow = Row) and (ACol = Col) then ShowStatus(ACol, ARow);
  if Assigned(FOnChanged) and not (csDesigning in ComponentState) then
    FOnChanged(Self, ACol, ARow);
end;

procedure TNumberGrid.Update;
begin
  CheckSizes;
  inherited Update;
end;

procedure TNumberGrid.CheckSizes;
var
  j: Longint;
  AChange: Boolean;
begin
  AChange := not Assigned(FMatrix)
             or (FMatrix.RowCount <> RowCount - FixedRows)
             or (FMatrix.ColCount <> ColCount - FixedCols);
  if not Assigned(FMatrix) then
    FMatrix := TMatrix.Create(RowCount-FixedRows, ColCount-FixedCols)
  else begin
    FMatrix.RowCount := RowCount - FixedRows;
    FMatrix.ColCount := ColCount - FixedCols;
  end;
  if Titles.Count >= ColCount then begin
    for j := Titles.Count-1 downto ColCount-1 do
      Titles.Delete(j);
    AChange := True;
  end;
  if Lbls.Count >= RowCount then begin
    for j := Lbls.Count-1 downto RowCount-1 do
      Lbls.Delete(j);
    AChange := True;
  end;
  if AChange then
    Invalidate;
end;

procedure TNumberGrid.SizeChanged(OldColCount, OldRowCount: Longint);
begin
  inherited SizeChanged(OldColCount, OldRowCount);
  CheckSizes;
end;

procedure TNumberGrid.CancelSelection;
var
  s:TGridRect;
begin
  s.Left := Col; s.Right := Col;
  s.Bottom := Row; s.Top := Row;
  Selection := s;
end;

constructor TNGInPlaceEdit.CreateWithGrid(AOwner:TNumberGrid);
begin
  Create(AOwner);
  Grid := AOwner;
end;

{$IFDEF LCL_}
procedure TNumberGrid.SelectEditor;
begin
  if Editor = nil then
    Editor := TNGInplaceEdit.CreateWithGrid(Self);
  // inherited SelectEditor;
end;
{$ELSE}
function TNumberGrid.CreateEditor: TInplaceEdit;
begin
  Result := TNGInplaceEdit.CreateWithGrid(Self);
end;

function TNumberGrid.GetEditMask(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF};
begin
  Result := '';
  if Assigned(FOnGetEditMask) then FOnGetEditMask(Self, ACol, ARow, Result);
end;
{$ENDIF}

procedure TNGInplaceEdit.PasteFromClipboard;
begin
  Grid.PasteFromClipboard; { Relies on PasteFromClipboard being a non-virtual function }
end;

procedure TNGInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE:
      begin
        Text := Grid.OriginalText;
        Grid.EditorMode := False;
        Grid.Modified := Grid.OriginalModified;
        Key := 0;
      end;
    VK_LEFT:
      if (SelLength = 0) and (SelStart = 0) then begin
        Grid.KeyDown(Key, Shift);
        Key := 0;
      end;
    VK_RIGHT:
      if SelStart = Length(Text) then begin
        Grid.KeyDown(Key, Shift);
        Key := 0;
      end;
  else
    inherited KeyDown(Key, Shift);
  end;
end;

{$IFDEF LCL_}
procedure TNumberGrid.ColRowMoved(IsColumn: Boolean; FromIndex, ToIndex: Integer);
begin
  inherited ColRowMoved(IsColumn, FromIndex, ToIndex);
  if IsColumn then
    ColumnMoved(FromIndex, ToIndex)
  else
    RowMoved(FromIndex, ToIndex)
end;
{$ENDIF}

procedure TNumberGrid.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  FMatrix.Move(FromIndex, ToIndex);
  if Assigned(FOnColumnMoved) then FOnColumnMoved(Self, FromIndex, ToIndex);
  Changed(-1, -1);
end;

procedure TNumberGrid.RowMoved(FromIndex, ToIndex: Longint);
var
  c: Longint;
  v: Double;
begin
  {$IFDEF MAKEFROMCOLVECTS}
  for c := 1 to ColCount-1 do
    FMatrix.Column[c].Move(FromIndex, ToIndex);
  {$ELSE}
  for c := 1 to ColCount-1 do
    if FMatrix.GetCell(FromIndex, c, v) then
      FMatrix.Value[ToIndex, c] := v
    else
      FMatrix.Valid[ToIndex, c] := False;
  {$ENDIF}
  if Assigned(FOnRowMoved) then FOnRowMoved(Self, FromIndex, ToIndex);
  Changed(-1, -1);
end;

{$IFDEF LCL_}
procedure TNumberGrid.EditorDoGetValue;
begin
  if Assigned(Editor) then
    SetEditText(Col, Row, Editor.Text);
end;

procedure TNumberGrid.EditorDoSetValue;
begin
  if Assigned(Editor) then
    Editor.Text := GetEditText(Col, Row);
end;

procedure TNumberGrid.FocusCell(ACol, ARow: Longint; UpdateSelection: Boolean);
begin
  if not UpdateSelection then
    SelectActive := False;
  MoveExtend(False, ACol, ARow);
end;

procedure TNumberGrid.DrawFocusRect(aCol, aRow: Integer; ARect: TRect);
begin
  inherited DrawFocusRect(aCol, aRow, ARect);
  // Copied from TDrawGrid
  // Draw focused cell if we have the focus
  if Self.Focused Or ((goEditing in Options)and(goAlwaysShowEditor in Options) and ((Editor=nil) or not Editor.Focused)) then
  begin
    Canvas.Pen.Color:=FocusColor;
    Canvas.Pen.Style:=psDot;
    if goRowSelect in Options then begin
      Canvas.MoveTo(GCache.FixedWidth+1, aRect.Top);
      Canvas.LineTo(GCache.MaxClientXY.x-2, aRect.Top);
      Canvas.LineTo(GCache.MaxClientXY.x-2, aRect.Bottom-2);
      Canvas.LineTo(GCache.FixedWidth+1, aRect.Bottom-2);
      Canvas.LineTo(GCache.FixedWidth+1, aRect.Top+1);
    end else begin
      Canvas.MoveTo(aRect.Left, aRect.Top);
      Canvas.LineTo(ARect.Right-2,aRect.Top);
      Canvas.LineTo(aRect.Right-2,aRect.bottom-2);
      Canvas.LineTo(aRect.Left, aRect.Bottom-2);
      Canvas.Lineto(aRect.left, aRect.top+1);
    end;
    Canvas.Pen.Style:=psSolid;
  end;
end;

procedure TNumberGrid.DoCopyToClipboard;
begin
  CopyToClipboard;
end;

procedure TNumberGrid.DoCutToClipboard;
begin
  CopyToClipboard;
  ClearSelection;
end;

procedure TNumberGrid.DoPasteFromClipboard;
begin
  PasteFromClipboard;
end;

{$ENDIF}

procedure TNumberGrid.SetEditText(ACol, ARow: Longint; const s: {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF});
var
  AValue: Double;
begin
  LastCol := -1;
  if Length(s) <> 0 then
    if TryStrToFloat(Trim(s), AValue) then
      SetCell(ARow, ACol, AValue, True)
    else begin
      SetValid(ARow, ACol, False);
      LastCol := ACol; { Save text, to allow user to edit it }
      LastRow := ARow;
      LastString := s;
    end
  else
    SetValid(ARow, ACol, False);
  if Assigned(FOnSetEditText) then FOnSetEditText(Self, ACol, ARow, s);
end;

function TNumberGrid.GetEditText(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF};
var
  v: Double;
  AEditor: TNGInplaceEdit;
begin
  Result := '';
  if EditorMode and ((Selection.Left <> ACol) or (Selection.Right <> ACol)
                      or (Selection.Top <> ARow) or (Selection.Bottom <> ARow)) then
    CancelSelection; { If typing, then only one cell should be selected }
  if (LastCol <> ACol) or (LastRow <> ARow) then begin
    if (LastCol >= 0) then begin
      InvalidateCell(LastCol, LastRow);
      LastCol := -1;
    end;
    if GetCell(ARow, ACol, v) then
      Result := FloatToStr(v);
    AEditor := InplaceEditor as TNGInplaceEdit;
    if AEditor <> nil then AEditor.ReadOnly := not CanChange(ARow, ACol);
  end else
    Result := LastString;
  if Assigned(FOnGetEditText) then FOnGetEditText(Self, ACol, ARow, Result);
  OriginalText := Result;
  OriginalModified := Modified;
end;

function TNumberGrid.CanEditShow: Boolean;
begin
  Result := inherited CanEditShow;
  if Result and Assigned(FOnChangeCellQuery) then begin
    FOnChangeCellQuery(Self, Col, Row, Result);
  end;
  if not Result then EditorMode := False;
end;

{$IFDEF UNIX}
type TOpenControl = class(TControl)
  property Caption;
  end;
{$ENDIF}

procedure TNumberGrid.ShowStatus(ACol, ARow: Longint);
var
  v: Double;
  Buffer: String;
begin
  if Assigned(FStatusWindow) then begin
    if not GetCellLabel(ARow, ACol, Buffer) then
      if GetCell(ARow, ACol, v) then
        Buffer := FloatToStr(v)
      else
        Buffer := '';
{$IFDEF MSWINDOWS}
{$IFDEF WIN32}
    FStatusWindow.SetTextBuf(PChar(Buffer));
{$ELSE}
    Buffer := Buffer + #0;
    FStatusWindow.SetTextBuf(@Buffer[1]);
{$ENDIF}
{$ENDIF}
{$IFDEF UNIX}
    TOpenControl(FStatusWindow).Caption := Buffer; // TODO: Check
{$ENDIF}
  end;
end;

function TNumberGrid.CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean;
var
  r, nr, c, nc: Integer;
  FirstCellSet, FirstTime: Boolean;
  v: Double;
begin
  Result := False;
  if Key = VK_END then begin
    if ssCtrl in Shift then begin
      with DataExtent.BottomRight do
        FocusCell(X, Y, not (ssShift in Shift));
    end else
      EndPending := True;
      Exit;
  end else if Key = VK_DELETE then
    ClearSelection
  //end else if Shift*[ssShift, ssAlt, ssCtrl] = ssCtrl then
    //case Key of
      //Ord('C'): CopyToClipboard;
      //Ord('V'): PasteFromClipboard;
      //Ord('X'): begin CopyToClipboard; ClearSelection; end;
    //else
      //Result := True;
  else if EndPending and (Key in [VK_DOWN, VK_UP, VK_LEFT, VK_RIGHT]) then begin
    { Find target cell }
    c := Col;
    r := Row;
    FirstCellSet := GetCell(r, c, v);
    FirstTime := True;
    repeat
      nc := c;
      nr := r;
      case Key of
        VK_DOWN:  if r >= RowCount-1 then Break else Inc(nr);
        VK_UP:    if r <= FixedRows then Break else Dec(nr);
        VK_LEFT:  if c <= FixedCols then Break else Dec(nc);
        VK_RIGHT: if c >= ColCount-1 then Break else Inc(nc);
      end;
      if GetCell(nr, nc, v) xor FirstCellSet then
        if FirstTime then
          FirstCellSet := not FirstCellSet
        else
          Break;
      FirstTime := False;
      c := nc;
      r := nr;
    until False;
    FocusCell(c, r, not (ssShift in Shift));
  end else
    Result := True;
  EndPending := False;
end;

function TNumberGrid.SelectCell(ACol, ARow: Longint): Boolean;
begin
  Result := True;
  if Assigned(FOnSelectCell) then FOnSelectCell(Self, ACol, ARow, Result);
  if Result then
    ShowStatus(ACol, ARow);
end;

procedure TNumberGrid.DrawCell(ACol, ARow: Longint; ARect: TRect;
  AState: TGridDrawState);
begin
  {$IFDEF LCL_}
  inherited DrawCell(ACol, ARow, ARect, AState);
  {$ENDIF}
  if DefaultDrawing then DefaultDrawCell(ACol, ARow, ARect, AState);
  if Assigned(FOnDrawCell) and not (csDesigning in ComponentState) then FOnDrawCell(Self, ACol, ARow, ARect, AState);
end;

procedure TNumberGrid.TopLeftChanged;
begin
  inherited TopLeftChanged;
  if Assigned(FOnTopLeftChanged) then FOnTopLeftChanged(Self);
end;

procedure TNumberGrid.SetTitles(l:TStringList);
begin
  FTitles.Assign(l);
  Changed(-1, 0);
end;

function TNumberGrid.GetTitle(i:Longint):string;
begin
  if (i > 0) and (i <= FTitles.Count) then
    Result := FTitles[i-1]
  else if i = 0 then
    Result := Caption
  else
    Result := '';
end;

procedure TNumberGrid.SetTitle(i:Longint; s:string);
begin
  if (i > FTitles.Count) and (i <= ColCount-1) then
    while FTitles.Count < i do FTitles.Add('');
  if (i > 0) and (i <= FTitles.Count) then
    FTitles[i-1] := s
  else if i = 0 then
    Caption := s;
  Changed(i, 0);
end;

procedure TNumberGrid.SetLbls(l:TStringList);
begin
  FLbls.Assign(l);
  Changed(0, -1);
end;

function TNumberGrid.GetLbl(i:Longint):string;
begin
  if (i > 0) and (i <= Lbls.Count) then
    Result := Lbls[i-1]
  else if i = 0 then
    Result := Caption
  else
    Result := '';
end;

procedure TNumberGrid.SetLbl(i:Longint; s:string);
begin
  if (i > Lbls.Count) and (i <= RowCount-1) then
    while Lbls.Count < i do Lbls.Add('');
  if (i > 0) and (i <= Lbls.Count) then
    Lbls[i-1] := s
  else if i = 0 then
    Caption := s;
  Changed(0, i);
end;

function TNumberGrid.GetTextN: TCaption;
begin
  Result := inherited Caption;
end;

procedure TNumberGrid.SetTextN(const Value: TCaption);
begin
  inherited Caption := Value;
  Changed(0, 0);
end;

procedure TNumberGrid.AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean);
var
  W, r: Integer;
  s: string;
begin
  r := 0;
  W := 0;
  { Autosize on column and row headings only }
  while (r < RowCount) and ( (r < FixedRows) or (aCol < FixedCols) ) do begin
    if GetCellLabel(r, aCol, s) then
      W := Max(Canvas.TextWidth(s), W);
    Inc(r);
  end;
  if W = 0 then
    W := DefaultColWidth
  else
    Inc(W, 4);
  if not GrowOnly or (ColWidths[aCol] < W) then
    ColWidths[aCol] := W;
end;

{$IFDEF VCL_}
{ // TODO: Linux? }
procedure TNumberGrid.CMFontChanged(var Message: TMessage);
begin
  inherited;
  CurrentZeroWidth := 0; { Flag that actual value needs to be determined }
end;
{$ENDIF}

function TNumberGrid.GetCellLabel(ARow, ACol: Longint;  var s: string): Boolean;
begin
  Result := True;
  if (ARow = LastRow) and (ACol = LastCol) then
    s := 'Bad Value'
  else if (ARow < FixedRows) or (ARow < FixedRows) then begin
    if ARow >= FixedRows - 1 then
      s := Title[ACol]
    else
      s := '';
    if (Length(s) = 0) and (ACol <> 0) then
      s := Char(Ord('A')+Acol-1);
  end else if ACol < FixedCols then begin
    if (ACol = FixedCols-1) and (ARow > 0) and (ARow <= Lbls.Count) then
      s := Lbls[ARow-1]
    else
      s := IntToStr(ARow);
  end else
    Result := False;
  if Assigned(FOnGetCellLabel) and not (csDesigning in ComponentState) then
    FOnGetCellLabel(Self, ACol, ARow, s, Result);
end;

{ Helper procedures for filling grid }
procedure TextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Alignment: TAlignment);
var
{$IFDEF LCL_}
  TextStyle: TTextStyle;
{$ENDIF}
{$IFDEF VCL_}
  uFormat: UINT;
  x: Integer;
{$ENDIF}
begin
  {$IFDEF VCL_}
  uFormat := GetTextAlign(Canvas.Handle);
  case Alignment of
    taLeftJustify:  x := ARect.Left + 2;
    taRightJustify: begin x := ARect.Right - 2; SetTextAlign(Canvas.Handle, uFormat or TA_RIGHT); end;
    taCenter: begin x := (ARect.Right+ARect.Left) div 2; SetTextAlign(Canvas.Handle, uFormat or TA_CENTER); end;
  end;
  ExtTextOut(Canvas.Handle, x, ARect.Top + 2,
             ETO_CLIPPED or ETO_OPAQUE, @ARect, AddressOfFirstChar(Text), Length(Text), nil);
  if Alignment <> taLeftJustify then
    SetTextAlign(Canvas.Handle, uFormat);
  {$ENDIF}
  {$IFDEF LCL_}
  TextStyle := Canvas.TextStyle;
  TextStyle.Alignment := taLeftJustify;
  TextStyle.Layout := tlTop;
  TextStyle.Clipping := True;
  TextStyle.Opaque := True;
  TextStyle.WordBreak := False;
  TextStyle.SingleLine := True;
  Dec(ARect.Right, 2);
  Dec(ARect.Bottom, 2);
  Canvas.TextRect(ARect, ARect.Left + 2, ARect.Top + 2, Text, TextStyle);
  {$ENDIF}
  {$IFDEF CLX_}
  Canvas.FillRect(ARect);
  Canvas.TextRect(ARect, ARect.Left + 2, ARect.Top + 2, Text,
                  Integer(AlignmentFlags_AlignLeft) or Integer(AlignmentFlags_AlignTop));
  {$ENDIF}
end;

procedure ColouredTextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Colour: TColor; Alignment: TAlignment);
var
  {$IFDEF VCL_}
  Current: LongInt;
  {$ENDIF}
  {$IFDEF LCL_}
  OldColour: TColor;
  OldStyle: TBrushStyle;
  {$ENDIF}
  {$IFDEF CLX_}
  OldColour: TColor;
  OldStyle: TBrushStyle;
  {$ENDIF}
begin
  {$IFDEF VCL_}
  Current := GetBkColor(Canvas.Handle);
  SetBkColor(Canvas.Handle, clRed);
  TextInGrid(Canvas, ARect, Text);
  SetBkColor(Canvas.Handle, Current);
  {$ENDIF}
  {$IFDEF LCL_}
  OldColour := Canvas.Brush.Color;
  OldStyle := Canvas.Brush.Style;
  Canvas.Brush.Color := clRed;
  Canvas.Brush.Style := bsSolid;
  TextInGrid(Canvas, ARect, Text);
  Canvas.Brush.Color := OldColour;
  Canvas.Brush.Style := OldStyle;
  {$ENDIF}
  {$IFDEF CLX_}
  OldColour := Brush.Color;
  OldStyle := Brush.Style;
  Brush.Color := clRed;
  Brush.Style := bsSolid;
  TextInGrid(Canvas, ARect, Text);
  Brush.Color := OldColour;
  Brush.Style := OldStyle;
  {$ENDIF}
end;

procedure TNumberGrid.DefaultDrawCell(ACol, ARow: Longint; ARect: TRect;
  AState: TGridDrawState);
var
  buffer: string;
  nchar: Integer;
  v: Double;
  ZeroSize: TSize;
begin
  if (ARow = LastRow) and (ACol = LastCol) then begin
    ColouredTextInGrid(Canvas, ARect, LastString, clRed);
    Exit;
  end
  else if not GetCellLabel(ARow, ACol, buffer) then
    if GetCell(ARow, ACol, v) then begin
      if CurrentZeroWidth = 0 then begin
        {$IFDEF VCL_}
        GetTextExtentPoint(Canvas.Handle, AddressOfFirstChar('0123456789'), 10, ZeroSize);
        {$ELSE}
        ZeroSize := Canvas.TextExtent('0123456789');
        {$ENDIF}
        CurrentZeroWidth := ZeroSize.cx div 10;
      end;
      nchar := (ARect.Right - ARect.Left - 4) div CurrentZeroWidth;
      buffer := DtosM(v, nchar, -1);
    end else
      buffer := '';
  TextInGrid(Canvas, ARect, buffer);
end;

function TNumberGrid.GetPointer(ARow, ACol:Longint):Pointer;
begin
  Result := FMatrix.Item[ARow-FixedRows+1, ACol-FixedCols+1];
end;

function TNumberGrid.CanChange(ARow, ACol:Longint):Boolean;
begin
  Result := (ARow >= FixedRows) and (ACol >= FixedCols);
end;

function TNumberGrid.GetCell(ARow, ACol:Longint; var v:Double):Boolean;
var
  p: ^Double;
begin
  p := Item[ARow, ACol];
  Result := not IsNanOrInf(p^);
  Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
  if Assigned(FOnGetCell) then FOnGetCell(Self, ACol, ARow, v, Result);
end;

procedure TNumberGrid.SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
var
  p: ^Double;
  OldV: Double;
  OldIsValid, Change: Boolean;
begin
  if not CanChange(ARow, ACol) then Exit;
  p := Item[ARow, ACol];
  OldIsValid := not IsNanOrInf(p^);
  if OldIsValid then
    OldV := p^;
  if Assigned(FOnSetCell) then FOnSetCell(Self, ACol, ARow, v, IsValid);
  { Note: To Set a specific invalid value, set IsValid := True
    and v to the required value }
  if IsValid then
    Move(v, p^, SizeOf(Double))
  else
    SetNan(p^);
  IsValid := not IsNanOrInf(p^);  { Check for a specific invalid value }
  Change := OldIsValid <> IsValid;
  if (LastCol = ACol) and (LastRow = ARow) then begin
    Change := True; { Turn off error indication }
    LastCol := -1;
  end;
  if not Change and IsValid then
    Change := v <> OldV;
  if Change then begin
    Modified := True;
    Changed(ACol, ARow);
  end;
end;

function TNumberGrid.GetValid(ARow, ACol:Longint):Boolean;
var
  v: Double;
begin
  Result := GetCell(ARow, ACol, v);
end;

procedure TNumberGrid.SetValid(ARow, ACol:Longint; IsValid:Boolean);
begin
  SetCell(ARow, ACol, 0, IsValid);
end;

function TNumberGrid.GetValue(ARow, ACol:Longint):Double;
begin
  GetCell(ARow, ACol, Result);
end;

procedure TNumberGrid.SetValue(ARow, ACol:Longint; const v:Double);
begin;
  SetCell(ARow, ACol, v, True);
end;

procedure TNumberGrid.ExchangeRows(i, j: Longint);
var
  ACol: Longint;
  v1, v2: Double;
begin
  for ACol := 1 to ColCount - 1 do
    if GetCell(i, ACol, v1) then
      if GetCell(j, ACol, v2) then begin
        Value[i, ACol] := v2;
        Value[j, ACol] := v1;
      end else begin
        Value[j, ACol] := v1;
        Valid[i, ACol] := False;
      end
    else if GetCell(j, ACol, v2) then begin
      Value[i, ACol] := v2;
      Valid[j, ACol] := False;
    { else no action } end;
end;

procedure TNumberGrid.Invalidate;
begin
  { Fudge to prevent some excess Invalidates }
  if not InhibitNextInvalidate then
    if FUpdateCount = 0 then
      inherited Invalidate;
  InhibitNextInvalidate := False;
end;

procedure TNumberGrid.SetFocus;
begin
  { Windows/MDI bug fix? - be more forcefull in getting focus }
  inherited;
  {$IFDEF MSWINDOWS}
  if Windows.GetFocus <> Handle then
    Windows.SetFocus(Handle);
  {$ENDIF}
end;

procedure TNumberGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell: TGridCoord;
  XFixed, YFixed:Boolean;
  NewSelection:TGridRect;
begin
  SavedOptions := [];
  InAreaSelection := asNone;
  Cell := MouseCoord(X, Y);
  if (Cell.X <> Col) or (Cell.Y <> Row) then EditorMode := False;
  if (Button = mbLeft) and not EditorMode and (goRangeSelect in Options)
     and not (ssCtrl in Shift) and not (ssAlt in Shift) then begin
    if (Cell.X <> Col) or (Cell.Y <> Row) then begin
      XFixed := (Cell.X < FixedCols) and (Cell.X >= 0);
      YFixed := (Cell.Y < FixedRows) and (Cell.Y >= 0);
      if XFixed or YFixed then begin
        if XFixed then begin
          if not YFixed then begin
            if not (goRowMoving in Options)then begin
              InAreaSelection := asRow;
              NewSelection.Right := FixedCols;
              NewSelection.Left := ColCount-1;
              NewSelection.Top := Cell.Y;
              NewSelection.Bottom := Cell.Y;
            end else
              InAreaSelection := asNone;
          end else begin
            InAreaSelection := asAll;
            NewSelection := DataExtent;
            Cell := NewSelection.TopLeft;
            NewSelection.TopLeft := NewSelection.BottomRight;
            NewSelection.BottomRight := Cell;
          end;
        end else begin { YFixed }
          if not (goColMoving in Options) then begin
            NewSelection.Left := Cell.X;
            NewSelection.Right := Cell.X;
            NewSelection.Bottom := FixedRows;
            NewSelection.Top := DataExtent.Bottom;
            InAreaSelection := asCol;
          end else
            InAreaSelection := asNone;
        end;
        if InAreaSelection <> asNone then begin
          InvalidateCell(Col, Row);
          SelectingAnchor := NewSelection.BottomRight;
          Selection := NewSelection;
          InvalidateCell(Col, Row);
        end;
      end else begin
        {$IFNDEF LCL_} { LCL does this anyway using ExtendedSelect }
          InhibitNextInvalidate := goEditing in Options;
          SavedOptions := Options*[goEditing];
          Options := Options - [goEditing];	{ Allow Selection of cells by mouse drag }
        {$ENDIF}
{       if ssShift in Shift then begin
          NewSelection := Selection;
          NewSelection.BottomRight := SelectingAnchor;
          Selection := NewSelection;
        end; }
      end;
    end;
  end else if (Button = mbRight) and not EditorMode then begin
    { Right click should ensure selection is under mouse }
    NewSelection := Selection;
    { Click test should include fixed cols }
    if NewSelection.Left = FixedCols then
      NewSelection.Left := 0;
    if NewSelection.Top = FixedRows then
      NewSelection.Top := 0;
    if (Cell.Y < NewSelection.Top) or (Cell.Y > NewSelection.Bottom)
       or (Cell.X < NewSelection.Left) or (Cell.X > NewSelection.Right) then
      if Cell.X < FixedCols then
        if Cell.Y < FixedRows then
          FocusCell(FixedCols, FixedRows, True)
        else
          FocusCell(FixedCols, Cell.Y, True)
      else if Cell.Y < FixedRows then
        FocusCell(Cell.X, FixedRows, True)
      else
        FocusCell(Cell.X, Cell.Y, True)
  end;
  inherited MouseDown(Button, Shift, X, Y);
  if (InAreaSelection <> asNone) and ((FGridState = gsRowSizing) or (FGridState = gsColSizing)) then begin
    CancelSelection;
    InAreaSelection := asNone;
  end;
end;

procedure TNumberGrid.TrackSelectedArea(X, Y: Integer);
var
  Cell:TGridCoord;
  NewSelection:TGridRect;
begin
  Cell := MouseCoord(X, Y);
  NewSelection.BottomRight := SelectingAnchor;
  case InAreaSelection of
   	asCol:
      if (Cell.X >= FixedCols) and (Cell.X < ColCount) and (NewSelection.Right <> Cell.X) then begin
        NewSelection.Left := Cell.X;
        NewSelection.Top := Selection.Bottom;
        InvalidateCell(Col, Row);
        Selection := NewSelection;
      end;
    asRow:
      if (Cell.Y >= FixedRows) and (Cell.Y < RowCount) and (NewSelection.Bottom <> Cell.Y) then begin
        NewSelection.Top := Cell.Y;
        NewSelection.Left := Selection.Right;
        InvalidateCell(Col, Row);
        Selection := NewSelection;
      end;
  end;
end;

procedure TNumberGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if InAreaSelection <> asNone then TrackSelectedArea(X, Y);
end;

procedure TNumberGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  InhibitNextInvalidate := SavedOptions <> [];
  Options := Options + SavedOptions;
  SavedOptions := [];
  if InAreaSelection <> asNone then TrackSelectedArea(X, Y);
  InAreaSelection := asNone;
end;

function TNumberGrid.DataExtent:TGridRect;
var
  r,c:Longint;
  s: string;
  { Can't use Max from Math unit due to clash with some other routine (IsNan?) }
  function Max(a,b: Longint): Longint;
  begin
    if a >= b then
      Result := a
    else
      Result := b;
  end;
begin
  Result.Right := FixedCols;
  Result.Left := ColCount-1;
  Result.Bottom := FixedRows;
  Result.Top := RowCount-1;
  for r := FixedRows to RowCount-1 do
    for c := FixedCols to ColCount-1 do begin
      if not GetValid(r,c) then
        { Not a number - check for non null label }
        if not GetCellLabel(r,c,s) or (s = '') then
          Continue;
      if Result.Right < c then Result.Right := c;
      if Result.Left > c then Result.Left := c;
      if Result.Top > r then Result.Top := r;
      if Result.Bottom < r then Result.Bottom := r;
    end;
end;

procedure TNumberGrid.SelectAll;
begin
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).SelectAll
  else begin
    InvalidateCell(Col, Row);
    Selection := DataExtent;
    InvalidateCell(Col, Row);
  end;
end;

procedure TNumberGrid.ClearArea(Area:TGridRect);
var
  r,c:Longint;
begin
  EditorMode := False;
  for r := Area.Top to Area.Bottom do
    for c := Area.Left to Area.Right do
      Valid[r,c] := False;
end;

procedure TNumberGrid.ClearSelection;
var
  Writeable: Boolean;
  r,c:Longint;
begin
  if not (goEditing in Options) then Exit;
  { Simulate DEL key }
  if EditorMode and (InplaceEditor is TCustomEdit) then
    with InplaceEditor as TCustomEdit do begin
      if SelLength <= 0 then
        SelLength := 1;
      ClearSelection;
    end
  else
    for r := Selection.Top to Selection.Bottom do
      for c := Selection.Left to Selection.Right do begin
        Writeable := True;
        if Assigned(FOnChangeCellQuery) then
          FOnChangeCellQuery(Self, c, r, Writeable);
        if Writeable then
          Valid[r,c] := False;
      end;
end;

procedure TNumberGrid.ClearColumn(ACol: Longint);
begin
  EditorMode := False;
  {$IFDEF MAKEFROMCOLVECTS}
  FMatrix.Column[ACol].Clear;
  {$ELSE}
  FMatrix.ClearRange(FMatrix.Index(ACol,1), FMatrix.RowCount);
  {$ENDIF}
  Changed(ACol, -1);
end;

procedure TNumberGrid.Clear;
{$IFDEF MAKEFROMCOLVECTS}
var
  c:Longint;
{$ENDIF}
begin
  EditorMode := False;
{$IFDEF MAKEFROMCOLVECTS}
  for c := 1 to ColCount-1 do
    FMatrix.Column[c].Clear;
{$ELSE}
  FMatrix.Clear;
{$ENDIF}
  Modified := False;
  Changed(-1, -1);
end;

procedure TNumberGrid.Blank;
begin
  RowCount := 2;
  ColCount := 2;
  Title[1] := ' ';
  Lbl[1] := ' ';
  Valid[1,1] := False;
  Caption := '';
end;

procedure TNumberGrid.SaveAreaToStream(Stream:TStream; Area:TGridRect);
var
  s:TTextStream;
  r,c:Longint;
  v: Double;
  buffer: string;
begin
  s := nil;
  try
    s := TTextStream.Create(Stream);
    for r := Area.Top to Area.Bottom do begin
      for c := Area.Left to Area.Right do begin
        if GetCellLabel(r, c, buffer) then
          s.PutString(buffer)
        else if GetCell(r, c, v) then
          s.PutFloat(v);
        if c <> Area.Right then s.Tab;
      end;
      s.Newline;
    end;
  finally
    s.Free;
  end;
end;

procedure TNumberGrid.SaveToStream(Stream:TStream);
begin
  SaveAreaToStream(Stream, DataExtent);
end;

procedure TNumberGrid.OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint);
var
  s:TTextStream;
  r,c:Longint;
  Writeable: Boolean;
begin
  s := nil;
  try
    BeginUpdate;
    s := TTextStream.Create(Stream);
    s.NULLisEOF := True;
    r := StartRow;
    repeat
      for c := StartCol to ColCount-1 do
        if s.GetFreeItem then begin
          Writeable := True;
          if Assigned(FOnChangeCellQuery) then
            FOnChangeCellQuery(Self, c, r, Writeable);
          if Writeable then begin
            if r >= RowCount then
              RowCount := RowCount * 2;
            Value[r,c] := s.RealToken;
          end;
        end;
      Inc(r);
    until s.NextLine;;
  finally
    EndUpdate;
    s.Free;
  end;
end;

procedure TNumberGrid.LoadFromStream(Stream:TStream);
begin
  OffsetLoadFromStream(Stream, 1, 1);
end;

function TNumberGrid.CanCopyToClipboard: Boolean;
begin
  if EditorMode then
    Result := (InplaceEditor <> nil) and (InplaceEditor is TCustomEdit) and (TCustomEdit(InplaceEditor).SelLength > 0)
  else
    Result := True; { Always one cell selected }
end;

procedure TNumberGrid.CopyToClipboard;
var
  s:THugeMemoryStream;
  ch:Char;
begin
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).CopyToClipboard
  else begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      SaveAreaToStream(s, Selection);
      ch := #0;
      s.Write(ch, 1);
      s.CopyToClipboard(CF_TEXT);
    finally
      s.Free;
    end;
  end;
end;

function TNumberGrid.CanPasteFromClipboard: Boolean;
begin
  Result := (goEditing in Options) and ClipboardHasText;
end;

procedure TNumberGrid.PasteFromClipboard;
var
  s: THugeMemoryStream;
  r, c: Longint;
  function MoreThanOneWord: Boolean;
  var
    p: PChar;
  begin
    p := s.Memory;
    while (p^ <> #0) and (p^ <= ' ') do Inc(p);
    while (p^ <> #0) and (p^ > ' ') do Inc(p);
    while (p^ <> #0) and (p^ <= ' ') do Inc(p);
    Result := p^ > ' ';
  end;
  { Can't use Min from Math unit due to clash with some other routine (IsNan?) }
  function Min(a,b: Longint): Longint;
  begin
    if a <= b then
      Result := a
    else
      Result := b;
  end;
begin
  if ClipboardHasText then begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      s.SetFromClipboard(CF_TEXT);
      { As a heuristic, assume anything with more than one word
        should be pasted into grid as a whole }
      if not EditorMode or MoreThanOneWord then begin
        EditorMode := False;
        r := Min(Selection.Top, Selection.Bottom);
        c := Min(Selection.Left, Selection.Right);
        OffsetLoadFromStream(s, r, c);
        Exit;
      end;
    finally
      s.Free;
    end;
  end;
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).PasteFromClipboard
end;

procedure TNumberGrid.MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
var
  Coord: TGridCoord;
begin
  Coord := MouseCoord(X, Y);
  ACol := Coord.X;
  ARow := Coord.Y;
end;

function TNumberGrid.NaturalWidth: Integer;
var
  i: Longint;
begin
  {$IFNDEF CLX_}
  if VisibleRowCount < RowCount-1 then
    Result := GetSystemMetrics(SM_CXVSCROLL)
  {$ELSE}
  if VScrollBar.Visible then
    Result := VScrollBar.Width
  {$ENDIF}
  else
    Result := 0;
  {$IFNDEF LCL_}
  if BorderStyle = bsSingle then
  {$IFDEF WIN32}
    Inc(Result, 4);
  {$ELSE}
    Inc(Result, 1);
  {$ENDIF}
  {$IFDEF UNIX}
  Inc(Result, -2);
  {$ENDIF}
  {$ENDIF}
  for i := 0 to ColCount - 1 do
    Inc(Result, ColWidths[i]+GridLineWidth);
end;

end.

