{ Loosely based on Z_Prof tool by A. Baars (antonie.baars@wxs.nl) }

unit ProfileForm;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Messages, Classes, Graphics, Controls, Forms, Dialogs,
     Grids, Buttons, ExtCtrls, StdCtrls, NGrid, GridEx;

type

  TProfilerForm = class(TForm)
    Panel: TPanel;
    Clearbutton: TSpeedButton;
    RadioGroup: TRadioGroup;
    Clock: TLabel;
    RecalibrateButton: TSpeedButton;
    OverheadE: TLabel;
    NG: TNumberGrid;
    Label1: TLabel;
    Label2: TLabel;
    OverheadI: TLabel;
    Label4: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure ClearbuttonClick(Sender: TObject);
    procedure RadioGroupClick(Sender: TObject);
    procedure Recalibrate(Sender: TObject);
    procedure tablecontent(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure NGDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure NGGetCellLabel(Sender: TObject; Col, Row: Integer;
      var s: String; var IsText: Boolean);
  public
    destructor Destroy; override;
    procedure UpdateActions; override;
    procedure PopulateTable;
  private
    CPUclock: Double;
  end;

  function ProfilerForm: TProfilerForm;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     Clock, Math, MathSup, SysUtils, Profile, UCompat;

{$IFDEF MSWINDOWS}
{$IFNDEF LCL_}
{$R *.DFM}
{$ENDIF}
{$ENDIF}

var
  TheProfilerForm: TProfilerForm;

function ProfilerForm: TProfilerForm;
begin
  if not Assigned(TheProfilerForm) then
    Result := TProfilerForm.Create(Application)
  else
    Result := TheProfilerForm;
end;

procedure TProfilerForm.FormCreate(Sender: TObject);
begin
  if not Assigned(TheProfilerForm) then
    TheProfilerForm := Self;
  Recalibrate(Self);
end;

procedure TProfilerForm.tablecontent(Sender: TObject; Col, Row: Longint; Rect: TRect; State: TGridDrawState);
var
  mtext,mmtext: string;
  div1: Double;
  Digits: Integer;
begin
  mtext:='';
  mmtext:='';
  div1:=1;
  Digits := 2;
  case radiogroup.itemindex of
     0: begin mmtext := '/s';  div1 :=1/CPUclock;   end;
     1: begin mmtext := '/ms'; div1 :=1e3/CPUclock; end;
     2: begin mmtext := '/s'; div1 :=1e6/CPUclock; end;
     3: begin mmtext := '';    div1 := 1; Digits := 0; end;
   end;
  if row=0 then
    case col of
      0: mtext:='Passes';
      1: mtext:='Time' + mmtext;
      2: mtext:='Last' + mmtext;
      3: mtext:='Mean' + mmtext;
      4: mtext:='Fastest'+ mmtext;
      5: mtext:='Routine';
    end
  else with Marks[row-1] do
    case col of
      0: mtext := inttostr(Passes);
      1: if passes>0 then mtext := FloatToStrF(div1*TotalTime, fffixed, 8, Digits);
      2: if passes>0 then mtext := FloatToStrF(div1*PrevTime, fffixed, 8, Digits);
      3: if passes>0 then mtext := Floattostrf(div1*TotalTime/passes, fffixed, 8, Digits);
      4: if passes>0 then mtext := FloatToStrF(div1*Fastest, fffixed, 8, Digits);
      5: mtext := Name;
    end;
  with Rect, TDrawgrid(sender).Canvas do
    TextRect(Rect, Left + (Right - Left - TextWidth(mtext)) div 2,
                 Top + (Bottom - Top - TextHeight(mtext)) div 2, mtext);
end;

procedure TProfilerForm.ClearbuttonClick(Sender: TObject);
var
  i: Integer;
begin
  for i :=0 to High(Marks) do
    with Marks[i] do begin
      Passes := 0;
      TotalTime := 0;
      Active := False;
    end;
  MarkLevel := 0;
  PopulateTable;
end;

procedure TProfilerForm.RadioGroupClick(Sender: TObject);
begin
  PopulateTable;
end;

procedure TProfilerForm.Recalibrate(Sender: TObject);
  function FormatTime(Cycles: Integer): string;
  var
    o: Double;
  begin
    o := Cycles/CPUclock*1e6;
    if o < 0.9 then
      Result := Format('%f ns, %d Cycles', [o*1e3, Cycles])
    else
      Result := Format('%f s, %d Cycles', [o, Cycles]);
  end;
begin
  CPUclock := ClockFrequency;
  if CPUclock >= 1e6 then
    Clock.caption:=Format('Clock: %f MHz', [CPUclock/1e6])
  else
    Clock.caption:=Format('Clock: %f Hz', [CPUclock]);
  CalculateOverhead;
  OverheadI.Caption := FormatTime(Overhead);
  OverheadE.Caption := FormatTime(ExternalOverhead);
  PopulateTable;
end;

destructor TProfilerForm.Destroy;
begin
  if TheProfilerForm = Self then
    TheProfilerForm := nil;
  inherited;
end;

procedure TProfilerForm.UpdateActions;
begin
  inherited;
  if TimeChanged then begin
    TimeChanged := False;
    PopulateTable;
  end;
end;

procedure TProfilerForm.PopulateTable;
var
  Units: string;
  Divisor: Double;
  Digits, i, j: Integer;
  procedure SetCol(Col, Dig: Integer; V: Double);
  begin
    V := RoundTo(V, -Dig);
    NG.Value[i,Col] := V;
    Dig := NG.Canvas.TextWidth(DTosM(V, -1, Dig))+10;
    if NG.ColWidths[Col-1] < Dig then
      Ng.ColWidths[Col-1] := Dig;
  end;
begin
  for i := 0 to High(Marks) do
    if Marks[i].Name = '' then begin
      NG.RowCount := Max(i+1,2);
      Break;
    end;
  Digits := 2;
  case RadioGroup.ItemIndex of
    0: begin Units := '/s';  Divisor :=1/CPUclock;   end;
    1: begin Units := '/ms'; Divisor :=1e3/CPUclock; end;
    2: begin Units := '/s'; Divisor :=1e6/CPUclock; end;
    3: begin Units := '';    Divisor := 1; Digits := 0; end;
  end;
  with NG do begin
    Title[0] := 'Passes';
    Title[1] := 'Time' + Units;
    Title[2] := 'Last' + Units;
    Title[3] := 'Mean' + Units;
    Title[4] := 'Fastest'+ Units;
    Title[5] := 'Routine';
    for i := 0 to 5 do
      AutoAdjustColumn(i);
    for i := 1 to Min(Length(Marks), RowCount-1) do
      with Marks[i-1] do
        if Passes > 0 then begin
          SetCol(1, 0, Passes);
          SetCol(2, Digits, TotalTime*Divisor);
          SetCol(3, Digits, TotalTime/Passes*Divisor);
          SetCol(4, Digits, PrevTime*Divisor);
          SetCol(5, Digits, Fastest*Divisor);
          NG.ColWidths[5] := Max(NG.ColWidths[5], NG.Canvas.TextWidth(Name)+4)
        end else
          for j := 1 to 5 do
            Valid[i,j] := False;
  end
end;

procedure TProfilerForm.NGDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
begin
  if (ARow > 0) and (ARow <= Length(Marks)) then
    if ACol = 5 then
      TextInGrid(NG.Canvas, Rect, Marks[ARow-1].Name)
    else if NG.Valid[ARow,ACol+1] then
      TextInGrid(NG.Canvas, Rect, DtosM(NG.Value[ARow,ACol+1], -1, -1), taRightJustify);
end;

procedure TProfilerForm.NGGetCellLabel(Sender: TObject; Col, Row: Integer; var s: String; var IsText: Boolean);
begin
  if (Col = 5) and (Row > 0) then begin
    IsText := True;
    s := Marks[Row-1].Name;
  end;
end;

initialization
  ShowTimingAtEnd := False;
{$IFDEF LCL_}
  {$I ProfileForm.lrs}
{$ENDIF}
end.
