unit ProjAbout;
{ Automatic version number control. This is based on TProject by
  T S Eriksen, though almost all of the code has been moved to a
  design time only unit. }
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses
  {$IFDEF WIN32} Windows, {$ENDIF}
  {$IFNDEF CLX_} StdCtrls, Buttons, Controls, ExtCtrls, Forms,
  {$ELSE} QStdCtrls, QButtons, QControls, QExtCtrls, QForms, {$ENDIF}
  Classes;

type
  TFileType = (ftProgram,ftLibrary);

{CMW}
  TAboutBox = class(TForm)
    Panel1: TPanel;
    ProgramIcon: TImage;
    ProductName: TLabel;
    Description: TLabel;
    Details: TLabel;
    Button1: TButton;
  end;

  TProject = class(TComponent)
  private
    { user defined properties }
    FResFile        : string; { Resourcefile name }
    FCompany        : string; { stored in Ini-file }
    FCopyright      : string; { stored in Ini-file }
    FComment        : string; { default : <comments about this file/product> }
    FDescription    : string; { default : <Description of file> }
    FSpecialBuild   : string; { default : <How this build differs from a standard build> }
    FPrivateBuild   : string; { default : <Private information> }
    FProductname    : string; { default : <name of product this file ships with> }
    FProductVersion : string; { default : <0.0.0.0> }
    FOriginalName   : string; { default : <name of .exe file > }
    FTrademarks     : string; { default : <Legal trademarks used in this build> }
    FLanguage       : string; { default : UK_English }
    FCharset        : string; { default : Multilinqual }
    FFileType       : TFileType; { default : ffProgram }
    { Auto defined properties }
    { Internalname, }         { TProject.name }
    { Fileversion , }         { TProject.version + TProject.build }
    { FileFlags }             { ffprerease when version < 1.0, special/private build }
    { property data buffers etc }
    FAutoNaming  : Boolean;   { Styrer automatisk navngivning af project }
    // FProjectPath : string;    { hvor de skal kopieres fra }
    FBuild       : Integer;   { build nr, incr. ved component write state}
    FVersion     : string;    { Version number }
    FBuildDate   : string;
  protected
    { Protected declarations }
    procedure   Loaded; override;
    procedure   WriteState(Writer:TWriter);    override;
  public
    constructor Create(AComponent:TComponent); override;
    procedure   Execute; {CMW}
    { Public declarations }
  published
    { Published declarations }
    property AutoNaming  : Boolean   read FAutoNaming  write FAutoNaming;
    property Build       : Integer   read FBuild       write FBuild default 0;
    property Version     : string    read FVersion     write FVersion;
    { version resource properties }
    // property ProjectPath : string   read FProjectPath    write FProjectPath;
    property Company       : string read FCompany        write FCompany;
    property Copyright     : string read FCopyright      write FCopyright;
    property Comment       : string read FComment        write FComment;
    property Description   : string read FDescription    write FDescription;
    property SpecialBuild  : string read FSpecialBuild   write FSpecialBuild;
    property PrivateBuild  : string read FPrivateBuild   write FPrivateBuild;
    property Productname   : string read FProductName    write FProductname;
    property ProductVersion: string read FProductVersion write FProductVersion;
    property OriginalName  : string read FOriginalName   write FOriginalName;
    property Trademarks    : string read FTrademarks     write FTrademarks;
    property Language      : string read FLanguage       write FLanguage;
    property Charset       : string read FCharset        write FCharset;
    property FileType      : TFileType read FFileType    write FFileType;
    property ResourceFile  : string    read FResFile     write FResFile;
    property BuildDate     : string    read FBuildDate   write FBuildDate;
  end;

  TProjectHelper = class
    procedure DoWriteState(Project:TProject); virtual; abstract;
    procedure DoCreate(Project:TProject); virtual; abstract;
  end;

var
  DesignProjectHelper: TProjectHelper;

implementation

Uses
  {$IFDEF DFS_WIN16} IniFiles, WinProcs, WinTypes, FileCtrl, {$ENDIF}
  {$IFDEF LCL_} LResources, {$ENDIF}
  SysUtils, Sysinfo, UCompat;

{CMW}
{$IFDEF MSWINDOWS}
{$IFNDEF LCL}
{$R *.dfm}
{$ENDIF}
{$ENDIF}
{$IFDEF CLX_}
{$R *.xfm}
{$ENDIF}

function CheckStr(const s:string):string;
begin
  if (s = '') or (s[1] = '<') then
    Result := ''
  else
    Result := s;
end;

procedure TProject.Execute; {CMW}
var
  About:TAboutBox;
  ProdName:string;
  procedure Add(s: string);
  var
    c: string;
    {$IFDEF FPC}
    i: Integer;
    {$ENDIF}
  begin
    if s <> '' then begin
      {$IFDEF FPC}
      {Strip UTF8 characters for the time being}
      for i := 1 to Length(s) do
        if Ord(S[i]) > 128 then
          s[i] := ' ';
      {$ENDIF}
      c := About.Details.Caption;
      if c <> '' then
       About.Details.Caption := Concat(c, LineEnd, s)
      else
        About.Details.Caption := s;
    end;
  end;
begin
  About := nil;
  try
    About := TAboutBox.Create(Application);
    ProdName := CheckStr(ProductName);
    if ProdName = '' then
      ProdName := Application.Title;
    About.ProductName.Caption := ProdName;
    if not Assigned(Application.Icon) or Application.Icon.Empty then begin
      if Assigned(Application.MainForm) and Assigned(TForm(Application.MainForm).Icon) and not TForm(Application.MainForm).Icon.Empty then
      About.ProgramIcon.Picture.Icon := TForm(Application.MainForm).Icon;
    end else
      About.ProgramIcon.Picture.Icon := Application.Icon;

    About.Description.Caption := Concat(CheckStr(Description), LineEnd, Version, '.', IntToStr(Build), ' ', BuildDate);
    About.Details.Caption := '';
    Add(CheckStr(Comment));
    Add(CheckStr(Copyright));
    Add(CheckStr(Company));
    Add(Concat(LineEnd, DetermineCpu));
    About.ShowModal;
  finally
    About.Free;
  end;
end;

{ protected procedures }

procedure TProject.WriteState(Writer:TWriter);
begin
  { When something on a form has changed all components (on it) will
    be saved when next compiled - before that save, we'll just
    increase the Build number }
  if csDesigning in ComponentState then begin
    Inc(FBuild);                        { increase build number }
    DateTimeToString(FBuildDate, 'd mmm yyyy h:nn', Now);
    if DesignProjectHelper <> nil then
      DesignProjectHelper.DoWriteState(Self);
  end;
  Inherited WriteState(Writer);
end;

procedure TProject.Loaded;
var
  s : string;
  a : integer;
begin
  inherited Loaded;
  If (csDesigning in ComponentState) then begin
    { Denne blok udfres kun ved
      1) Load af projekt
      2) Visning af form/modul hvor TProject er lagret  }
    Inc(FBuild);                     { increase build number }
  end else begin
    { Denne blok udfrers kun ved
      1) Start af applikation }
    If (AutoNaming) then begin
      application.Title:=Name+' ver. '+Version+'.'+
                         IntToStr(Build);
      s:=application.title;
      For a:=1 to length(s) do if s[a]='_' then s[a]:=' ';
      application.title:=s;
    end;
  end;
end;

constructor TProject.Create(AComponent:TComponent);
begin
  Inherited Create(AComponent);
  { Only do this if designing - if runtime, it's not nesseary }
  If (csDesigning in ComponentState) and (DesignProjectHelper <> nil) then
    DesignProjectHelper.DoCreate(Self);
end;

initialization
  {$IFDEF LCL_}
  {$I projabout.lrs}
  {$ENDIF}
end.


