unit SlaveEdit;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface
{$IFDEF UNIX} uses X; {$ENDIF}


type
  TWindowMessage = {$IFDEF MSWINDOWS} Cardinal {$ELSE} TAtom {$ENDIF};

procedure LaunchEditor(FileName: string);
function SaveAllWindows: Boolean;
procedure BroadcastNotification(WindowMessage: TWindowMessage);
function BroadcastName(const Name: string):TWindowMessage;

implementation

uses {$IFDEF MSWINDOWS} ShellAPI, Windows, {$ENDIF}
     {$IFDEF UNIX}
       Xlib, Xatom, CompatMsg,
      {$IFDEF LCL_} GTKProc, {$ENDIF}
     {$ENDIF}
     Controls, Dialogs, SysUtils, StrUtils, Forms, LongName, Detach;


type

{ Typedef to help compatibility - the X Window/Windows Window Type}
{$IFDEF MSWINDOWS}
  TWnd = HWND;
{$ENDIF}
{$IFDEF UNIX}
  TWnd = {$IFDEF CLX_}Window{$ENDIF}{$IFDEF LCL_}TWindow{$ENDIF};
{$ENDIF}

  TWindowEnumerator = class
  protected
    ResultHWND: TWnd;
    function CallBack(hWnd: TWnd): Boolean; overload; virtual;
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; overload; virtual; abstract;
  public
    function DoEnumerate: TWnd;
  end;

  TWindowFinder = class(TWindowEnumerator)
  private
    FText: string;
  protected
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; override;
  public
    function FindWindow(Text: string): TWnd;
  end;

  TSaveAllWindows = class(TWindowEnumerator)
  public
    First: Boolean;
  protected
    function CallBack(hWnd: TWnd): Boolean; override;
    {$IFDEF UNIX}
    function StartSave: TWnd;
    function FinishSave: TWnd;
    {$ENDIF}
  end;

procedure BringWindowToFront(hWnd: TWnd);
begin
  {$IFDEF MSWINDOWS}
    {$IFDEF VER80}
      SetActiveWindow(hWnd);
    {$ELSE}
      SetForegroundWindow(hWnd);
    {$ENDIF}
    ShowWindow(hWnd, SW_RESTORE);
  {$ENDIF}
  {$IFDEF UNIX}
    XRaiseWindow(X11Display, hWnd);
    XSetInputFocus(X11Display, hWnd, RevertToNone, CurrentTime);
    XSync(X11Display, False);
  {$ENDIF}
end;

procedure LaunchEditor(FileName: string);
var
  Window: TWnd;
  Handle: Integer;
begin
  if FileName = '' then
    raise Exception.Create('Please choose a name for the file before editing it');
  with TWindowFinder.Create do
    try
      Window := FindWindow(ExtractFileName(FileName));
      if Window <> 0 then begin
        BringWindowToFront(Window);
        Exit;
      end;
    finally
      Free;
    end;
  if not FileExists(FileName) then
    if MessageDlg(Format('File %s does not exist. Create it?', [FileName]),
                  mtConfirmation, [mbYes, mbNo], 0) = mrYes then begin
      Handle := FileCreate(FileName);
      if Handle >= 0 then
        FileClose(Handle)
      else
        RaiseLastOSError;
    end;
  {$IFDEF MSWINDOWS}
    { Try default application associated with file }
    if ShellExecute(0, nil, PChar(ExtractFileName(FileName)), nil, PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
      { else try slave }
      if ShellExecute(0, nil, 'slave', PChar(ExtractFileName(FileName)), PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
        { use notepad as a last resort }
        if WinExec(PChar('notepad ' + FileName), SW_SHOW) < 32 then
          RaiseLastOSError;
  {$ENDIF}
  {$IFDEF UNIX}
  if not Detach1('slave', FileName) then
    if not Detach1('gedit', FileName) then
      if not Detach1('kedit', FileName) then
        if not Detach1('xedit', FileName) then
          RaiseLastOSError;
  {$ENDIF}
end;

{ TWindowEnumerator }

{$IFDEF MSWINDOWS}
function EnumWindowsProc(hWnd: TWnd; lParam: LPARAM): BOOL; export; stdcall;
begin
  Result := TWindowEnumerator(lParam).CallBack(hWnd);
end;
{$ENDIF}

{$IFDEF UNIX}
{ Duplicate logic of XmuClientWindow }
function ClientWindow(Window: TWindow): TWindow;
var
  WM_STATE: TAtom;
  function GetProperty(AWindow: TWindow): Boolean;
  var
    AType: TAtom;
    Format: Integer;
    nItems, After: DWord;
    data: PChar;
  begin
    AType := None;
    XGetWindowProperty(X11Display, AWindow, WM_STATE, 0, 0, False, AnyPropertyType,
                       @AType, @Format, @nItems, @After, @data);
    Result := Atype <> 0;
  end;
  function SearchChildren(AWindow: TWindow): TWindow;
  var
    Root, Parent: TWindow;
    Children: PWindow;
    nChildren, i: Integer;
  begin
    Result := 0;
    if (XQueryTree(X11Display, AWindow, @Root, @Parent, @Children, @nChildren) = 0) or not Assigned(Children) then
      Exit;
    try
      for i := 0 to nChildren-1 do
        if GetProperty(Children[i]) then begin
          Result := Children[i];
          Exit;
        end;
      for i := 0 to nChildren-1 do begin
        Result := SearchChildren(Children[i]);
        if Result <> 0 then
          Exit;
        end;
     finally
       XFree(children);
     end;
  end;
begin
  Result := Window; { Default is current window }
  WM_STATE := XInternAtom(X11Display, 'WM_STATE', True);
  if WM_STATE = 0 then Exit;
  if GetProperty(Window) then Exit; { On current Window }
  Result := SearchChildren(Window);
  if Result = 0 then
    Result := Window;
end;
{$ENDIF}

function TWindowEnumerator.CallBack(hWnd: TWnd): Boolean;
var
  buffer: string;
  n: Integer;
  {$IFDEF UNIX}
  Name: PChar;
  {$ENDIF}
begin
  Result := True; { Continue enumeration by default }
  {$IFDEF MSWINDOWS}
    if not IsWindowVisible(hWnd) then Exit;
    n := GetWindowTextLength(hWnd);
    if n <= 0 then Exit;
    SetLength(buffer, n);
    n := GetWindowText(hWnd, PChar(buffer), Length(buffer)+1);
    if n <= 0 then Exit;
    SetLength(buffer, n);
  {$ENDIF}
  {$IFDEF UNIX}
  if (XFetchName(X11Display, hWnd, @Name) <> 0) and Assigned(Name) then begin
    buffer := Name;
    XFree(Name);
  end else
    Exit; { No name - skip }
  {$ENDIF}
  Result := CallBack(hWnd, buffer);
end;

function TWindowEnumerator.DoEnumerate: TWnd;
{$IFDEF VCL_}
var
  EnumProc: Pointer;
{$ENDIF}
{$IFDEF UNIX}
var
  Root, Parent: TWindow;
  Children: PWindow;
  nChildren, i: Integer;
{$ENDIF}
begin
  ResultHWND := 0;
  {$IFDEF MSWINDOWS}
    {$IFDEF LCL_}
      EnumWindows(@EnumWindowsProc, LPARAM(self));
    {$ELSE}
      EnumProc := MakeProcInstance(@EnumWindowsProc, Application.Handle);
      try
        EnumWindows(EnumProc, LPARAM(self));
      finally
        FreeProcInstance(EnumProc);
      end;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF UNIX}
    // XSync(X11Display, False);
    if (XQueryTree(X11Display, RootWindow, @Root, @Parent, @Children, @nChildren) <> 0) and Assigned(Children) then
      try
        for i := 0 to nChildren-1 do
          if not CallBack(ClientWindow(Children[i])) then
            Break;
      finally
        XFree(Children);
      end;
  {$ENDIF}
  Result := ResultHWND;
end;

{ TWindowFinder }

function TWindowFinder.CallBack(hWnd: TWnd; WindowName: string): Boolean;
var
  p: Integer;
  Tail: string;
begin
  p := Pos(FText, AnsiLowerCase(WindowName));
  Result := True;
  if p = 0 then Exit;
  if p < Length(WindowName) - Length(FText) + 1 then begin
    { Not at the end of the string - look for "file - application name" }
    Tail := Copy(WindowName, p+Length(FText), 3);
    if Tail <> ' - ' then
      Exit;
  end;
  ResultHWND := hWnd;
  Result := False;
end;

function TWindowFinder.FindWindow(Text: string): TWnd;
begin
  FText := AnsiLowerCase(Text);
  Result := DoEnumerate;
end;

{ TSaveAllWindows }

var
  WM_Save_All: TWindowMessage = 0;

function TSaveAllWindows.CallBack(hWnd: TWnd): Boolean;
var
{$IFDEF MSWINDOWS}
  r: LRESULT;
{$ENDIF}
{$IFDEF UNIX}
  Prop: Integer;
  Event: TXClientMessageEvent;
  function GetProperty: Integer;
  var
    nItems, BytesLeft: Integer;
    ActualType: TAtom;
    ActualFormat: Integer;
    Items: PInteger;
  begin
    Result := 0;
    if (XGetWindowProperty(X11Display, hWnd, WM_Save_All, 0, 1, False, XA_INTEGER,
                           @ActualType, @ActualFormat, @nItems, @BytesLeft, @Items) = Success)
       and Assigned(Items) then begin
      Result := Items^;
      XFree(Items);
    end;
  end;
{$ENDIF}
begin
  Result := True;
  {$IFDEF MSWINDOWS}
    r := SendMessage(hWnd, WM_Save_All, 0, 0);
    if r = 272727 then begin
      ResultHWND := hWnd;
      Result := False;
    end;
  {$ENDIF}
  {$IFDEF UNIX}
    Prop := GetProperty;
    if Prop <> 0 then begin
      WriteLn( hWnd, ' ', Prop);
      ResultHWND := hWnd;
      if First then begin
        Event.display := X11Display;
        Event._type := ClientMessage;
        Event.message_type := WM_Save_All;
        Event.format := 32;
        Event.data.l[0] := MainWindow;
        Event.window := hWnd;
        XSendEvent(X11Display, hWnd,  False, NoEventMask, @Event);
      end else
        Result := False;
    end;
  {$ENDIF}
end;

{$IFDEF UNIX}
function TSaveAllWindows.StartSave: TWnd;
begin
  First := True;
  Result := DoEnumerate;
  XFlush(X11Display);
end;

function TSaveAllWindows.FinishSave: TWnd;
var
  Start: TDateTime;
begin
  First := False;
  Start := Now;
  repeat
    XFlush(X11Display);
    Application.ProcessMessages;
    Result := DoEnumerate;
  until (Result = 0) or (Now - Start > 2/SecsPerDay);
end;
{$ENDIF}

function SaveAllWindows: Boolean;
var
  Window: TWnd;
begin
  Result := True; { Assume OK }
  if WM_Save_All = 0 then
    WM_Save_All :=
  {$IFDEF MSWINDOWS}
        RegisterWindowMessage('WM_Save_All');
  {$ENDIF}
  {$IFDEF UNIX}
        XInternAtom(X11Display, 'WM_Save_All', False);
  {$ENDIF}
  with TSaveAllWindows.Create do
    try
      {$IFDEF UNIX}
      Window := StartSave;
      if Window <> 0 then
        Window := FinishSave;
      {$ENDIF}
      {$IFDEF MSWINDOWS}
      Window := DoEnumerate;
      {$ENDIF}
      if Window <> 0 then begin
        BringWindowToFront(Window);
        Result := False;
      end;
    finally
      Free;
    end;
end;

procedure BroadcastNotification(WindowMessage: TWindowMessage);
begin
  {$IFDEF MSWINDOWS}
  PostMessage(HWND_BROADCAST, WindowMessage, 0, 0);
  {$ENDIF}
  { TODO: UNIX }
end;

function BroadcastName(const Name: string):TWindowMessage;
begin
  Result :=
  {$IFDEF MSWINDOWS}
        RegisterWindowMessage(PChar(Name));
  {$ENDIF}
  {$IFDEF UNIX}
        XInternAtom(X11Display, PChar(Name), False);
  {$ENDIF}
  BroadcastNotification(Result);
end;

end.
